/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "proto.h"

void assemble(void) {
	char *line, *token, *operand;
	struct instruction *ins;
	struct pseudo *psop;
	struct segment *seg = NULL;
	int linelen;
	u16 val;
	
	rewind_buffer();

	loc = 0;

	line = get_buffer_line();
	while(line) {
		linelen = strlen(line) - 6;
		
		if(*line == '/') {
			get_first_token(line);
			file_add(get_token());
			}
		else if(*line == '\\') {
			file_remove();
			}
		else {
			file_top->lineno = atoi(get_first_token(line));

			token = get_token();
			if(token) {
		
				if(islabel(token)) {
					if((token[0] == '_') && (token[1] == '_')) {
						token[strlen(token) - 1] = '\0';
						add_symbol_value(token, loc);
						}
					token = get_token();
					}
			
				else if(issymbol(token)) {
					get_token();
					operand = get_token();

					if(token[0] == '@')
						add_symbol_value(token, new_atoi(operand, 0xffff));

					token = NULL;
					}

				if(token) {
					if(get_opcode(token)) {
						operand = strtok(NULL, "\0\r\n");
						ins = get_ins(token, operand);
				
						mem[loc] = ins->opcode;
						loc++;
						if(ins->length > 1) {
							mem[loc] = ins->oper_low;
							loc++;
							if(ins->length > 2) {
								mem[loc] = ins->oper_high;
								loc++;
								}
							}
						}
				
					else if((psop = get_pseudo(token))) {
						switch(psop->magic) {
					
							case P_ORG:
								if(!seg && loc)
									seg = add_segment(0);
								if(seg) seg->len = loc - seg->org;

								seg = add_segment(new_atoi(get_token(), 0xffff));
								loc = seg->org;

								break;
							
							case P_BYTE:
								token = strtok(NULL, ",\r\n");
								while(token) {
									mem[loc] = eval(token, 0xff);
									loc++;

									token = strtok(NULL, ",\r\n");
									}
								break;
							
							case P_WORD:
								token = strtok(NULL, ",\r\n");
								while(token) {
									val = eval(token, 0xffff);
									
									mem[loc] = val & 0xff;
									mem[loc + 1] = val >> 8;
									loc += 2;
									
									token = strtok(NULL, ",\r\n");
									}
								break;
							
							case P_ASCII:
								token = strchr(token, '\0');

								token++;

								while((*token != '\"') && (*token != '\0')) token++;
								token++;

								while((*token != '\"') && (*token != '\0')) {
									if(*token == '\\') token++;

									mem[loc] = *token;
									loc++;
									
									token++;
									}
								token++;
							
								while(isspace(*token)) token++;

								while(token)
									token = get_token();

								break;
							
							case P_EXPORT:
								token = get_token();
								break;
							}
						}

					else
						do_error("invalid opcode '%s'\n", token);
					}
				}
			}

		line = get_buffer_line();
		}
	
	if(!seg && loc)
		seg = add_segment(0);
	if(seg) seg->len = loc - seg->org;
	}
