/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "proto.h"

void build_symtab(void) {
	char *line, *token, *operand;
	struct instruction *ins;
	struct pseudo *psop;
	struct symbol *sym;
	int linelen;
	u16 val;

	rewind_buffer();

	loc = 0;

	line = get_buffer_line();
	while(line) {
		linelen = strlen(line) - 6;

		if(*line == '/') {
			get_first_token(line);
			file_add(get_token());
			}
		else if(*line == '\\') {
			file_remove();
			}
		else {
			file_top->lineno = atoi(get_first_token(line));
			
			token = get_token();
			if(token) {
				
				/* if this is a label, add it to the symtab */
				if(islabel(token)) {
					token[strlen(token) - 1] ='\0';
					add_symbol_value(token, loc);

					token = get_token();
					}

				/* might be a symbol assignment */
				else if(issymbol(token)) {
					operand = get_token();
					if(!operand)
						do_error("bare symbol '%s'\n", token);
					if(strcmp(operand, "="))
						do_error("expected assignment operator\n");

	                                operand = get_token();
					if(!operand)
						do_error("invalid assignment\n");

					val = new_atoi(operand, 0xffff);
					if((sym = get_symbol(token)) && (val <= 0xff) && !sym->vol)
						do_warning("possible bad forward reference");

					add_symbol_value(token, val);

					token = NULL;
					}

				/* otherwise, its an instruction */
				if(token) {

					/* if this thing is an opcode, handle it */
					if(get_opcode(token)) {
						operand = strtok(NULL, "\0\r\n");
						ins = get_ins(token, operand);
						loc += ins->length;
						}

					/* otherwise its a pseudo */
					else if((psop = get_pseudo(token))) {
						switch(psop->magic) {

							case P_ORG:
								token = get_token();
								if(!token)
									do_error("org directive has no operand\n");

								loc = new_atoi(token, 0xffff);
								break;

							case P_BYTE:
								token = strtok(NULL, ",\r\n");
								if(!token)
									do_error("byte directive must be followed by one or more values\n");
								while(token) {
									loc++;
									token = strtok(NULL, ",\r\n");
									}
								break;
							
							case P_WORD:
								token = strtok(NULL, ",\r\n");
								if(!token)
									do_error("word directive must be followed by one or more values\n");
								while(token) {
									loc += 2;
									token = strtok(NULL, ",\r\n");
									}
								break;
							
							case P_ASCII:
								token = strchr(token, '\0');
								if(token == (char *) ((line) + linelen - 1)) goto bad_ascii;
								token++;

								while((*token != '\"') && (*token != '\0')) token++;
								if(*token != '\"') goto bad_ascii;
								token++;

								while((*token != '\"') && (*token != '\0')) {
									if(*token == '\\') token++;
									token++;
									loc++;
									}
								if(*token != '\"') goto bad_ascii;
								token++;
							
								while(isspace(*token)) token++;
								if(!isspace(*token) && (*token != '\0')) goto bad_ascii;

								while(token)
									token = get_token();

								break;

								bad_ascii:
									do_error("ascii directive must be followed by a string enclosed in quotes\n");
								break;
							
							case P_EXPORT:
								token = get_token();
								if((token[0] == '_') && (token[1] == '_'))
									do_error("can't export a volatile symbol\n");

								if(!(sym = get_symbol(token)))
									sym = add_symbol(token);
								sym->exported = 1;
								break;
							}
						}
					else
						do_error("invalid label or opcode '%s'\n", token);
					}
				}

			token = get_token();
			if(token)
				do_error("unexpected extra tokens in input file\n");

			}

		line = get_buffer_line();
		}
	}
