/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "types.h"
#include "proto.h"

void do_error(char *format, ...) {
	va_list args;
	struct fileinfo *scan;

	if(file_top) {
		sprintf(buffer, "%s:%d: ", file_top->filename, file_top->lineno);
	
		va_start(args, format);
		vsprintf(strchr(buffer, '\0'), format, args);
		va_end(args);

		fputs(buffer, stderr);

		for(scan = file_top->prev; scan; scan = scan->prev)
			printf("\tin %s:%d\n", scan->filename, scan->lineno);
		}
	else {
		sprintf(buffer, "orgasm: ");
	
		va_start(args, format);
		vsprintf(strchr(buffer, '\0'), format, args);
		va_end(args);

		fputs(buffer, stderr);
		}
	
	free_all();

	exit(1);
	}

void do_warning(char *format, ...) {
	va_list args;
	struct fileinfo *scan;

	if(file_top) {
		sprintf(buffer, "%s:%d: warning: ", file_top->filename, file_top->lineno);
	
		va_start(args, format);
		vsprintf(strchr(buffer, '\0'), format, args);
		va_end(args);

		fputs(buffer, stderr);

		for(scan = file_top->prev; scan; scan = scan->prev)
			printf("\tin %s:%d\n", scan->filename, scan->lineno);
		}
	else {
		sprintf(buffer, "orgasm: warning: ", file_top->filename, file_top->lineno);
	
		va_start(args, format);
		vsprintf(strchr(buffer, '\0'), format, args);
		va_end(args);

		fputs(buffer, stderr);
		}
	}

void do_oom() {
	fputs("orgasm: out of memory\n", stderr);
	exit(2);
	}
