/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef HAVE_DMALLOC_H
 #include <dmalloc.h>
#endif

#include "types.h"
#include "proto.h"

struct fileinfo *file_top = NULL;

struct fileinfo *file_add(char *filename) {
	struct fileinfo *fileinfo;

	if(!(fileinfo = (struct fileinfo *) malloc(sizeof(struct fileinfo))))
		do_oom();
	
	if(!(fileinfo->filename = strdup(filename)))
		do_oom();

	fileinfo->lineno = 0;

	fileinfo->prev = file_top;
	file_top = fileinfo;

	return fileinfo;
	}

void file_remove(void) {
	struct fileinfo *fileinfo;

	if(file_top) {
		fileinfo = file_top;
		file_top = fileinfo->prev;

		free(fileinfo->filename);
		free(fileinfo);
		}
	}

void file_free(void) {
	struct fileinfo *scan, *prev;

	for(scan = file_top; scan; scan = prev) {
		prev = scan->prev;

		free(scan->filename);
		free(scan);
		}
	
	file_top = NULL;
	}
