/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "types.h"
#include "proto.h"

char *read_line(FILE *stream) {
	char *string = NULL;
	int string_length = 0;
	char c = '\0';

	while(!feof(stream) && (c != '\n')) {
   		c=fgetc(stream);
      
		if(c != EOF) {
			string_length++;
         
			if(!(string = (char *) realloc(string, (sizeof(char) * string_length))))
				do_oom();
         
			string[string_length - 1] = c;
			}
		}

	if(!string_length)
		return NULL;

	if(!(string=(char *) realloc(string, (sizeof(char) * (string_length + 1)))))
		do_oom();
		
	string[string_length] = '\0';

	return string;
	}

void strip_comment(char *line) {
	char *scan;

	for(scan = line; *scan; scan++)
		if((*scan == ';') || (*scan == '/') || (*scan == '\\'))
			*scan = '\0';
	}

inline char *get_first_token(char *line) {
	return strtok(line, " \t\r\n");
	}

inline char *get_token(void) {
	return strtok(NULL, " \t\r\n");
	}

int startswith(char *str, char *start) {
	int i, j;
	
	for(i = 0; isspace(str[i]); i++);

	for(j = 0; j < strlen(start); i++, j++)
		if((i == strlen(str)) || (tolower(str[i]) != tolower(start[j])))
			return 0;

	return 1;
	}

int endswith(char *str, char *end) {
	int i, j;
	
	for(i = strlen(str) - 1; isspace(str[i]); i--);

	for(j = strlen(end) - 1; j >= 0; i--, j--)
		if((i < 0) || (tolower(str[i]) != tolower(end[j])))
			return 0;

	return 1;
	}

void cut(char *str, int n) {
	int i;

	for(i = strlen(str) - 1; isspace(str[i]); i--);

	if((n <= 0) || (i = (i - n + 1)) < 0)
		return;
	
	str[i] = '\0';
	}

void chop(char **str) {
	char *end;
	
	for(; isspace(**str); (*str)++);

	end = strchr(*str, '\0');
	end--;

	for(; isspace(*end); *end = '\0', end--);
	}
