/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>

#include "types.h"
#include "proto.h"

void preprocess(char *infilename) {
	FILE *file;
	char *read, *line, *token;
	int end = 0, lineno = 0, debug = 0;
	struct macro *macro = NULL;

	file_add(infilename);
	
	if(!(file = fopen(infilename, "r")))
		do_error("couldn't open file for reading\n");
	
	sprintf(buffer, "/ %s", infilename);
	add_buffer_line(buffer);

	loc = 0;
	
	read = read_line(file);

	while(read && !end) {
		line = read;
		lineno++;
		file_top->lineno++;
		
		strip_comment(line);
		chop(&line);

		if(startswith(line, ".macro")) {
			if(macro)
				do_error("cannot nest macro definitions\n");

			token = get_first_token(line);
			token = get_token();
			if(!token)
				do_error("expected macro name\n");

			macro = add_macro(token);
			file_add(token);

			line = NULL;
			}
		else if(!strcmp(line, ".endmacro")) {
			if(!macro)
				do_error("not in a macro definition\n");

			macro = NULL;

			file_remove();

			line = NULL;
			}
		else if(startswith(line, ".include")) {
			if(macro)
				do_error("can't include inside a macro\n");

			token = get_first_token(line);
			if(!(token = get_token()))
				do_error("must specify a filename for inclusion\n");

			sprintf(buffer, "%05d", lineno);
			add_buffer_line(buffer);

			preprocess(token);

			line = NULL;
			}
		else if(startswith(line, ".end")) {
			end = 1;

			line = NULL;
			}
		else if(startswith(line, ".debug")) {
			if(debug)
				do_error("already in debug mode\n");
			if(!opt.debug)
				debug = 1;

			line = NULL;
			}
		else if(startswith(line, ".enddebug")) {
			if(!debug)
				do_error("not in debug mode\n");

			debug = 0;

			line = NULL;
			}
		
		if(line && !debug) {
			if(*line == '%') {
				line++;
				token = get_first_token(line);
				
				if(!(macro = get_macro(token)))
					do_error("undefined macro '%s'\n", token);

				sprintf(buffer, "/ %s(macro)", token);
				add_buffer_line(buffer);

				add_buffer_macro(macro);

				sprintf(buffer, "\\");
				add_buffer_line(buffer);

				macro = NULL;
				}

			else if(*line) {
				if(!macro) {
					sprintf(buffer, "%05d %s", lineno, line);
					add_buffer_line(buffer);
					}
				else {
					sprintf(buffer, "%05d %s", file_top->lineno, line);
					add_macro_line(macro, buffer);
					}
				}
			}

		free(read);
		read = read_line(file);
		}
	
	if(read) free(read);
	
	fclose(file);

	file_remove();

	sprintf(buffer, "\\");
	add_buffer_line(buffer);
	}
