/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#ifndef __ORGASM_PROTO_H__
#define __ORGASM_PROTO_H__

#include "types.h"

/* asm.c */
extern u16		loc;
extern u8		mem[];
extern struct options	opt;
extern char		buffer[];
extern void		free_all(void);

/* build_symtab.c */
extern void		build_symtab(void);

/* error.c */
extern void		do_error(char *, ...);
extern void		do_warning(char *, ...);
extern void		do_oom(void);

/* eval.c */
extern long int		new_atoi(char *expr, long int ubound);
extern u16		eval(char *expr, u16 ubound);

/* fileinfo.c */
extern struct fileinfo	*file_top;
extern struct fileinfo	*file_add(char *);
extern void		file_remove(void);
extern void		file_free(void);

/* macro.c */
extern struct macro	*add_macro(char *);
extern struct macro	*get_macro(char *);
extern void		init_macro(struct macro *);
extern void		add_macro_line(struct macro *, char *);
extern void		macro_free(void);

/* opcode.c */
extern struct opcode	*get_opcode(char *);
extern struct pseudo	*get_pseudo(char *);
extern struct instruction *get_ins(char *, char *);

/* parse.c */
extern char		*read_line(FILE *);
extern void		strip_comment(char *);
extern char		*get_first_token(char *);
extern char		*get_token(void);
extern int		startswith(char *, char *);
extern int		endswith(char *, char *);
extern void		cut(char *, int);
extern void		chop(char **);

/* preprocess.c */
extern void		preprocess(char *);

/* segment.c */
extern int		num_segs;
extern struct segment	*add_segment(u16);
extern void		save(char *);
extern void		segment_free(void);

/* symbol.c */
extern struct symbol	*add_symbol(char *);
extern struct symbol	*add_symbol_value(char *, u16);
extern struct symbol	*get_symbol(char *);
extern void		validate_symtab(void);
extern void		dump_symtab(void);
extern void		free_symtab(void);
extern void		export_symbols(char *);
extern int		issymchar(char);
extern int		issymbol(char *);
extern int		islabel(char *);

/* text_buffer.c */
extern void		add_buffer_line(char *);
extern void		add_buffer_macro(struct macro *);
extern void		rewind_buffer(void);
extern char		*get_buffer_line(void);
extern void		free_buffer(void);

/* additional headers for systems with broken functions */

#ifndef HAVE_STRDUP
extern char		*strdup(const char *);
#endif

#endif
