/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "proto.h"

char **line_buffer = NULL;
int buffer_num_lines = 0;
int buffer_lineno = 0;

/* add a line to the buffer */

void add_buffer_line(char *line) {
	buffer_num_lines++;
	if(!(line_buffer = (char **) realloc(line_buffer, sizeof(char *) * buffer_num_lines)))
		do_oom();
	
	if(!(line_buffer[buffer_num_lines - 1] = strdup(line)))
		do_oom();
	}

/* add a macro to the buffer */

void add_buffer_macro(struct macro *macro) {
	char *token;
	int i = 0;

	token = strtok(NULL, ",\0");
	while(token) {
		sprintf(buffer, "0 @P%d = %s", i, token);
		add_buffer_line(buffer);
		i++;
		token = strtok(NULL, ",\r\n");
		}
	
	if(!(line_buffer = (char **) realloc(line_buffer, sizeof(char *) * (buffer_num_lines + macro->num_lines))))
		do_oom();
	
	for(i = 0; i < macro->num_lines; i++)
		add_buffer_line(macro->line[i]);
	}

inline void rewind_buffer(void) {
	buffer_lineno = 0;
	}

char *get_buffer_line(void) {
	char *line;

	if(buffer_lineno < buffer_num_lines) {
		if(!(line = strdup(line_buffer[buffer_lineno])))
			do_oom();
		buffer_lineno++;
		}
	else
		line = NULL;
	
	return line;
	}

void free_buffer(void) {
	int i;

	for(i = 0; i < buffer_num_lines; i++)
		free(line_buffer[i]);
	
	free(line_buffer);
	}
