/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#ifndef __ORGASM_TYPES_H__
#define __ORGASM_TYPES_H__

#include <stdio.h>

/* 8 and 16 bit types */
typedef unsigned char u8;
typedef unsigned short int u16;

struct opcode {
	char	*name;
	u8	implied;
	u8	relative;
	u8	immediate;
	u8	zeropage;
	u8	zeropagex;
	u8	zeropagey;
	u8	absolute;
	u8	absolutex;
	u8	absolutey;
	u8	indirect;
	u8	indirectx;
	u8	indirecty;
	};

#define P_ORG		(0x100)
#define P_BYTE		(0x101)
#define P_WORD		(0x102)
#define P_ASCII		(0x103)
#define P_EXPORT	(0x104)

struct pseudo {
	char	*name;
	int	magic;
	};

struct instruction {
	u8 opcode;
	u8 oper_low;
	u8 oper_high;
	int length;
	};

struct symbol {
	char	*name;
	u16	value;
	int	defined;
	int	exported;
	int	vol;
	char	*filename;
	int	lineno;

	struct symbol *next;
	};

struct segment {
	u16	org;
	u16	len;

	struct segment *next;
	};

#define F_CBM	(0)
#define F_SEG	(1)
#define F_MEM	(2)
#define F_MEM8	(3)
#define F_MEM16	(4)
#define F_MEM32	(5)

struct macro {
	char *name;
	char **line;
	int num_lines;
	int lineno;

	struct macro *next;
	};

struct options {
	char *infilename;
	char *outfilename;
	int format;
	int debug;
	int verbose;
	int export;
	u16 baseaddr;
	};

struct fileinfo {
	char *filename;
	int lineno;

	struct fileinfo *prev;
	};

#endif
