#include <stdio.h>
#include <stdlib.h>
#include "csim.h"

#define DUMP_LXT (0)
int quiet = 0;
int ignore_clocks=1;

int cycle=0;
char force_node_init_value='0';
int errors=0;

unsigned char mem[65536];

void *lxt=NULL;

int irnow=0, icnt=0;

int
clk,
reset, 
rdy,
d_H,
d_L, 
dout_H,
dout_L, 
ma_H,
ma_L,
rw,
isync,
a_H, a_L,
pc_H, pc_L,
x_H, x_L,
y_H, y_L,
sp_H, sp_L,
nf, vf, cf, zf, df, im,
ir_H, ir_L,
dummy
;


int GFI(char *n, int indx)
{
int rc=GetFacilityIndex(n, indx);
if(rc<0)
	{
	fprintf(stderr, "FAC '%s'[%d] not found!\n", n, indx);
	errors++;
	}

return(rc);
}

void init(void)
{
int i;

InitializeModelValue(force_node_init_value);
if(!ImportModel("circuit.csim"))
	{
	exit(255);
	}

SetSimClockMode(ignore_clocks);
SetGetFacMode(1);		/* autoeval if model dirty */
}


void tick(void)
{
int addr, isread;

if(ignore_clocks)
	{
	if(lxt)
		{
		SimulateCones();
		WriteLXTFrame(cycle);
		SimulateLatches();
		IncrementCycleCount(1);
		cycle++;
		}
		else
		{
		SimulateModel(1);
		cycle+=1;
		}
	}
	else
	{
	if(lxt)
		{
		SimulateCones();
		WriteLXTFrame(cycle);
		StkFac(0x0, clk, clk);

		SimulateLatches();
		IncrementCycleCount(1);
		cycle++;

		WriteLXTFrame(cycle);
		StkFac(0x1, clk, clk);		/* we're posedge */
		SimulateModel(1);
		cycle++;
		}
		else
		{
		StkFac(0x0, clk, clk);
		SimulateModel(1);
		StkFac(0x1, clk, clk);
		SimulateModel(1);
		cycle+=2;
		}
	}


addr = GetFac(ma_H, ma_L);
isread = GetFac(rw, rw);
irnow = GetFac(ir_H, ir_L);

if(isread)
	{
	int is = GetFac(isync, isync);
	icnt += is;

	if(!quiet)
		{
		if(is)
			{
			int a = GetFac(a_H, a_L);
			int pc = GetFac(pc_H, pc_L);
			int x = GetFac(x_H, x_L);
			int y = GetFac(y_H, y_L);
			int sp = GetFac(sp_H, sp_L);
	
			printf("\t(PC: %04X)(A: %02X)(X: %02X)(Y: %02X)(SP: %02X)   %c%c%c%c%c %c\n", pc, a, x, y, sp, 
					GetFac(nf, nf) ? 'N' : '-',
					GetFac(vf, vf) ? 'V' : '-',
					GetFac(cf, cf) ? 'C' : '-',
					GetFac(zf, zf) ? 'Z' : '-',
					GetFac(df, df) ? 'D' : '-',
					GetFac(im, im) ? 'I' : '-'
					);
			}
	
		printf("R: %04x (%02x) %s\n", addr, mem[addr],  is?"*":"");
		}
	StkFac(mem[addr], d_H, d_L);
	}
	else
	{
	mem[addr] = GetFac(dout_H, dout_L);
	if(!quiet)
		{
		printf("W: %04x (%02x)\n", addr, mem[addr]);
		}

	if(addr==0xf004)
		{
		int x = mem[0xf000] + 256*mem[0xf001];
		int y = mem[0xf002] + 256*mem[0xf003];
		printf("PLOT: %3d %3d\n", x, y);
		}
	}
}


int main(int argc, char **argv)
{
int i;
FILE *handle;

atexit((void(*)(void))UnloadModel);
init();

clk = GFI("cpu6502/clk", 0);
reset = GFI("cpu6502/reset", 0);
rdy = GFI("cpu6502/rdy", 0);
d_H = GFI("cpu6502/di", 7);
d_L = GFI("cpu6502/di", 0);
dout_H = GFI("cpu6502/do", 7);
dout_L = GFI("cpu6502/do", 0);
ma_H = GFI("cpu6502/ma", 15);
ma_L = GFI("cpu6502/ma", 0);
rw = GFI("cpu6502/rw", 0);
isync = GFI("cpu6502/sync", 0);

a_H = GFI("cpu6502/a", 7);
a_L = GFI("cpu6502/a", 0);
pc_H = GFI("cpu6502/pc", 15);
pc_L = GFI("cpu6502/pc", 0);
x_H = GFI("cpu6502/x", 7);
x_L = GFI("cpu6502/x", 0);
y_H = GFI("cpu6502/y", 7);
y_L = GFI("cpu6502/y", 0);
sp_H = GFI("cpu6502/sp", 7);
sp_L = GFI("cpu6502/sp", 0);

nf = GFI("cpu6502/nf", 0);
vf = GFI("cpu6502/vf", 0);
cf = GFI("cpu6502/cf", 0);
zf = GFI("cpu6502/zf", 0);
df = GFI("cpu6502/df", 0);
im = GFI("cpu6502/im", 0);

ir_H = GFI("cpu6502/ir", 7);
ir_L = GFI("cpu6502/ir", 0);

if(errors)
	{
	fprintf(stderr, "Errors encountered during init, exiting.\n");
	exit(255);
	}

SetLXTtmpsigMode(1);
SetLXTremoveNonvectorZeroBits(1);
if(DUMP_LXT) lxt=OpenLXT("zzz.lxt");

#if 0
mem[0xfffc]=0x12;
mem[0xfffd]=0xc0;

mem[0xc012]=0xa9;
mem[0xc013]=0x00;

mem[0xc014]=0xD0;
mem[0xc015]=0xfe;

mem[0xc016]=0x18;

mem[0xc017]=0x8d;
mem[0xc018]=0x06;
mem[0xc019]=0x10;

mem[0xc01a]=0x20;
mem[0xc01b]=0x04;
mem[0xc01c]=0x50;

mem[0xc01d]=0x4c;
mem[0xc01e]=0x12;
mem[0xc01f]=0xc0;

mem[0x5004]=0xE8;
mem[0x5005]=0xD0;
mem[0x5006]=0x01;
mem[0x5007]=0x02;
mem[0x5008]=0x60;
#endif

handle=fopen("mem.img", "rb");
if(handle)
	{
	int addr_h, addr_l, addr, ch;
	addr_l = fgetc(handle);
	addr_h = fgetc(handle);
	addr = addr_h*256 + addr_l;

	mem[0xfffc]=addr_l;
	mem[0xfffd]=addr_h;

	if(addr>=0)
	while(!feof(handle))
		{
		ch = fgetc(handle);
		mem[addr] = ch;
		addr++;
		addr &= 0xffff;
		}
	fclose(handle);
	}
	else
	{
	fprintf(stderr, "Could not open mem.img file, exiting!\n");
	exit(255);
	}

StkFac(0x1, reset, reset);
tick();
StkFac(0x0, reset, reset);
StkFac(0x1, rdy, rdy);

for(i=0;;i++) 
	{
	if(irnow == 0x02)
		{
		printf("Machine halted (%d instructions executed)\n", icnt);
		break;
		}
	tick();
	}

/**************************************/

exit(0);
}



