#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>

#include <sys/time.h>
#include <unistd.h>


int quiet = 1;
int cycle=0;
unsigned char mem[65536];
int irnow=0, icnt=0;



struct timeval tv_orig, tv_now;
XEvent evt;

/*------------------------------------------------------------------ */
#ifndef ITIMER_REAL              /* fixes compiles on suns and linux */
#define ITIMER_REAL     0
#define ITIMER_VIRTUAL  1
#define ITIMER_PROF     2

struct timeval {
        long    tv_sec;         /* seconds */
        long    tv_usec;        /* and microseconds */
};


struct  itimerval {
        struct  timeval it_interval;    /* timer interval */
        struct  timeval it_value;       /* current value */
};
#endif
/*------------------------------------------------------------------ */

#define XWIN_SIZE 256
#define YWIN_SIZE 200
int numpixels=0;

struct sigaction Sig;

/* 
 * Globals...
 */
Display *display;
int numplanes;
Window window;
Pixmap buffer;
int screen;
GC clear, background;
GC pixor;


void tick(void)
{
int addr, isread;

sim();

cycle++;
if((cycle%50000)==0)
	{
	unsigned long long tvorig = (tv_orig.tv_sec*1000000ULL) + tv_orig.tv_usec;
	unsigned long long tvnow, tvdelta;
	float tvfloat;

	gettimeofday(&tv_now, NULL);
	tvnow = (tv_now.tv_sec*1000000ULL) + tv_now.tv_usec;
	tvdelta = tvnow - tvorig;
	tvfloat = ((float)tvdelta)/1000000.0;

	printf("Cycle: %d Time: %.2f sec (%.2f cycles/sec)\n", cycle, tvfloat, ((float)cycle)/tvfloat);
	}

addr = getint_le_ma();
isread = getint_le_rw();
irnow = getint_le_ir();

if(isread)
	{
	int is = getint_le_sync();
	icnt += is;

	if(!quiet)
		{
		if(is)
			{
			int a = getint_le_a();
			int pc = getint_le_pc();
			int x = getint_le_x();
			int y = getint_le_y();
			int sp = getint_le_sp();
	
			printf("\t(PC: %04X)(A: %02X)(X: %02X)(Y: %02X)(SP: %02X)   %c%c%c%c%c %c\n", pc, a, x, y, sp, 
					getint_le_nf() ? 'N' : '-',
					getint_le_vf() ? 'V' : '-',
					getint_le_cf() ? 'C' : '-',
					getint_le_zf() ? 'Z' : '-',
					getint_le_df() ? 'D' : '-',
					getint_le_im() ? 'I' : '-'
					);
			}
	
		printf("R: %04x (%02x) %s\n", addr, mem[addr],  is?"*":"");
		}
	setint_le_di(mem[addr]);
	}
	else
	{
	mem[addr] = getint_le_do();
	if(!quiet)
		{
		printf("W: %04x (%02x)\n", addr, mem[addr]);
		}

        if(addr==0xf004)
                {
                int x = mem[0xf000] + 256*mem[0xf001];
                int y = mem[0xf002] + 256*mem[0xf003];
		if(!quiet)
			{
	                printf("PLOT: %3d %3d\n", x, y);
			}
		numpixels++;
		XDrawPoint(display,window,pixor,x,y);
                }
	}
}


/*
 * 25Hz screen IRQ...
 */
char init=0;
int lines=0;

void update()
{
int i;

while(1)
{
	tick();
        XFlush(display); 

        if(numpixels==10000)
        	{
		numpixels=0;
        	XCopyArea(display,buffer,window,pixor,0,0,XWIN_SIZE,YWIN_SIZE,0,0);
        	XFlush(display);
        	}

        while(XCheckMaskEvent(display,EnterWindowMask|LeaveWindowMask,&evt));
}
}


int main(int argc, char **argv)
{
int ii;
FILE *handle;

struct itimerval time;
XGCValues val;
unsigned long msk;
unsigned long i,j;

XTextProperty wtitle;
char *window_name="CPU6502 Simulation...";

sim_init();

if((display=XOpenDisplay((char *)0))==(Display *)0)
   {
   fprintf(stdout,"Cannot Open Display!\n");
   exit(0);
   }

screen=DefaultScreen(display);
numplanes=DisplayPlanes(display,screen);

window=XCreateSimpleWindow(display,RootWindow(display,screen),0,
       0,XWIN_SIZE,YWIN_SIZE, 1, WhitePixel(display,screen), 
WhitePixel(display,screen));

background=XCreateGC(display, window, 0, (XGCValues *)0);
XSetForeground(display,background,WhitePixel(display,screen));

msk=GCFunction;
val.function=GXclear;
clear=XCreateGC(display, window, msk, &val);


val.function=GXcopy;
pixor=XCreateGC(display, window, msk, &val);

XSetForeground(display, pixor, BlackPixel(display,screen));
XSetBackground(display, pixor, BlackPixel(display,screen));

XMapWindow(display, window);
XStringListToTextProperty(&window_name,1,&wtitle);
XSetWMName(display,window,&wtitle);

XSelectInput(display, window, EnterWindowMask|LeaveWindowMask);
XNextEvent(display, &evt);
XSelectInput(display, window, NoEventMask);

XMapWindow(display, window);
buffer=XCreatePixmap(display,window,XWIN_SIZE,YWIN_SIZE,numplanes);

XFillRectangle(display, buffer, clear, 0, 0, XWIN_SIZE, YWIN_SIZE);

handle=fopen("mem.img", "rb");
if(handle)
        {
        int addr_h, addr_l, addr, ch;
        addr_l = fgetc(handle);
        addr_h = fgetc(handle);
        addr = addr_h*256 + addr_l;

        mem[0xfffc]=addr_l;
        mem[0xfffd]=addr_h;

        if(addr>=0)
        while(!feof(handle))
                {
                ch = fgetc(handle);
                mem[addr] = ch;
                addr++;
                addr &= 0xffff;
                }
        fclose(handle); 
        }
        else
        {
        fprintf(stderr, "Could not open mem.img file, exiting!\n");
        exit(255);
        }

gettimeofday(&tv_orig, NULL);
setint_le_reset(1);
tick();
tick();
setint_le_reset(0);
setint_le_rdy(1);

XFillRectangle(display, buffer, background, 0, 0, XWIN_SIZE, YWIN_SIZE);
XFlush(display);




        Sig.sa_handler = update;
        Sig.sa_flags = SA_RESTART;
        sigaction((int)SIGALRM,&Sig,0);

time.it_interval.tv_sec=0;
time.it_interval.tv_usec=40000;  /* make this match your frame rate */
time.it_value.tv_sec=0;
time.it_value.tv_usec=1;

setitimer(ITIMER_REAL, &(time), (struct itimerval *)0);

while(1) { pause(); }

/**************************************/

for(ii=0;;ii++) 
	{
	if(irnow == 0x02)
		{
		printf("Machine halted (%d instructions executed)\n", icnt);
		break;
		}
	tick();
	}

/**************************************/

printf("Executed %d simulation cycles.\n", cycle);
exit(0);
}
