#include "aet.h"
#include "mem.h"

struct symbol **sym=NULL;
struct symbol **facs=NULL;
int numfacs=0;

struct symbol *firstnode=NULL;	/* 1st sym in aet */
struct symbol *curnode=NULL;	/* current loaded sym in aet loader */


/*
 * Generic hash function for symbol names...
 */
int hash(char *s)
{
char *p;
unsigned int h=0, g;
for(p=s;*p;p++)
        {
        h=(h<<4)+(*p);
        if(g=h&0xf0000000)
                {
                h=h^(g>>24);
                h=h^g;
                }   
        }
return(h%SYMPRIME);
}


/*
 * add symbol to table.  no duplicate checking
 * is necessary as aet's are "correct."
 */
struct symbol *symadd(char *name, int hv)
{
struct symbol *s;

s=(struct symbol *)Vcalloc(1,sizeof(struct symbol));
strcpy(s->name=(char *)Valloc(strlen(name)+1,0),name);
s->next=sym[hv];
sym[hv]=s;
return(s);
}


/*
 * find a slot already in the table...
 */
struct symbol *symfind(char *s)
{
int hv;
struct symbol *temp;

hv=hash(s);
if(!(temp=sym[hv])) return(NULL); /* no hash entry, add here wanted to add */

while(temp)
        {
        if(!strcmp(temp->name,s))
                {
                return(temp); /* in table already */    
                }
        if(!temp->next) break;
        temp=temp->next;
        }

return(NULL); /* not found, add here if you want to add*/
}

/*
 * build the aet database...
 */
int aet(char *fname)
{
FILE *fh;
char *aethdr="#aet";
char line[1024];
int regions=0;
int delta=0, t, i;
int whichsym, curtime=0;
struct Node *n;
struct symbol *s;
char *aetbuff;
char ch, heval;
struct HistEnt *he;

if(!(fh=fopen(fname,"rb")))
	{
        printf("Error opening .aet file '%s'.\n",fname);
        exit(0);
	}

fscanf(fh,"%s",line);
if(strcmp(aethdr,line))
	{
	printf("File missing #aet header.\n");
	exit(0);
	}

while(1)
	{
	if(feof(fh))
		{
		printf("Short read: file missing #aet footer.\n");
		exit(0);
		}

	fscanf(fh,"%s",line);
	if(!strcmp(aethdr,line)) break;	/* found footer */

	s=symadd(line,hash(line));
	numfacs++;

	if(!firstnode)
		{
		firstnode=curnode=s;
		}
		else
		{
		curnode->nextinaet=s;
		curnode=s;
		}

	n=(struct Node *)Vcalloc(1,sizeof(struct Node));
	n->nname=s->name;
	s->n=n;
	}

printf("[%d] facilities found in aet.\n",numfacs);

facs=(struct symbol **)Valloc(numfacs*sizeof(struct symbol *), 0);
curnode=firstnode;
for(i=0;i<numfacs;i++)
	{
	facs[i]=curnode;
	curnode=curnode->nextinaet;
	}
quicksort(facs,0,numfacs-1);

aetbuff=(char *)Valloc(numfacs+1, 0);	/* allocate aet single cycle buffer */

while(!feof(fh))
	{
	t=-1;
	fscanf(fh,"%d %s",&t,aetbuff);
	if(t==-1) break;		/* eof */

	if(curtime) { delta=t-curtime; }

	curtime=t*10;
/*	if(curtime==0) curtime=1; */  /* doesn't seem necessary */
	if(min_time<0) {min_time=curtime;}

	whichsym=0;

	s=firstnode;
	while(s)
		{
		n=s->n;
		ch=aetbuff[whichsym];		

		if(!s->h)
			{
			he=(struct HistEnt *)Vcalloc(1,sizeof(struct HistEnt));
			he->time=curtime;
			if(ch=='0') he->val=0; else
			if(ch=='1') he->val=3; else
			if(ch=='x') he->val=1; else
			he->val=2;

			n->curr=he;
			n->head.next=he;

			s->h=he;
			regions++;
			}
			else 
			{
			if(ch=='0') heval=0; else
			if(ch=='1') heval=3; else
			if(ch=='x') heval=1; else
			heval=2;

			if(s->h->val!=heval) /* same region == go skip */ 
				{
				he=(struct HistEnt *)Vcalloc(1,sizeof(struct HistEnt));
				he->time=curtime;
				he->val=heval;

				s->h->next=he;
				s->h=he;
				regions++;
				}
			}

		s=s->nextinaet;
		whichsym++;
		}
	}

printf("[%d] regions found in aet.\n",regions);

s=firstnode;
while(s)
	{
	if(!delta) delta=1000;

	n=s->n;
	he=(struct HistEnt *)Vcalloc(1,sizeof(struct HistEnt));
	he->time=curtime+(4*delta);
	he->val=1;

	if(!s->h)
		{
		n->curr=he;
		n->head.next=he;

		s->h=he;
		}
		else
		{
		s->h->next=he;
		s->h=he;
		}
	
	s=s->nextinaet;
	}

fclose(fh);
printf("[%.2f] start time.\n[%.2f] end time.\n",
		((float)min_time)/100.0,((float)curtime)/100.0);

if(min_time==curtime)
	{
	printf("times range is equal to zero.  exiting.\n");
	exit(0);
	}

return(curtime);
}
