/* 
 * Copyright (c) Tony Bybell, Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include "cyco.h"

FILE *chandle=NULL, *hhandle=NULL;
char *classname=NULL;

char *empty_string="";
char *bddprefix=NULL;

int nonstick_facs=1;	/* set to zero if you want stkfac */
int wireopt_mode=0; 	/* set this when you want to generate a mostly wirefree ivf file */
int dangle_mode=0;  	/* set when you want a dangle-free model */
int compressed_ids=0;	/* set when FIS section detected */
int io_generate=0; 	/* set to generate i/o access functions */
int bdd_mode=0;		/* set to generate BDD model (-bdd option) */
int mvl_mode=0;		/* set to generate MVL model (-m option) */
int abel_mode=0;        /* set to generate ABEL code (-a option) */
int cyco_mode=0;        /* set to generate branch-less C-code */
int ivf_mode=0;         /* set to generate IVF code (-w2 option) */
int parallel_sim=0;	/* make parallel simulation c code (-p option) */
int opt_tgcc=0;		/* make ternary ops out of and/or (-tgcc option) */
int opt_fastmux=0;	/* dispense with "& (1)" in favor for "< (0)" */
int alias_disable=0;	/* set to disable aliases (-a option) */
int no_model_extract=0; /* do not extract adders etc. from netlist */
int no_tidy=0;          /* do not tidy-up the netlist */
int tidy_before=0;      /* tidy-up before extracting models */
int no_const_prop=0;    /* no constant propagation */
int remove_latches=0;   /* remove unused latches */
int no_strength_reduce=0;/* Do not reduce complex gates to simple gates */
int no_tempsmash=0;	/* Do not reduce simple gates to simple gates */
int no_varreduct=0;	/* Do not allow overlap scheduling */
int cone_mode=0;	/* set to enable cone tracing (-c option) */
int warn_disable=0;     /* do not print warning messages */
int no_facs=0;		/* don't generate facs in C models */
int sim_gran=-1;	/* -1 == infinite block size for C models, else -s sets this to 100 */

static ds_Tree *symtree=NULL;


FILE *handle;
char *tok=NULL;
struct port *ports=NULL;

struct symbol *symchain=NULL;
struct symbol *symnumberone=NULL;

struct gate *gatehead=NULL;
char *buff=NULL;	/* for gate depth indent */

struct gate **sortedgates=NULL;

int numgates=0, num_reverse_latches=0;

int numsyms=0;
int maxdepth=0;
int maxdepth2=0;

int nummode = 0;
static int cstate_sim_which;	/* updates globally because gateprint() is mutually recursive with portprint() */


char *gatenames[]={"NL_GATE", "NL_OPTIMIZED", "NOT", "AND", "OR", "XOR", "NAND", "NOR", "XNOR", "LATCH",
        "LATCH_N", "LATCH_SR", "LATCH_SR_N",
	"NEGC", "ADDC", "SUBC",
	"BUF", "BUFIF0", "PMOS", "BUFIF1", "NMOS", "NOTIF0", "NOTIF1", "MUX",
        "WIRE", "TRI", "WAND", "TRIAND", "WOR", "TRIOR", "TRI0", "TRI1", "SUPPLY0", "SUPPLY1"
};
int maxdelay=0, maxdelaydepth=0;


/*********************************************************************************************/

int is_latch(struct gate *g)
{
   int t = g->type;
   return (t==LATCH) || (t==LATCH_N) || (t==LATCH_SR) || (t==LATCH_SR_N);
}

/*********************************************************************************************/

static int cone_gates=0;
static int cone_count=0;

int cone_2(struct gate *g, int num)
{
struct gateport *gp;
struct symbol *s;
struct sensitivity *se;
int coned=0;

if((!g->cone)&&(!is_latch(g)))
	{
	g->cone=num;
	coned=1;
	cone_gates++;

	gp=g->in;
	while(gp)
		{
		s=gp->port;
		if(s)	/* formerly while() on s, s->next */
			{
			se=s->fromgate;
			while(se)
				{
				cone_2(se->gate, num);
				se=se->next;
				}

			}
		gp=gp->next;					
		}

	gp=g->out;
	while(gp)
		{
		s=gp->port;
		if(s)	/* formerly while() on s, s->next */
			{
			se=s->togate;
			while(se)
				{
				cone_2(se->gate, num);
				se=se->next;
				}
			}
		gp=gp->next;					
		}
	}

return(coned);
}

void cone(void)
{
struct gate *g=gatehead;
cone_count=1;

while(g)
	{
	if(!g->cone)
		{
		if(cone_2(g, cone_count)) cone_count++;
		}
	g=g->next;
	}

cone_count--;
fprintf(stderr,"-- INFO: logic distribution is %d cone%s across %d gate%s\n",
		cone_count, (cone_count!=1)?"s":"", cone_gates, (cone_gates!=1)?"s":"");
}

/*********************************************************************************************/
int queuedepth=0;
struct gatequeue *queueroot=NULL, *queuecurrent=NULL;
struct gatequeue *pass2=NULL;

void queuegate(struct gate *g, char mode)
{
struct gatequeue *q;

if(!mode)
	{
	g->visited=~0;
	}
	else
	{
	g->visited2=~0;
	if(is_latch(g)) 
		{
		num_reverse_latches++;
		return; /* don't add latches */
		}
	}

queuedepth++;
if(!queuecurrent)
	{
	q=(struct gatequeue *)calloc(1,sizeof(struct gatequeue));
	q->gate=g;
	queueroot=queuecurrent=q;
	}
	else
	{
	q=(struct gatequeue *)calloc(1,sizeof(struct gatequeue));
	q->gate=g;
	queuecurrent->next=q;
	queuecurrent=q;
	}
}

void add_in_ports_to_queue(void)
{
struct port *p;

p=ports;
while(p)
	{
	if(!(p->type)) /* in */
		{
		struct symbol *s;
		struct sensitivity *sens;

		s=p->port;
		s->is_valid=~0;
		sens=s->togate;

		while(sens)
			{
			struct gate *g;

			g=sens->gate;
			if(!(g->visited))
				{
				queuegate(g,0);
				}
			sens=sens->next;
			}
		}
	p=p->next;
	}
}

void add_out_ports_to_queue(void)
{
struct port *p;

queueroot=queuecurrent=NULL;
queuedepth=0;

p=ports;
while(p)
	{
	if((p->type)) /* out */
		{
		struct symbol *s;
		struct sensitivity *sens;

		s=p->port;
		s->is_valid2=~0;
		sens=s->fromgate;

		while(sens)
			{
			struct gate *g;

			g=sens->gate;
			if(!(g->visited2))
				{
				queuegate(g,1);
				}
			sens=sens->next;
			}
		}
	p=p->next;
	}
}

void add_dangling_nets_to_queue(void)
{
struct symbol *s;

s=symnumberone;

while(s)
	{
	struct sensitivity *sens;

	if((s->togate==NULL)&&(!s->isport)&&(!s->is_valid2))
		{
                if (!warn_disable && !(s->flags && SYMF_IS_ALIAS))
		fprintf(stderr,"WARNING: Dangling output net '%s' found.\n",s->name);

		s->is_valid2=~0;
		sens=s->fromgate;

		while(sens)
			{
			struct gate *g;

			g=sens->gate;
			if(!(g->visited2))
				{
				queuegate(g,1);
				}
			sens=sens->next;
			}
		}
	s=s->prev;
	}
}

void add_const_nets_to_queue(void)
{
struct symbol *s;

s=symnumberone;

while(s)
	{
	struct sensitivity *sens;
	int isdollar;

	if(((isdollar=(*(s->name)=='$'))||(s->fromgate==NULL))&&(!s->zapped))
		{
		if((!isdollar)&&(!s->isport))
                if (!warn_disable && !(s->flags && SYMF_IS_ALIAS))
			{
			fprintf(stderr,"WARNING: Floating net '%s'",s->name);
			if(!mvl_mode&&!abel_mode) fprintf(stderr," defaults to zero.\n");
			else fprintf(stderr,".\n");
			}

		s->is_valid=~0;
		sens=s->togate;

		while(sens)
			{
			struct gate *g;

			g=sens->gate;
			if(!(g->visited))
				{
				queuegate(g,0);
				}
			sens=sens->next;
			}
		}
	s=s->prev;
	}
}

void add_child_gates_to_queue(struct gate *l)
{
struct gateport *gp;

gp=l->out;
while(gp)
	{
	struct symbol *s;
	struct sensitivity *sens;

	s=gp->port;
	s->is_valid=~0;
	sens=s->togate;
	while(sens)
		{
		struct gate *g;

		g=sens->gate;
		if(!(g->visited))
			{
			queuegate(g,0);
			}
		sens=sens->next;
		}
	gp=gp->next;
	}		
}

void add_parent_gates_to_queue(struct gate *l)
{
struct gateport *gp;

gp=l->in;
while(gp)
	{
	struct symbol *s;
	struct sensitivity *sens;

	s=gp->port;
	s->is_valid2=~0;
	sens=s->fromgate;
	while(sens)
		{
		struct gate *g;

		g=sens->gate;
		if(!(g->visited2))
			{
			queuegate(g,1);
			}
		sens=sens->next;
		}
	gp=gp->next;
	}		
}

void add_latches_to_queue(void)
{
struct gate *l;

l=gatehead;
while(l)
	{
	if(is_latch(l))
		{
		add_child_gates_to_queue(l);
		l->depth=numgates;
		}
	l=l->next;
	}
}

void breadth_first_search(void)
{
int depth=1;
int watchdog=0;
struct gatequeue *q;

q=queueroot;
while(q)
	{
	struct gate *g;
	struct gateport *gp;
	char failed;

	failed=0;
	g=q->gate;
	gp=g->in;
	while(gp)
		{
		struct symbol *s;
		s=gp->port;
		if(!(s->is_valid))
			{
			failed=~0;
			break;
			}
		gp=gp->next;
		}
	if(!failed)
		{
		struct gatequeue *qt;
		g->depth=depth++;
		queuedepth--;	/* remove from count */
		watchdog=0;
		add_child_gates_to_queue(g);		
		qt=q->next;
		free(q);
		q=qt;
		}
		else
		{
		struct gatequeue *qt;
		qt=q->next;
		queuecurrent->next=q;
		queuecurrent=q;
		q->next=NULL;
		q=qt;
		watchdog++;
		}

	if((queuedepth)&&(watchdog==queuedepth))
		{
		struct gatequeue *qt;
		int i=1;
		fprintf(stderr,"Possible loops found...\n");
		fprintf(stderr,"Watchdog awakened at queuedepth: %d\n",queuedepth);

		qt=q;
		while(qt)
			{
			struct gate *g;
			g=qt->gate;
			fprintf(stderr,"| %d) %s.%d #%d\n",i++,gatenames[g->type],g->number,g->delay);
			qt=qt->next;
			}

		exit(0);
		}
	}

maxdepth=depth;
}

void breadth_first_search_rvs(void)
{
int depth=1;
int watchdog=0;
struct gatequeue *q;

q=queueroot;
while(q)
	{
	struct gate *g;
	struct gateport *gp;
	char failed;

	failed=0;
	g=q->gate;
	gp=g->out;
	while(gp)
		{
		struct symbol *s;
		s=gp->port;
		if(!(s->is_valid2))
			{
			failed=~0;
			break;
			}
		gp=gp->next;
		}
	if(!failed)
		{
		struct gatequeue *qt;
		g->depth2=depth++;
		queuedepth--;	/* remove from count */
		watchdog=0;
		add_parent_gates_to_queue(g);		
		qt=q->next;

		q->next=pass2;	/* stack in rvs order */
		pass2=q;

		q=qt;
		}
		else
		{
		struct gatequeue *qt;
		qt=q->next;
		queuecurrent->next=q;
		queuecurrent=q;
		q->next=NULL;
		q=qt;
		watchdog++;
		}

	if((queuedepth)&&(watchdog==queuedepth))
		{
		struct gatequeue *qt;
		int i=1;
		fprintf(stderr,"Possible loops found in reverse pass...\n");
		fprintf(stderr,"Watchdog awakened at queuedepth: %d\n",queuedepth);

		qt=q;
		while(qt)
			{
			struct gate *g;
			g=qt->gate;
			fprintf(stderr,"| %d) %s.%d #%d\n",i++,gatenames[g->type],g->number,g->delay);
			qt=qt->next;
			}

		exit(0);
		}
	}

maxdepth2=depth;
}

/*********************************************************************************************/


/*
 * handle debugging stuff...
 */
#ifndef DEBUG_FEEDBACK
void DEBUG(char *dummy, ...) { }
#endif

/*
 * write model with wires optimized out to the screen..
 */
char *backtrace(struct symbol *from)
{
while(from->back) { from=from->back; }
return(from->name);
}

void wireopt(char *model)
{
struct gate *g;
struct symbol *src, *dest, *sym, *backtrack;
struct port *p, *ptemp, *porthead=NULL;
static char *dir[]={"IN","OUT"};
char *tempswap, *tempcompare;
struct gateport *gpin, *gpout;
struct alias *alias, *alias2;

g=gatehead;
while(g)
	{
	if(g->type==NL_GATE)
		{
		src=g->in->port;
		dest=g->out->port;

		if(!dest->isport)
			{
			dest->back=src;
			dest->zapped=~0;
			g->type=NL_OPTIMIZED;

			tempcompare=(src->id)?src->id:src->name;
			if(strchr(tempcompare,'!'))	
				{
				/* hashing *will* be broken after this, but we don't need it again */
				tempswap=src->name;
				src->name=dest->name;
				dest->name=tempswap;

				tempswap=src->id;
				src->id=dest->id;
				dest->id=tempswap;
				}

			if(!alias_disable)
				{
				backtrack=src;
				while(backtrack->back) backtrack=backtrack->back;

				alias=calloc(1,sizeof(struct alias));	/* link in alias if enabled */
				alias->next=backtrack->alias;
				alias->sym=dest;
				backtrack->alias=alias;

				if(dest->alias)				/* link in any forgotten chains */
					{
					alias2=dest->alias;
					while(alias2->next) alias2=alias2->next;
					alias2->next=backtrack->alias;
					backtrack->alias=dest->alias;
					dest->alias=NULL;
					}
				}
			}
		}
	g=g->next;
	}

printf("MX %s{\nM%sP{",model,(compressed_ids)?"I":"");
p=ports;
while(p)	/* invert ports to normal order */
	{
	ptemp=(struct port *)malloc(sizeof(struct port));
	ptemp->port=p->port;
	ptemp->type=p->type;
	ptemp->next=porthead;
	porthead=ptemp;
	p=p->next;
	}

while(porthead)
	{
	struct symbol *s;
	s=symfind(porthead->port->name, 1);

	if((!s)||(!s->id))
		{
		printf("%s %s;",porthead->port->name, dir[porthead->type]);
		}
		else
		{
		printf("%s %s %s;",s->name, s->id, dir[porthead->type]);
		}
	ptemp=porthead->next;
	free(porthead);
	porthead=ptemp;
	}
printf("}\n}\n\nMF{\n");

sym=symnumberone;

if(!compressed_ids)
while(sym)
	{
	if((!(sym->isport))&&(!(sym->zapped)))
		{
		if(!sym->alias)
			{
			printf("FS{%s }\n",sym->name);
			}
			else
			{
			printf("FSA{%s ",sym->name);
			alias=sym->alias;
			while(alias)
				{
				printf("%s ",alias->sym->name);
				alias=alias->next;
				}
			printf("}\n");
			}
		}
	sym=sym->prev;
	}
else
while(sym)
	{
	if((!(sym->isport))&&(!(sym->zapped)))
		{
		if(sym->id)	/* $0, $1, ... won't have ids */
			{
			if(!sym->alias)
				{
				printf("FIS{%s %s }\n",sym->name, sym->id);
				}
				else
				{
				printf("FISA{%s %s ",sym->name, sym->id);
				alias=sym->alias;
				while(alias)
					{
					printf("%s ",alias->sym->id?alias->sym->id:alias->sym->name);
					alias=alias->next;
					}
				printf("}\n");
				}

			}
			else
			{
			if(!sym->alias)
				{
				printf("FS{%s }\n",sym->name);
				}
				else
				{
				printf("FSA{%s ",sym->name);
				alias=sym->alias;
				while(alias)
					{
					printf("%s ",alias->sym->id?alias->sym->id:alias->sym->name);
					alias=alias->next;
					}
				printf("}\n");
				}
			}
		}
	sym=sym->prev;
	}


g=gatehead;
while(g)
	{
	if(g->type!=NL_OPTIMIZED)
		{
		printf("FG %s.%d{%s %d I %d ",gatenames[g->type],g->number,
			gatenames[g->type],g->delay,g->incnt);
		gpin=g->in;
		while(gpin)
			{
			printf("%s ",backtrace(gpin->port));
			gpin=gpin->next;
			}
		printf("%d ",g->outcnt);
		gpout=g->out;
		while(gpout)
			{
			printf("%s ",backtrace(gpout->port));
			gpout=gpout->next;
			}
		printf("}\n");
		}
	g=g->next;
	}
printf("}\n");
}

/*
 * Quicksort algorithm from p154 of 
 * "Introduction to Algorithms" by Cormen, Leiserson, and Rivest.
 * Comparison signs flipped to do a reverse sort..
 */
int partition(struct gate **a, int p, int r)
{
struct gate *x, *t;
int i,j;

x=a[p];
i=p-1;
j=r+1;

while(1)
	{
	do
		{
		j--;
		} while(a[j]->depth>x->depth);	

	do	{
		i++;
		} while(a[i]->depth<x->depth);

	if(i<j)
		{
		t=a[i];
		a[i]=a[j];
		a[j]=t;
		}
		else
		{
		return(j);
		}
	}
}

void quicksort(struct gate **a, int p, int r)
{
int q;

if(p<r)
	{
	q=partition(a,p,r);
	quicksort(a,p,q);
	quicksort(a,q+1,r);
	}
}



/*
 * forward depthsort gates for c code emission... 
 */
void sortdepth(void)
{
int i;
struct gate *g;

sortedgates=(struct gate **)malloc(numgates*sizeof(struct gate *));

i=0;
g=gatehead;

while(g)
	{
	sortedgates[i++]=g;
	g=g->next;
	}

quicksort(sortedgates,0,numgates-1);
}

/* ######################################################################################## */

#ifdef ALLOW_SORTSTATS
/*
 * Print stats on sorted gate intereconnects...
 */
void sortstats(void)
{
struct gate *g;
struct gateport *gp;
struct sensitivity *sens;
int i,o,t;

for(t=0;t<numgates;t++)
	{
	g=sortedgates[t];

	fprintf(stderr,"%s.%d.%d #%d\n",gatenames[g->type],g->number,g->cone,g->delay);
	fprintf(stderr,"\t[D] %d\n",g->depth);

	gp=g->in;
	for(i=0;i<g->incnt;i++)
		{
		fprintf(stderr,"\t[I] %-20s from ",gp->port->name);

		sens=gp->port->fromgate;
		if(!sens)
			{
			if(gp->port->isport)
				{
				if(gp->port->isport->type) fprintf(stderr,"OUT PORT");
				else fprintf(stderr,"IN PORT");
				}
				else
				{
				fprintf(stderr,">FLOATING<");
				}
			}
		else
		while(sens)
			{
			fprintf(stderr,"(%s.%d)",gatenames[sens->gate->type],
				sens->gate->number);
			sens=sens->next;
			}
		fprintf(stderr,"\n");

		gp=gp->next;
		}	
	gp=g->out;
	for(o=0;o<g->outcnt;o++)
		{
		fprintf(stderr,"\t[O] %-20s to ",gp->port->name);

		sens=gp->port->togate;
		if(!sens)
			{
			if(gp->port->isport)
				{
				if(gp->port->isport->type) fprintf(stderr,"OUT PORT");
				else fprintf(stderr,"IN PORT");
				}
				else
				{
				fprintf(stderr,">FLOATING<");
				}
			}
		else
		while(sens)
			{
			fprintf(stderr,"(%s.%d)",gatenames[sens->gate->type],
				sens->gate->number);
			sens=sens->next;
			}
		fprintf(stderr,"\n");
		gp=gp->next;
		}	
	}
}
#endif

/* ######################################################################################## */

/*
 * print out the ports for a C model gate (used in gateprint())...
 */
void gateprint(struct gate *g);

void portprint(struct gateport *gp, int invertz)
{
if(*(gp->port->name)!='$')
	{
	if(gp->expr)
		{
		fprintf(classname ? chandle: stdout, " (");
		gateprint(gp->expr);	/* recurse into subexpression */
		fprintf(classname ? chandle: stdout, ")");
		return;
		}

	if(nonstick_facs)
		{
		fprintf(classname ? chandle: stdout, " t[%d]",gp->port->number);
		}
		else
		{
		fprintf(classname ? chandle: stdout, " ((s[%d]&(z=v[%d]))|(~z&t[%d]))", gp->port->number, gp->port->number, gp->port->number);
		}
	}
else
	{
	if(!invertz)
		{
		if((*(gp->port->name+1))=='1')
			{
			if((parallel_sim)||(opt_fastmux))
				fprintf(classname ? chandle: stdout, " (~0)");
				else
				fprintf(classname ? chandle: stdout, " (1)");
			}
			else
			{
			fprintf(classname ? chandle: stdout, " 0");
			}
		}
	else
		{
		if((*(gp->port->name+1))=='0')
			{
			fprintf(classname ? chandle: stdout, " 0");
			}
			else
			{
			if((parallel_sim)||(opt_fastmux))
				fprintf(classname ? chandle: stdout, " (~0)");
				else
				fprintf(classname ? chandle: stdout, " (1)");
			}
		}
	}
}


/*
 * Print c statements...
 */
void gateprint(struct gate *g)
{
struct gateport *gp, *gpcache=NULL;
int i,o;
int invertz=0;
int maxin;
int negcontrol=0;
char oper;
int recursive = (g->flags & GF_EXPRESSION);
const char *terminal = recursive ? "" : ";\n";

cstate_sim_which++;

maxin=g->incnt;
	{
	if(!recursive)
		{
		gp=g->out;
		for(o=0;o<g->outcnt;o++)
			{
			if(is_latch(g))
				{
				fprintf(classname ? chandle: stdout, "l%d =",gp->port->number);
				}
				else
				{
				fprintf(classname ? chandle: stdout, "t[%d] =",gp->port->number);
				}
			gp=gp->next;
			}	
		}

	switch(g->type)
		{
		case ADDC:
			{
			struct gateport *gpa, *gpb, *gpc;
			gpa=g->in;
			gpb=gpa->next;
			gpc=gpb->next;
			
			fprintf(classname ? chandle: stdout, " (");
			fprintf(classname ? chandle: stdout, " ("); portprint(gpa,0);	fprintf(classname ? chandle: stdout, " &"); portprint(gpb,0); fprintf(classname ? chandle: stdout, " ) |");
			fprintf(classname ? chandle: stdout, " ("); portprint(gpa,0);	fprintf(classname ? chandle: stdout, " &"); portprint(gpc,0); fprintf(classname ? chandle: stdout, " ) |");
			fprintf(classname ? chandle: stdout, " ("); portprint(gpb,0);	fprintf(classname ? chandle: stdout, " &"); portprint(gpc,0); fprintf(classname ? chandle: stdout, " )");
			fprintf(classname ? chandle: stdout, " ) %s", terminal);

			return;
			}

		case SUBC:
			if(maxin==3)
			{
			struct gateport *gpa, *gpb, *gpc;
			gpa=g->in;
			gpb=gpa->next;
			gpc=gpb->next;
			
			fprintf(classname ? chandle: stdout, " (");
			fprintf(classname ? chandle: stdout, " ("); portprint(gpa,0);	fprintf(classname ? chandle: stdout, " & ~"); portprint(gpb,0); fprintf(classname ? chandle: stdout, " ) |");
			fprintf(classname ? chandle: stdout, " ("); portprint(gpa,0);	fprintf(classname ? chandle: stdout, " &"); portprint(gpc,0); fprintf(classname ? chandle: stdout, " ) |");
			fprintf(classname ? chandle: stdout, " ( ~"); portprint(gpb,0);fprintf(classname ? chandle: stdout, " &"); portprint(gpc,0); fprintf(classname ? chandle: stdout, " )");
			fprintf(classname ? chandle: stdout, " ) %s", terminal);

			return;
			}
			else
			{
			struct gateport *gpa, *gpb;
			gpa=g->in;
			gpb=gpa->next;
			
			fprintf(classname ? chandle: stdout, " ("); portprint(gpa,0);	fprintf(classname ? chandle: stdout, " | ~"); portprint(gpb,0); 
			fprintf(classname ? chandle: stdout, " ) %s", terminal);

			return;
			}

		case BUFIF0:
			{
			struct gateport *gpa, *gpb;
			gpa=g->in;
			gpb=gpa->next;
			
			fprintf(classname ? chandle: stdout, " ("); portprint(gpa,0);	fprintf(classname ? chandle: stdout, " & ~"); portprint(gpb,0); 
			fprintf(classname ? chandle: stdout, " ) %s", terminal);

			return;
			}

		case LATCH_SR:
		case LATCH_SR_N:
			{
			struct gateport *gpa, *gpb, *gpc;
			gpa=g->in;
			gpb=gpa->next;
			gpc=gpb->next;
			
			fprintf(classname ? chandle: stdout, " (");
			fprintf(classname ? chandle: stdout, " ("); portprint(gpa,0);	fprintf(classname ? chandle: stdout, " ) | (");
			portprint(gpb,0);
			fprintf(classname ? chandle: stdout, " ) ) & ( ~(");
			portprint(gpc,0); fprintf(classname ? chandle: stdout, " ) ) %s", terminal);

			return;
			}

		case NEGC:
		case NOTIF1:
			oper='&';
			fprintf(classname ? chandle: stdout, " ( ~ ");
			break;
		case NL_GATE:
		case TRI0:
		case SUPPLY0:
		case BUF:
			oper=' ';
			fprintf(classname ? chandle: stdout, " (");
			break;
		case NOT:
		case TRI1:
		case SUPPLY1:
			oper=' ';
			fprintf(classname ? chandle: stdout, " ~ (");			
			break;
		case AND:
			if((opt_tgcc)&&(g->incnt>=opt_tgcc))
				{
				struct gateport *gp = g->in;
				int portsanity = 0;
				fprintf(classname ? chandle: stdout, " (");
				do
					{
					fprintf(classname ? chandle: stdout, " (");
					portprint(gp, 0);
					fprintf(classname ? chandle: stdout, " )");
					gp=gp->next;
					if(gp) 
						{
						portsanity++;
						if(opt_fastmux)
							{
							fprintf(classname ? chandle: stdout, " < (0) ?");
							}
							else
							{
							fprintf(classname ? chandle: stdout, " & (1) ?");
							}
						}
					} while(gp);			

				for(i=0;i<portsanity;i++)
					{
					fprintf(classname ? chandle: stdout, " : (0)");
					}
								
				fprintf(classname ? chandle: stdout, " ) %s", terminal);
				return;
				}
				/* else fallthrough to normal '&' type */
		case BUFIF1:
			oper='&';
			fprintf(classname ? chandle: stdout, " (");
			break;
		case NMOS:
			invertz=1;
			oper='&';
			fprintf(classname ? chandle: stdout, " (");
			break;
		case WAND:
		case TRIAND:
			invertz=~0;
			oper='&';
			fprintf(classname ? chandle: stdout, " (");
			break;
		case OR:
			if((opt_tgcc)&&(g->incnt>=opt_tgcc))
				{
				struct gateport *gp = g->in;
				fprintf(classname ? chandle: stdout, " (");
				do
					{
					fprintf(classname ? chandle: stdout, " (");
					portprint(gp, 0);
					fprintf(classname ? chandle: stdout, " )");
					gp=gp->next;
					if(gp) 
						{
						if(opt_fastmux)
							{
							fprintf(classname ? chandle: stdout, " < (0) ?(~0):");
							}
							else
							{
							fprintf(classname ? chandle: stdout, " & (1) ?:");
							}
						}
					} while(gp);			
								
				fprintf(classname ? chandle: stdout, " ) %s", terminal);
				return;
				}
				/* else fallthrough to normal '|' type */
		case WOR:
		case TRIOR:
		case WIRE:
		case TRI:
			oper='|';
			fprintf(classname ? chandle: stdout, " (");
			break;
		case XOR:
			oper='^';
			fprintf(classname ? chandle: stdout, " (");
			break;
		case NAND:
			oper='&';
			fprintf(classname ? chandle: stdout, " ~ (");
			break;

		case NOR:
		case NOTIF0:
			oper='|';
			fprintf(classname ? chandle: stdout, " ~ (");
			break;

		case XNOR:
			oper='^';
			fprintf(classname ? chandle: stdout, " ~ (");
			break;
		case LATCH:
		case LATCH_N:
			oper=' ';
			fprintf(classname ? chandle: stdout, " (");
			break;

		case PMOS:
			maxin--;
			negcontrol=~0;
			oper='&';
			fprintf(classname ? chandle: stdout, " (");
			break;

		case MUX:
			oper='?';
			fprintf(classname ? chandle: stdout, " (");
			break;

		default:
			oper=' ';	/* more gates to support here */
			break;
		}

	gp=g->in;
	for(i=0;i<maxin;i++)
		{
		if(parallel_sim)
		if(g->type==MUX)
			{
			if(i==0)
				{
				gpcache=gp;
				oper='&';
				}
			else
			if(i==1)
				{
				oper='|';
				}
			}


		muxfake:
		if(negcontrol)
			{
			fprintf(classname ? chandle: stdout, " (~");
			}

		portprint(gp, invertz);

		if(negcontrol)
			{
			fprintf(classname ? chandle: stdout, " )");
			negcontrol=0;
			}

		gp=gp->next;
		if((i!=(maxin-1)) && !is_latch(g))
			/* patch over clock pin for now */
			{
			if(oper=='?')
				{
				if(opt_fastmux)
					{
					fprintf(classname ? chandle: stdout, " < (0) %c",oper); /* not parallel sim */
					}
					else
					{
					fprintf(classname ? chandle: stdout, " & (1) %c",oper);	/* not parallel sim */
					}
				oper=':';
				}
				else
				{
				fprintf(classname ? chandle: stdout, " %c",oper);
				}
			}
			else
			{
			if((parallel_sim)&&(gpcache))
				{
				gp=gpcache;
				gpcache=NULL;
				fprintf(classname ? chandle: stdout, " & ~"); 
				goto muxfake;
				}

			fprintf(classname ? chandle: stdout, " ) %s", terminal);
			if(is_latch(g)) break;
			}
		}	
	}
}

void cstate_facname(char *name)
{
int i, len;
unsigned char ch;

if(name)
	{
	if((len=strlen(name)))	
		{
		fprintf(classname ? chandle: stdout, "\"");
		for(i=0;i<len;i++)
			{
			ch=(unsigned char)*(name++);
			if((ch<32)||(ch>127)||(ch=='\"')||(ch=='\\'))
				{
				fprintf(classname ? chandle: stdout, "\\%-03o",ch);
				}
				else
				{
				fprintf(classname ? chandle: stdout, "%c",ch);
				}
			}
		fprintf(classname ? chandle: stdout, "\"");
		}
	}
}

void cstate(void)
{
struct gate *g;
struct gateport *gp;
int i,o,t,prcone,cone_min;
struct symbol *ss;
int sim_comb_max=0, sim_latch_max=0, sim_comb_postlatch_max=0;

if(classname)
	{
	fprintf(hhandle, 
			"#ifndef %s_CLASSFILE_HH\n"
			"#define %s_CLASSFILE_HH\n\n"
			"class %s\n"
			"{\npublic:\n\n"
			"%s();\n"
			"~%s();\n\n",
			classname, classname, classname, classname, classname);

	fprintf(chandle, "#include \"%s.H\"\n\n", classname);
	}

if((!no_facs)&&(no_varreduct))
	{
	if(classname)
		{
		fprintf(hhandle, "char **sim_facs;\n");
		fprintf(chandle, "static char *local_sim_facs[]=\n\t{\n");
		fprintf(chandle, "\t/* %-4d */  \">> FACILITIES <<\",\n",0);
		}
		else
		{
		printf("char *sim_facs[]=\n\t{\n");
		printf("\t/* %-4d */  \">> FACILITIES <<\",\n",0);
		}
	ss=symnumberone;
	while(ss)
	{
	if(!ss->zapped)
	if(classname)
		{
		if(!ss->id)
			{
			fprintf(chandle, "\t/* %-4d */  ",ss->number);
			cstate_facname(ss->name); fprintf(chandle, ",\n");
			}
			else
			{
			fprintf(chandle, "\t/* %-4d */  ",ss->number);
			cstate_facname(ss->id); fprintf(chandle, ",\n");
			}
		}
		else
		{
		if(!ss->id)
			{
			printf("\t/* %-4d */  ",ss->number);
			cstate_facname(ss->name); printf(",\n");
			}
			else
			{
			printf("\t/* %-4d */  ",ss->number);
			cstate_facname(ss->id); printf(",\n");
			}
		}

	ss=ss->prev;
	}

	if(classname)
		{
		fprintf(chandle, "\t/* %-4d */  %d,\n",-1,0);
		fprintf(chandle, "\t};\n\n");
		}
		else
		{
		printf("\t/* %-4d */  %d,\n",-1,0);
		printf("\t};\n\n");
		}
	}

if(classname)
	{
	fprintf(hhandle, "%ssigned char *sim_tvalues, *sim_svalues, *sim_vvalues;\n", opt_fastmux ? "" : "un");
	fprintf(chandle, "static %ssigned char t[%d]", opt_fastmux ? "" : "un", numsyms+1);
	if(!nonstick_facs)
		{
		fprintf(chandle, ", s[%d], v[%d]", numsyms+1, numsyms+1);
		}
	}
	else
	{
	printf("%ssigned char *sim_tvalues=(%ssigned char *)0, *sim_svalues=(%ssigned char *)0, *sim_vvalues=(%ssigned char *)0;\n", 
				opt_fastmux ? "" : "un",
				opt_fastmux ? "" : "un",
				opt_fastmux ? "" : "un",
				opt_fastmux ? "" : "un");
	printf("static %ssigned char t[%d]", opt_fastmux ? "" : "un", numsyms+1);
	if(!nonstick_facs)
		{
		printf(", s[%d], v[%d]", numsyms+1, numsyms+1);
		}
	}

i=0;
for(t=0;;t++)
	{
	lattop:
	g=sortedgates[i];
	if(!is_latch(g))
		{
		i++;
		if(i<numgates) goto lattop;
		else break;
		}

	if (!(t&7)) { fprintf(classname ? chandle: stdout, " ;\nstatic %ssigned char l%d", opt_fastmux ? "" : "un", g->out->port->number); }
	else { fprintf(classname ? chandle: stdout, ", l%d",g->out->port->number); }

	i++;
	if(i==numgates) break;
	}

fprintf(classname ? chandle: stdout, " ;\n\n");
if(classname)
	{
	fprintf(hhandle, "unsigned int Num_Syms;\n\n",numsyms);
	}
	else
	{
	printf("const unsigned int Num_Syms = %d ;\n\n",numsyms);
	}

if(classname)
	{
	fprintf(chandle, "void %s::sim_comb_0()\n{\n", classname);
	fprintf(hhandle, "void sim_comb_0();\n");
	if(!nonstick_facs)
		{
		fprintf(chandle, "%ssigned char z;\n", opt_fastmux ? "" : "un");
		}
	}
	else
	{
	printf("void sim_comb_0(void)\n{\n");
	if(!nonstick_facs)
		{
		printf("%ssigned char z;\n", opt_fastmux ? "" : "un");
		}
	}

cstate_sim_which=0;
for(i=cone_count;i>-1;i--)
	{
	fprintf(classname ? chandle: stdout, "/* Cone: %d */\n",i);
	for(t=0;t<numgates;t++)
		{
		struct gate *g;
		g=sortedgates[t];
	
		if(g->cone==i)
		if(g->depth)
			{
			if ((g->flags & GF_EXPRESSION)==0)
				{
				gateprint(g);
				if((sim_gran>0)&&(cstate_sim_which>=sim_gran)&&(t!=(numgates-1)))
					{
					cstate_sim_which=0;
					if(classname)
						{
						fprintf(chandle, "}\n\nvoid %s::sim_comb_%d()\n{\n",classname, ++sim_comb_max);
						fprintf(hhandle, "void sim_comb_%d();\n",sim_comb_max);
						}
						else
						{
						printf("}\n\nvoid sim_comb_%d(void)\n{\n",++sim_comb_max);
						}
					if(!nonstick_facs)
						{
						fprintf(classname ? chandle: stdout, "%ssigned char z;\n", opt_fastmux ? "" : "un");
						}
					}
				}
			}
			else
			{
			/* should never happen since we zeroed dangling nets */
                        if (!warn_disable)
			fprintf(stderr,"WARNING: %s.%d #%d is unreachable.\n",gatenames[g->type],g->number,g->delay);
			}
		}
	}

if(classname)
	{
	fprintf(chandle, "}\n\nvoid %s::sim_comb_postlatch_0()\n{\n", classname); /* postlatch */
	fprintf(hhandle, "void sim_comb_postlatch_0();\n");
	}
	else
	{
	printf("}\n\nvoid sim_comb_postlatch_0(void)\n{\n"); /* postlatch */
	}
if(!nonstick_facs)
	{
	fprintf(classname ? chandle: stdout, "%ssigned char z;\n", opt_fastmux ? "" : "un");
	}

if(cone_mode) cone_min=0; else cone_min=-1;

cstate_sim_which=0;
if(num_reverse_latches)
for(i=cone_count;i>cone_min;i--)
	{
	prcone=0;
	for(t=0;t<numgates;t++)
		{
		struct gate *g;
		g=sortedgates[t];

		if(g->cone==i)
		if(g->depth2)
			{
			if ((g->flags & GF_EXPRESSION)==0)
				{
				if(!prcone) { prcone=1; fprintf(classname ? chandle: stdout, "/* Cone: %d */\n",i);; }
				gateprint(g);
				if((sim_gran>0)&&(cstate_sim_which>=sim_gran)&&(t!=(numgates-1)))
					{
					cstate_sim_which=0;
					if(classname)
						{
						fprintf(chandle, "}\n\nvoid %s::sim_comb_postlatch_%d()\n{\n",classname, ++sim_comb_postlatch_max);
						fprintf(hhandle, "void sim_comb_postlatch_%d();\n",sim_comb_postlatch_max);
						}
						else
						{
						printf("}\n\nvoid sim_comb_postlatch_%d(void)\n{\n",++sim_comb_postlatch_max);
						}
					if(!nonstick_facs)
						{
						fprintf(classname ? chandle: stdout, "%ssigned char z;\n", opt_fastmux ? "" : "un");
						}
					}
				}
			}
		}
	}

if(classname)
	{
	fprintf(chandle, "}\n\nvoid %s::sim_latch_0()\n{\n", classname); /* latch assignment */
	fprintf(hhandle, "void sim_latch_0();\n"); /* latch assignment */
	}
	else
	{
	printf("}\n\nvoid sim_latch_0(void)\n{\n"); /* latch assignment */
	}

if(!nonstick_facs)
	{
	fprintf(classname ? chandle: stdout, "%ssigned char z;\n", opt_fastmux ? "" : "un");
	}

cstate_sim_which=0;
for(t=0;t<numgates;t++)
	{
	g=sortedgates[t];

	gp=g->out;
	for(o=0;o<g->outcnt;o++)
		{
		if(is_latch(g))
			{

			if(nonstick_facs)
				{
				fprintf(classname ? chandle: stdout, "t[%d] = ( l%d ) ;\n",gp->port->number, gp->port->number);
				}
				else
				{
				fprintf(classname ? chandle: stdout, "t[%d] = ( (s[%d]&(z=v[%d]))|(~z&l%d) ) ;\n", gp->port->number, gp->port->number, gp->port->number, gp->port->number);
				}

			cstate_sim_which++;
			if((sim_gran>0)&&(cstate_sim_which>=sim_gran)&&(t!=(numgates-1)))
				{
				cstate_sim_which=0;
				if(classname)
					{
					fprintf(chandle, "}\n\nvoid %s::sim_latch_%d()\n{\n",classname, ++sim_latch_max);
					fprintf(hhandle, "void sim_latch_%d();\n", sim_latch_max);
					}
					else
					{
					printf("}\n\nvoid sim_latch_%d(void)\n{\n",++sim_latch_max);
					}
				if(!nonstick_facs)
					{
					fprintf(classname ? chandle: stdout, "%ssigned char z;\n", opt_fastmux ? "" : "un");
					}
				}
			}
		gp=gp->next;
		}	
	}

if(classname)
	{
	fprintf(chandle, "}\n\nvoid %s::sim()\n{\n", classname);
	fprintf(hhandle, "void sim();\n");
	for(t=0;t<=sim_comb_max;t++)
		fprintf(chandle, "sim_comb_%d();\n",t);
	for(t=0;t<=sim_latch_max;t++)
		fprintf(chandle, "sim_latch_%d();\n",t);
	for(t=0;t<=sim_comb_postlatch_max;t++)
		fprintf(chandle, "sim_comb_postlatch_%d();\n",t);
	fprintf(chandle, "}\n");

	fprintf(chandle, "\n\n%s::~%s() { /* nothing */ }\n", classname, classname);
	fprintf(chandle, "%s::%s()\n{\n", classname, classname);
	fprintf(chandle, "int i;\n"
		"Num_Syms = %d;\n"
		"for(i=1;i<%d;i++)\n\t{\n"
		"\tt[i]=", numsyms, numsyms+1);
	if(!nonstick_facs)
		{
		fprintf(chandle, "s[i]=v[i]=");
		}
	
	fprintf(chandle, "0;\n\t}\nsim_tvalues=t;\n");

	if(!nonstick_facs)
		{
		fprintf(chandle, "sim_svalues=s;\nsim_vvalues=v;\n");
		}
	fprintf(chandle, "}\n");
	}
	else
	{
	printf("}\n\nvoid sim(void)\n{\n");
	for(t=0;t<=sim_comb_max;t++)
		printf("sim_comb_%d();\n",t);
	for(t=0;t<=sim_latch_max;t++)
		printf("sim_latch_%d();\n",t);
	for(t=0;t<=sim_comb_postlatch_max;t++)
		printf("sim_comb_postlatch_%d();\n",t);
	printf("}\n");
	
	printf("\n\nvoid sim_init(void)\n{\n");
	printf("int i;\n"
		"for(i=1;i<%d;i++)\n\t{\n"
		"\tt[i]=", numsyms+1);
	if(!nonstick_facs)
		{
		printf("s[i]=v[i]=");
		}
	
	printf("0;\n\t}\nsim_tvalues=t;\n");

	if(!nonstick_facs)
		{
		printf("sim_svalues=s;\nsim_vvalues=v;\n");
		}
	printf("}\n");
	}

if(io_generate)
for (ss = symnumberone; ; ss = ss->prev)
{
   struct port *port;
   char *p1, *p2;
   int len;
   static int curr_len = 0, curr_type, curr_sz, curr_num;
   char curr_name[200];
   if (ss != 0)
      port = ss->isport;
   if (ss == symnumberone)
      curr_name[0] = 0;              /* initialize */
   if (ss && !port)
      continue;
   if (ss)
   {
   if(!ss->id)
	{
      	p1 = strchr(ss->name,'/');
      	p2 = strrchr(ss->name,'/');
	}
	else
	{
	p1 = strchr(ss->id,'/');
	p2 = strrchr(ss->id,'/');
	}
   }
   len = p2 - p1 - 1;
   if (ss && (curr_len == len) && (strncmp(curr_name, p1 + 1, len) == 0))
   {                            /* we found next index in current port */
      curr_sz += 1;
      curr_num = ss->number;
   }
   else
   {                            /* we found a new port */
      if (curr_len != 0)
      {
         int j;
         int num = curr_num;
	if(classname)
		{
		fprintf(hhandle, "char *get_le_%s();\n", curr_name);
		fprintf(chandle, "\n\nchar *%s::get_le_%s()\n{\n"
                	"  static char s[%d];\n  ", classname, curr_name, curr_sz+1);
		}
		else
		{
         	printf("\n\nchar *get_le_%s(void)\n{\n"
                	"  static char s[%d];\n  ", curr_name, curr_sz+1);
		}
         for (j = 0; j < curr_sz; j += 1)
		{
		if(opt_fastmux)
			{
		        fprintf(classname ? chandle: stdout, "s[%d]='0'+((t[%d]>>7)&1);",j, num-j);
			}
			else
			{
		        fprintf(classname ? chandle: stdout, "s[%d]='0'+(t[%d]&1);",j, num-j);
			}
		}
         fprintf(classname ? chandle: stdout, "s[%d]=0;\n  return s;\n}",curr_sz);

	if(classname)
		{
         	fprintf(hhandle, "char *get_be_%s();\n", curr_name);
         	fprintf(chandle, "\n\nchar *%s::get_be_%s(void)\n{\n"
                	"  static char s[%d];\n  ", classname, curr_name, curr_sz+1);
		}
		else
		{
         	printf("\n\nchar *get_be_%s(void)\n{\n"
                	"  static char s[%d];\n  ", curr_name, curr_sz+1);
		}

         for (j = 0; j < curr_sz; j += 1)
		{
		if(opt_fastmux)
			{
	            	fprintf(classname ? chandle: stdout, "s[%d]='0'+((t[%d]>>7)&1);",j, num-(curr_sz-j)+1);
			}
			else
			{
	            	fprintf(classname ? chandle: stdout, "s[%d]='0'+(t[%d]&1);",j, num-(curr_sz-j)+1);
			}
		}
         fprintf(classname ? chandle: stdout, "s[%d]=0;\n  return s;\n}",curr_sz);

	 if(curr_sz<=32)
		{
		if(classname)
			{
		        fprintf(hhandle, "unsigned int getint_be_%s();\n", curr_name);
		        fprintf(chandle, "\n\nunsigned int %s::getint_be_%s()\n{\n", classname, curr_name);
			}
			else
			{
		        printf("\n\nunsigned int getint_be_%s(void)\n{\n", curr_name);
			}
		fprintf(classname ? chandle: stdout, "return( ");
	         for (j = 0; j < curr_sz; j += 1)
			{
			if(opt_fastmux)
				{
		            	fprintf(classname ? chandle: stdout, "(((t[%d]>>7)&1)<<%d)",num-j, j);
				}
				else
				{
		            	fprintf(classname ? chandle: stdout, "((t[%d]&1)<<%d)",num-j, j);
				}
			if(j!=(curr_sz-1)) fprintf(classname ? chandle: stdout, " | ");
			}
	        fprintf(classname ? chandle: stdout, " );\n}");

		if(classname)
			{
		        fprintf(hhandle, "unsigned int getint_le_%s();\n", curr_name);
		        fprintf(chandle, "\n\nunsigned int %s::getint_le_%s()\n{\n", classname, curr_name);
			}
			else
			{
		        printf("\n\nunsigned int getint_le_%s(void)\n{\n", curr_name);
			}
		fprintf(classname ? chandle: stdout, "return( ");
	         for (j = 0; j < curr_sz; j += 1)
			{
			if(opt_fastmux)
				{
		            	fprintf(classname ? chandle: stdout, "(((t[%d]>>7)&1)<<%d)",num-(curr_sz-j)+1, j);
				}
				else
				{
		            	fprintf(classname ? chandle: stdout, "((t[%d]&1)<<%d)",num-(curr_sz-j)+1, j);
				}
			if(j!=(curr_sz-1)) fprintf(classname ? chandle: stdout, " | ");
			}
	        fprintf(classname ? chandle: stdout, " );\n}");
		}

         if (curr_type == 0)
         {                      /* input port also has set_* function */
	if(classname)
		{
          	fprintf(hhandle, "void set_le_%s(const char *s);\n", curr_name);
          	fprintf(chandle, "\n\nvoid %s::set_le_%s(const char *s)\n{\n  ", classname, curr_name);
		}
		else
		{
          	printf("\n\nvoid set_le_%s(const char *s)\n{\n  ", curr_name);
		}
          for (j = 0; j < curr_sz; j += 1)
		{
		if(opt_fastmux)
			{
	             	fprintf(classname ? chandle: stdout, "t[%d]=(s[%d]-'0')<<7;",num-j,j);
			}
			else
			{
	             	fprintf(classname ? chandle: stdout, "t[%d]=s[%d]-'0';",num-j,j);
			}
		}
          fprintf(classname ? chandle: stdout, "\n}");

	if(classname)
		{
          	fprintf(hhandle, "void set_be_%s(const char *s);\n", curr_name);
          	fprintf(chandle, "\n\nvoid %s::set_be_%s(const char *s)\n{\n  ", classname, curr_name);
		}
		else
		{
          	printf("\n\nvoid set_be_%s(const char *s)\n{\n  ", curr_name);
		}
          for (j = 0; j < curr_sz; j += 1)
		{
		if(opt_fastmux)
			{
             		fprintf(classname ? chandle: stdout, "t[%d]=(s[%d]-'0')<<7;",num-(curr_sz-j)+1,j);
			}
			else
			{
             		fprintf(classname ? chandle: stdout, "t[%d]=s[%d]-'0';",num-(curr_sz-j)+1,j);
			}
		}
          fprintf(classname ? chandle: stdout, "\n}");

	 if(curr_sz<=32)
		{
		if(classname)
			{
		        fprintf(hhandle, "void setint_be_%s(unsigned int v);\n", curr_name);
		        fprintf(chandle, "\n\nvoid %s::setint_be_%s(unsigned int v)\n{\n", classname, curr_name);
			}
			else
			{
		        printf("\n\nvoid setint_be_%s(unsigned int v)\n{\n", curr_name);
			}
	         for (j = 0; j < curr_sz; j += 1)
			{
			if(opt_fastmux)
				{
		            	fprintf(classname ? chandle: stdout, "t[%d] = ((v>>%d)<<7); ",num-j, j);
				}
				else
				{
		            	fprintf(classname ? chandle: stdout, "t[%d] = (v>>%d)&1; ",num-j, j);
				}
			}
	        fprintf(classname ? chandle: stdout, "\n}");

		if(classname)
			{
		        fprintf(hhandle, "void setint_le_%s(unsigned int v);\n", curr_name);
		        fprintf(chandle, "\n\nvoid %s::setint_le_%s(unsigned int v)\n{\n", classname, curr_name);
			}
			else
			{
		        printf("\n\nvoid setint_le_%s(unsigned int v)\n{\n", curr_name);
			}
	         for (j = 0; j < curr_sz; j += 1)
			{
			if(opt_fastmux)
				{
		            	fprintf(classname ? chandle: stdout, "t[%d] = ((v>>%d)<<7); ",num-(curr_sz-j)+1, j);
				}
				else
				{
		            	fprintf(classname ? chandle: stdout, "t[%d] = (v>>%d)&1; ",num-(curr_sz-j)+1, j);
				}
			}
	        fprintf(classname ? chandle: stdout, "\n}");
		}
         }
      }
      if (ss)
      {
         strcpy(curr_name, p1 + 1);
         curr_name[len] = 0;
         curr_type = port->type;
         curr_len = len;
         curr_sz = 1;
         curr_num = ss->number;
      }
      else
         break;
   }
}

if(classname)
	{
	fprintf(hhandle, "};");
	}
}

/**************************************************************************************************/

int sigcmp(char *s1, char *s2)
{
unsigned char c1, c2;
int u1, u2;
                                
for(;;)
        {
        c1=(unsigned char)*(s1++);
        c2=(unsigned char)*(s2++);
                                
        if((!c1)&&(!c2)) return(0);
        if((c1<='9')&&(c2<='9')&&(c2>='0')&&(c1>='0'))
                {
                u1=(int)(c1&15);
                u2=(int)(c2&15);
         
                while(((c2=(unsigned char)*s2)>='0')&&(c2<='9'))
                        {
                        u2*=10;
                        u2+=(unsigned int)(c2&15);
                        s2++;
                        }
                 
                while(((c2=(unsigned char)*s1)>='0')&&(c2<='9'))
                        {
                        u1*=10;
                        u1+=(unsigned int)(c2&15);
                        s1++;
                        }
 
                if(u1==u2) continue;
                        else return((int)u1-(int)u2);
                }
                else
                {
                if(c1!=c2) return((int)c1-(int)c2);
                }
        }
}


/*
 * Quicksort algorithm from p154 of
 * "Introduction to Algorithms" by Cormen, Leiserson, and Rivest.
 * 05-jul-97bsi
 */
                        
int ss_partition(struct symbol **a, int p, int r)
{
struct symbol *x, *t;
int i,j;
                         
x=a[p];
i=p-1;
j=r+1;
                         
while(1)
        {
        do
                {
                j--;
                } while(sigcmp(!a[j]->id?a[j]->name:a[j]->id,
				!x->id?x->name:x->id
				)>0);
                
        do      {
                i++;
                } while(sigcmp(!a[i]->id?a[i]->name:a[i]->id,
				!x->id?x->name:x->id
				)<0);

        if(i<j)  
                {
                t=a[i];
                a[i]=a[j];
                a[j]=t;
                }
                else
                {       
                return(j);
                }
        }
}


void ss_quicksort(struct symbol **a, int p, int r)
{
int q;
                         
if(p<r) 
        {
        q=ss_partition(a,p,r);
        ss_quicksort(a,p,q);
        ss_quicksort(a,q+1,r);
        }
}             


void mvl_gateprint(struct gate *g)
{
struct gateport *gp;
int i;

printf("%s", gatenames[g->type]);

printf(" %d",g->incnt);
gp=g->in;
for(i=0;i<g->incnt;i++)
	{
	printf(" %d",gp->port->number);
	gp=gp->next;
	}

printf(" %d", g->outcnt);
gp=g->out;
for(i=0;i<g->outcnt;i++)
	{
	printf(" %d",gp->port->number);
	gp=gp->next;
	}
printf("\n");
}


static char *compress_fac_str = NULL;
static int   compress_fac_len = 0;
void compress_fac(char *str)
{
int i;
int len = strlen(str);
int minlen = (len<compress_fac_len) ? len : compress_fac_len;

if(minlen>(126-33)) minlen=(126-33);	/* keep in printable range--most hierarchies won't be this big anyway */

if(compress_fac_str)
	{
	for(i=0;i<minlen;i++)
		{
		if(compress_fac_str[i]!=str[i]) break;
		}	
	printf("%c%s\n", (i+33), str+i);
	free(compress_fac_str);
	}
	else
	{
	printf("%c%s\n", (0+33), str);
	}

compress_fac_str = (char *) malloc((compress_fac_len=len)+1);
strcpy(compress_fac_str, str);	
}

void mvlstate(void)
{
struct gate *g;
struct gateport *gp;
int i,t;
struct symbol *ss;
struct symbol **facs;
int numfacs;
char *subst, ch;
int numlatches=0;
int pre_instruction_count=0, post_instruction_count=0, latch_instruction_count=0;
int aliases=0;
struct alias *a;

for(t=0;t<numgates;t++)
	{
	g=sortedgates[t];

	if(is_latch(g))
		{
		gp=g->out;
		for(i=0;i<g->outcnt;i++)
			{
			if(!gp->port->is_latch)
				{
				gp->port->is_latch=~0;
				numlatches++;
				}
			gp=gp->next;
			}
		}
	}


facs=(struct symbol **)calloc(numfacs=numsyms, sizeof(struct symbol *));

ss=symnumberone; i=0;
while(ss)
	{
	facs[i++]=ss;

	if(!ss->id) subst=ss->name; else subst=ss->id;
        while((ch=(*subst)))
                {
                if(ch=='/') { *subst=0x01; } /* forces sort at hier boundaries */
                subst++;
                }

	ss=ss->prev;
	}
ss_quicksort(facs,0,numfacs-1);
ss=symnumberone;
while(ss)
	{
	if(!ss->id) subst=ss->name; else subst=ss->id;
        while((ch=(*subst)))
                {
                if(ch==0x01) { *subst='/'; } /* reconstruct normal hier boundaries */
                subst++;
                }

	ss=ss->prev;
	}

printf("SYM %d\n",numfacs);
for(i=0;i<numfacs;i++)
{
ss=facs[i]; ss->number=i;
printf("FAC %c%s %d ", (ss->isport)?((ss->isport->type)?'O':'I'):('W'), (ss->is_latch)?"R":"", i);

compress_fac((!ss->id) ? ss->name : ss->id);
if(!ss->back)
	{
	if(ss->alias) aliases++;
	}
}

if(aliases)
	{
	for(i=0;i<numfacs;i++)
		{
		ss=facs[i];

		if(!ss->back)
			{
			if(ss->alias)
				{
				a=ss->alias;
				t=1;
				while(a->next)
					{
					t++;
					a=a->next;
					}

				a=ss->alias;
				printf("ALI %d %d", ss->number, t);
				while(a)
					{
					printf(" %d",a->sym->number);
					a=a->next;
					}
				printf("\n");				
				}
			}		
		}
	}

for(t=0;t<numgates;t++)
	{
	g=sortedgates[t];
	if(g->depth)
		{
		if(!is_latch(g))
			{
			pre_instruction_count += (3 + g->incnt + g->outcnt);
			}
			else
			{
			latch_instruction_count += (3 + g->incnt + g->outcnt);
			}
		}

	if(g->depth2)
		{
		if(!is_latch(g))
		post_instruction_count += (3 + g->incnt + g->outcnt);
		}
	}

if(pre_instruction_count)
	{
	printf("#\n# Prelatch combinational logic\n#\nPRE %d\n", pre_instruction_count);
	for(t=0;t<numgates;t++)
		{
		g=sortedgates[t];
	
		if(g->depth)
			{
			if(!is_latch(g))
			mvl_gateprint(g);
			}
			else
			{
			/* should never happen since we zeroed dangling nets */
                        if (!warn_disable)
			fprintf(stderr,"WARNING: %s.%d #%d is unreachable.\n",gatenames[g->type],g->number,g->delay);
			}
		}
	}

if(numlatches)
	{
	printf("#\n# Latches\n#\n");
	printf("LAT %d\n",latch_instruction_count);
	for(t=0;t<numgates;t++)
		{
		g=sortedgates[t];
	
		if(g->depth)
			{
			if(is_latch(g))
			mvl_gateprint(g);
			}
		}
	}

if(num_reverse_latches)
	{
	printf("#\n# Postlatch combinational logic\n#\nPST %d\n",post_instruction_count);
	for(t=0;t<numgates;t++)
		{
		if(sortedgates[t]->depth2) mvl_gateprint(sortedgates[t]);
		}
	}
free(facs);
}

/**************************************************************************************************/

void bdd_gateprint(struct symbol **facs, struct gate *g, int level)
{
struct gateport *gp, *gpin;
int i, j;

switch(g->type)
	{
	case AND:
	case WAND:
	case TRIAND:
	case BUFIF1:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = (", bddprefix, facs[gp->port->number]->id, level);
				gpin=g->in;
				for(j=0;j<g->incnt;j++)
					{
					printf(" %s%s_X%d",bddprefix, facs[gpin->port->number]->id, level);
					gpin=gpin->next;
					if(gpin) printf(" AND");
					}
				printf(" ) .\n");
				gp=gp->next;
				}
			break;
	case OR:
	case WOR:
	case TRIOR:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				for(j=0;j<g->incnt;j++)
					{
					printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
					gpin=gpin->next;
					if(gpin) printf(" OR");
					}
				printf(" ) .\n");
				gp=gp->next;
				}
			break;
	case XOR:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				for(j=0;j<g->incnt;j++)
					{
					printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
					gpin=gpin->next;
					if(gpin) printf(" XOR");
					}
				printf(" ) .\n");
				gp=gp->next;
				}
			break;
	case NOR:
	case NOTIF0:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = NOT (",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				for(j=0;j<g->incnt;j++)
					{
					printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
					gpin=gpin->next;
					if(gpin) printf(" OR");
					}
				printf(" ) .\n");
				gp=gp->next;
				}
			break;
	case NOT: 
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = NOT (",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				for(j=0;j<g->incnt;j++)
					{
					printf(" %s%s_X%d",bddprefix, facs[gpin->port->number]->id, level);
					gpin=gpin->next;
					}
				printf(" ) .\n");
				gp=gp->next;
				}
			break;
	case MUX:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = ite (",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				for(j=0;j<g->incnt;j++)
					{
					printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
					gpin=gpin->next;
					if(gpin) printf(" ,");
					}
				printf(" ) .\n");
				gp=gp->next;
				}
			break;

	case LATCH:
	case LATCH_N:
	case BUF:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level+1);
				gpin=g->in;
				printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
				printf(" ) .\n");
				gp=gp->next;
				}
			break;

	case NL_GATE:
	case NL_OPTIMIZED:
	case WIRE:
	case TRI:
	case TRI0:
	case TRI1:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
				printf(" ) .\n");
				gp=gp->next;
				}
			break;

	case NAND:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = NOT (",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				for(j=0;j<g->incnt;j++)
					{
					printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
					gpin=gpin->next;
					if(gpin) printf(" AND");
					}
				printf(" ) .\n");
				gp=gp->next;
				}
			break;
	case XNOR:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = NOT (",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				for(j=0;j<g->incnt;j++)
					{
					printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
					gpin=gpin->next;
					if(gpin) printf(" XOR");
					}
				printf(" ) .\n");
				gp=gp->next;
				}
			break;

	case SUPPLY0:
	case SUPPLY1:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);
				printf(" %s", (g->type==SUPPLY0)?"0":"1");
				printf(" ) .\n");
				gp=gp->next;
				}
			break;

	case NEGC:
	case NOTIF1:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				printf("let %s%s_X%d = ( (NOT",bddprefix,facs[gp->port->number]->id, level);
				gpin=g->in;
				printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
				gpin=gpin->next;
				printf(") AND ");
				printf(" %s%s_X%d",bddprefix,facs[gpin->port->number]->id, level);
				printf(" ) ) .\n");
				gp=gp->next;
				}
			break;

	case ADDC:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				struct gateport *gpa = g->in;
				struct gateport *gpb = gpa->next;
				struct gateport *gpc = gpb->next;

				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);

				printf(" ( %s%s_X%d AND %s%s_X%d) OR",bddprefix,facs[gpa->port->number]->id, level, bddprefix,facs[gpb->port->number]->id, level);
				printf(" ( %s%s_X%d AND %s%s_X%d) OR",bddprefix,facs[gpa->port->number]->id, level, bddprefix,facs[gpc->port->number]->id, level);
				printf(" ( %s%s_X%d AND %s%s_X%d)",bddprefix,facs[gpb->port->number]->id, level, bddprefix, facs[gpc->port->number]->id, level);
				printf(" ) .\n");
				gp=gp->next;
				}
			break;

	case SUBC:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				struct gateport *gpa = g->in;
				struct gateport *gpb = gpa->next;

				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);

				if(g->incnt==3)
					{
					struct gateport *gpc = gpb->next;
					printf(" ( %s%s_X%d AND (NOT %s%s_X%d) ) OR",bddprefix,facs[gpa->port->number]->id, level, bddprefix,facs[gpb->port->number]->id, level);
					printf(" ( %s%s_X%d AND %s%s_X%d) OR",bddprefix,facs[gpa->port->number]->id, level, bddprefix,facs[gpc->port->number]->id, level);
					printf(" ( (NOT %s%s_X%d) AND %s%s_X%d)",bddprefix,facs[gpb->port->number]->id, level, bddprefix,facs[gpc->port->number]->id, level);
					}
					else
					{
					printf(" ( %s%s_X%d OR (NOT %s%s_X%d) )",bddprefix,facs[gpa->port->number]->id, level, bddprefix,facs[gpb->port->number]->id, level);
					}
				printf(" ) .\n");
				gp=gp->next;
				}
				break;

	case LATCH_SR:
	case LATCH_SR_N:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				struct gateport *gpd = g->in;
				struct gateport *gps = gpd->next;
				struct gateport *gpr = gps->next;

				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);
				printf(" (%s%s_X%d OR %s%s_X%d) AND",bddprefix,facs[gpd->port->number]->id, level, bddprefix,facs[gps->port->number]->id, level);
				printf(" (NOT %s%s_X%d)",bddprefix,facs[gpr->port->number]->id, level);
				printf(" ) .\n");
				gp=gp->next;
				}
				break;
	case BUFIF0:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				struct gateport *gpd = g->in;
				struct gateport *gpc = gpd->next;

				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);
				printf(" %s%s_X%d AND",bddprefix,facs[gpd->port->number]->id, level);
				printf(" (NOT %s%s_X%d)",bddprefix,facs[gpc->port->number]->id, level);
				printf(" ) .\n");
				gp=gp->next;
				}
			break;

	case PMOS:
	case NMOS:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				fprintf(stderr, "Gate type '%s' not supported in assignment to %s%s_X%d, using '0'.\n", 
						gatenames[g->type], 
						bddprefix, facs[gp->port->number]->id, level);
				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);
				printf(" 0 ) .\n");
				gp=gp->next;
				}
			break;

	default:
			gp=g->out;
			for(i=0;i<g->outcnt;i++)
				{
				fprintf(stderr, "Gate type %d not supported in assignment to %s%s_X%d, using '0'.\n", 
						g->type, 
						bddprefix, facs[gp->port->number]->id, level);
				printf("let %s%s_X%d = (",bddprefix,facs[gp->port->number]->id, level);
				printf(" 0 ) .\n");
				gp=gp->next;
				}
			break;
	}
}

void bddstate(int maxlevel)
{
struct gate *g;
struct gateport *gp;
int i,t;
struct symbol *ss;
struct symbol **facs;
int numfacs;
char *subst, ch;
int numlatches=0;
int pre_instruction_count=0, latch_instruction_count=0;
int level;

int konst0index = -1, konst1index = -1;

for(t=0;t<numgates;t++)
	{
	g=sortedgates[t];

	if(is_latch(g))
		{
		gp=g->out;
		for(i=0;i<g->outcnt;i++)
			{
			if(!gp->port->is_latch)
				{
				gp->port->is_latch=~0;
				numlatches++;
				}
			gp=gp->next;
			}
		}
	}


facs=(struct symbol **)calloc(numfacs=numsyms, sizeof(struct symbol *));

ss=symnumberone; i=0;
while(ss)
	{
	facs[i++]=ss;

	if(!ss->id) subst=ss->name; else subst=ss->id;

        while((ch=(*subst)))
                {
                if(ch=='/') { *subst=0x01; } /* forces sort at hier boundaries */
                subst++;
                }

	ss=ss->prev;
	}
ss_quicksort(facs,0,numfacs-1);
ss=symnumberone;
while(ss)
	{
	if(!ss->id) subst=ss->name; else subst=ss->id;
        while((ch=(*subst)))
                {
                if(ch==0x01) { *subst='$'; } /* reconstruct normal hier boundaries */
                else if(ch=='!') { *subst='_'; } /* subst over the temp prefix */
                subst++;
                }

	ss=ss->prev;
	}

for(i=0;i<numfacs;i++)
{
ss=facs[i]; ss->number=i;

ss->id = (!ss->id) ? ss->name : ss->id;

if(*ss->id=='$')
	{
	if((konst0index<0)&&(!strcmp("$0", ss->id))) konst0index=i;
	else
	if((konst1index<0)&&(!strcmp("$1", ss->id))) konst1index=i;
	}

if((ss->isport)||(ss->is_latch))
	{
	printf("# FAC %c%s T%d ", (ss->isport)?((ss->isport->type)?'O':'I'):('W'), (ss->is_latch)?"R":"", i);
	printf("%s%s\n", bddprefix, ss->id);
	}
}

for(t=0;t<numgates;t++)
	{
	g=sortedgates[t];
	if(g->depth)
		{
		if(!is_latch(g))
			{
			pre_instruction_count += (3 + g->incnt + g->outcnt);
			}
			else
			{
			latch_instruction_count += (3 + g->incnt + g->outcnt);
			}
		}
	}

for(level=0;level<maxlevel;level++)
	{
	printf("\n#\n# BDD Logic Cycle %d\n#\n", level);

	if(konst0index>=0)
		{
		printf("\nlet $0_X%d = 0 .\n", level);
		}

	if(konst1index>=0)
		{
		if(konst0index<0) printf("\n");
		printf("let $1_X%d = 1 .\n", level);
		}

	if(pre_instruction_count)
		{
		printf("\n");
		for(t=0;t<numgates;t++)
			{
			g=sortedgates[t];
		
			if(g->depth)
				{
				if(!is_latch(g))
				bdd_gateprint(facs, g, level);
				}
				else
				{
				/* should never happen since we zeroed dangling nets */
	                        if (!warn_disable)
				fprintf(stderr,"WARNING: %s.%d #%d is unreachable.\n",gatenames[g->type],g->number,g->delay);
				}
			}
		}
	
	if(numlatches)
		{
		printf("\n");
		for(t=0;t<numgates;t++)
			{
			g=sortedgates[t];
		
			if(g->depth)
				{
				if(is_latch(g))
				bdd_gateprint(facs, g, level);
				}
			}
		}
	}

free(facs);
}

/**************************************************************************************************/

/*
 * add symbol to table.  no duplicate checking
 * is necessary as this is only called by symfind.
 */
struct symbol *symadd(const char *name)
{
struct symbol *s;

s=(struct symbol *)calloc(1,sizeof(struct symbol));
strcpy(s->name=(char *)malloc(strlen(name)+1),name);

s->number=++numsyms;	/* for temporary variable creation for c generator */

if(symchain)
	{
	symchain->prev=s;
	}
	else
	{
	symnumberone=s;
	}

s->nextinsymchain=symchain;
symchain=s;

return(s);
}

/*
 * find a slot already in the table...
 */
struct symbol *symfind(const char *s, int add_if_not_found)
{
struct symbol *sym;

symtree=ds_splay((char *)s, symtree);
if(!ds_success)
	{
	if(add_if_not_found)
		{
		sym=symadd(s);
		symtree=ds_insert(sym->name, symtree, (void *)sym);
		return(sym);
		}
		else
		{
		return(NULL);
		}
	}

return((struct symbol *)symtree->pnt);
}


/*
 * simple little tokenizer that reads ver generated
 * ivf files.
 */
int sep=0;

int nexttoken(void)
{
char *bptr;
char ch,typ;

if(feof(handle))
	{
	fprintf(stderr,"Short read.\n");
	exit(0);
	}

bptr=tok;
if(sep)
{
*(bptr++)=sep;
sep=0;
}
else
while(!feof(handle))
	{
	ch=(char)fgetc(handle);
	switch(ch)
		{
		case ' ':
		case '\t':
		case '\r':
		case '\n':
			typ=0;
			break;

		case '{':
		case '}':
		case ';':
		case '.':
			typ=1;
			break;

		default: 
			*(bptr++)=ch;
			typ=2;
			break;
		}

	if(typ==2) continue;
	if((bptr!=tok)&&(typ==0)) break;
	if(typ==1)
		{
		if(bptr==tok)
			{
			*(bptr++)=ch;
			break;
			}
			else
			{
			sep=ch;
			break;
			}
		}
	}
*bptr=0x00;

return(bptr!=tok);
}


/*
 * Mainline...
 */
int main(int argc, char **argv)
{
char *buff, *filename=NULL;
int i;
int bdd_levels;

tok=(char *)malloc(8192);
buff=(char *)malloc(8192);

for(i=1;i<argc;i++)
	{
	if((!strcmp(argv[i],"-h"))||(!strcmp(argv[i],"-help")))
                {
                printf( "Usage:\n"
                        "------\n"
                        "%s [input_filename] [options]\n\n"
			"-\t\taccept input file from stdin\n"
			"-a\t\tdisable alias generation when -w used in cyco chain\n"
			"-b\t\tgenerate Abel netlist (if -w not used)\n"
			"-bdd count\tgenerate BDD model with count sim cycles (if -w not used)\n"
			"-bddprefix name\tprepend prefix to all BDD signal names\n"
                        "-c\t\tenable coning locality code generation for c models\n"
                        "-class name\tspecify classname (generate c++ instead of c)\n"
			"-d\t\tensure dangling nets are not backtraced\n"
			"-fm\t\tuse fastmux compare semantics in c model (overrides -p)\n"
                        "-h[elp]\t\tprints this screen\n"
			"-i\t\tgenerate i/o port access functions\n"
                        "-L\t\talso remove unused latches (needs -w2)\n"
			"-m\t\tgenerate multivalue model (if -w not used)\n"
			"-nf\t\tdon't generate facs string array in c models\n"
			"-nts\t\tdo not smash temp variables out of c model\n"
			"-nvr\t\tdo not perform varreduct scheduling with c model\n"
			"-p\t\tuse parallel simulation semantics\n"
			"-P\t\tno constant propagation\n"
			"-s\t\tsplit c model functions into 100 line gcc optimizable blocks\n"
                        "-S\t\tno strength reduction (requires -w2)\n"
			"-stk\t\tbuild in stkfac capability for c models\n"
                        "-t\t\tdo not tidy netlist (leave unused gates)\n"
                        "-T\t\ttidy before extracting models (default: after)\n"
			"-tgcc width\tconvert AND/OR to ?: ops (requires gcc)\n"
                        "-w\t\trewrite ivf to stdio with most wires optimized out\n"
                        "-w2\t\trewrite ivf to stdio with additional optimizations\n"
                        "-W\t\tsuppress warnings about dangling nets\n"
                        "-x\t\tdo not extract models (i.e. add/subtract)\n"
                        "\n", argv[0]);

                exit(0);
                }
	else
	if(!strcmp(argv[i],"-a"))
		{
		alias_disable=~0;
		}
	else
	if(!strcmp(argv[i],"-b"))
		{
		abel_mode=~0;
		}
	else
	if(!strcmp(argv[i],"-nf"))
		{
		no_facs=~0;
		}
	else
	if(!strcmp(argv[i],"-x"))
		{
		no_model_extract=~0;
		}
	else
	if(!strcmp(argv[i],"-t"))
		{
		no_tidy=~0;
		}
	else
	if(!strcmp(argv[i],"-P"))
		{
		no_const_prop=~0;
		}
	else
	if(!strcmp(argv[i],"-L"))
		{
		remove_latches=~0;
		}
	else
	if(!strcmp(argv[i],"-T"))
		{
		tidy_before=~0;
		}
	else
        if(!strcmp(argv[i],"-c"))
                {
                cone_mode=~0;
                }
	else
	if(!strcmp(argv[i],"-p"))
		{
		parallel_sim=~0;
		}
	else
	if(!strcmp(argv[i],"-fm"))
		{
		opt_fastmux=~0;
		}
	else
	if(!strcmp(argv[i],"-m"))
		{
		mvl_mode=~0;
		}
	else
	if(!strcmp(argv[i],"-bdd"))
		{
		bdd_mode=~0;
		if(i!=(argc-1))
			{
			i++;
			sscanf(argv[i], "%d", &bdd_levels);
			}
			else
			{
			fprintf(stderr, "** Missing level depth after -bdd in command, exiting.\n");
			exit(255);
			}

		if(!bddprefix)
			{
			bddprefix=malloc(1);
			bddprefix[0]=0;		/* give it a null string */
			}
		}
	else
	if(!strcmp(argv[i],"-bddprefix"))
		{
		if(!bdd_mode) { bdd_mode=~0; bdd_levels=1; }
		if(i!=(argc-1))
			{
			int len=strlen(argv[++i]);
			if(bddprefix) 
				{
				if(bddprefix[0]!=0) fprintf(stderr, "** Specified -bddprefix at command line multiple times, using most recent.\n");
				free(bddprefix);
				}
			bddprefix = malloc(len+3+1);
			sprintf(bddprefix, "%s_$_", argv[i]);
			}
			else
			{
			fprintf(stderr, "** Missing hierarchy name after -bdd in command, exiting.\n");
			exit(255);
			}
		}
	else
	if(!strcmp(argv[i],"-w"))
               	{
                wireopt_mode=~0;
                }
	else
	if(!strcmp(argv[i],"-w2"))
               	{
                ivf_mode=~0;
                }
	else
	if(!strcmp(argv[i],"-s"))
               	{
                sim_gran=100;	/* block size to make gcc happy...else unlimited */
                }
	else
	if(!strcmp(argv[i],"-stk"))
               	{
                nonstick_facs=0;
                }
	else
	if(!strcmp(argv[i],"-S"))
               	{
                no_strength_reduce=~0;
                }
	else
	if(!strcmp(argv[i],"-nts"))
               	{
                no_tempsmash=~0;
                }
	else
	if(!strcmp(argv[i],"-nvr"))
               	{
                no_varreduct=~0;
                }
	else
	if(!strcmp(argv[i],"-tgcc"))
               	{
		if(i!=(argc-1))
			{
			opt_tgcc = atoi(argv[++i]);
			if(opt_tgcc<2) 
				{
				opt_tgcc = 0;
				}
			}
			else
			{
			fprintf(stderr, "** Missing width specifier after -tgcc in command, exiting.\n");
			exit(255);
			}
                }
	else
	if(!strcmp(argv[i],"-W"))
               	{
                warn_disable=~0;
                }
        else
	if(!strcmp(argv[i],"-d"))
               	{
                dangle_mode=~0;
                }
        else
	if(!strcmp(argv[i],"-i"))
               	{
                io_generate=~0;
                }
	else
	if(!strcmp(argv[i],"-class"))
		{
		if(classname)
			{
			fprintf(stderr, "** Classname specified twice (previously '%s'), exiting.\n", classname);
			exit(255);
			}
		else
		if(i!=(argc-1))
			{
			int clen, pos;
			int fail=0;
			i++;
			classname = argv[i];
			clen = strlen(classname);			
			fail = (clen<1);

			for(pos=0;pos<clen;pos++)
				{
				if(classname[pos]=='_') continue;
				if(!isalnum(classname[pos])) { fail = 1; break; }
				}

			if(!fail)
				{
				char *cname = (char *)malloc(clen+3);
				char *hname = (char *)malloc(clen+3);

				sprintf(cname, "%s.C", classname);
				sprintf(hname, "%s.H", classname);

				chandle=fopen(cname, "wb");
				hhandle=fopen(hname, "wb");
				if((!chandle)||(!hhandle))				
					{
					fprintf(stderr, "** Problem opening class output files, exiting.\n");
					exit(255);
					}

				free(cname); free(hname);
				}
				else
				{
				fprintf(stderr, "** Classname of '%s' after -class in command is illegal, exiting.\n");
				exit(255);
				}
			}
			else
			{
			fprintf(stderr, "** Missing classname after -class in command, exiting.\n");
			exit(255);
			}
		}
        else
        if(filename)
                {
		if(!strcmp(filename,"-"))
			{
			fprintf(stderr,"Using \"-\" option and a filename are mutually exclusive.\n");
			exit(0);
			}
			else
			{
                	fprintf(stderr,"Second filename encountered: \"%s\".\n",argv[i]);
                	exit(0);
			}
                }
        else
                {
                filename=argv[i];
                }
	}

if((!nonstick_facs)&&(sim_gran>0))
	{
	sim_gran=50;
	}	

if(filename)
        {
	if(!strcmp(filename,"-"))
		{
		handle=stdin;
		}
	else
        if(!(handle=fopen(filename,"rb")))
                {
                fprintf(stderr,"Could not open file \"%s\" for input!\n",filename);
                exit(0);
                }
	}
	else
	{
	fprintf(stderr, "Cyco: "CYCO_VERSION"\n");
	fprintf(stderr,"No filename specified.  Use '%s -h' for options.\n", argv[0]);
	exit(0);
	}

symtree=NULL;

if (wireopt_mode) nummode += 1;
if (mvl_mode) nummode += 1;
if (bdd_mode) nummode += 1;
if (abel_mode) nummode += 1;
if (ivf_mode) nummode += 1;
if (nummode > 1)
   {
   fprintf(stderr, "** incompatible command-line options, type `cyco -h' for help\n");
   exit(1);
   }

if (!wireopt_mode && !mvl_mode && !abel_mode && !ivf_mode & !bdd_mode)
   cyco_mode = 1;

if(wireopt_mode)
   {
   fprintf(stderr,"Cyco: IVF Wire Optimization "CYCO_VERSION"\n");
   }
else if(mvl_mode)
   {
   printf("#\n# Cyco: IVF MVL Cycle Compiler "CYCO_VERSION"\n#\n");
   fprintf(stderr,"Cyco: IVF MVL Cycle Compiler "CYCO_VERSION"\n");
   }
else if(bdd_mode)
   {
   printf("#\n# Cyco: IVF BDD Cycle Compiler "CYCO_VERSION"\n#\n");
   fprintf(stderr,"Cyco: IVF BDD Cycle Compiler "CYCO_VERSION"\n");
   }
else if(abel_mode)
   {
   printf("\"\n\" Cyco: IVF Abel translator "CYCO_VERSION"\n\"\n");
   fprintf(stderr,"Cyco: IVF Abel translator "CYCO_VERSION"\n");
   }
else if(ivf_mode)
   {
   fprintf(stderr,"Cyco: IVF Wire Optimization (+const prop) "CYCO_VERSION"\n");
   }
else if(cyco_mode)
   {
   fprintf(stderr,"Cyco: IVF Cycle Compiler "CYCO_VERSION"\n");
   fprintf(classname ? hhandle : stdout, "/*\nCyco: IVF Cycle Compiler "CYCO_VERSION"\n");
   if((opt_fastmux)&&(parallel_sim))
	{
	fprintf(stderr, "-- INFO: -p and -fm options both specified, disabling -p!\n");
	parallel_sim = 0;
	}
   if((opt_tgcc)&&(parallel_sim))
	{
	fprintf(stderr, "-- INFO: -p and -tgcc %d options both specified, creating\n", opt_tgcc);
	fprintf(stderr,"          non-parallel muxes for short-circuit gate ops!\n");
	}
   }


nexttoken();
if(strcmp("MX",tok)) { fprintf(stderr,"MX preamble not found.\n"); exit(0); }
nexttoken();
strcpy(buff,tok);
nexttoken();
if(strcmp(".",tok)) { fprintf(stderr,"'.' missing in model name.\n"); exit(0); }
strcat(buff,tok);
nexttoken();
strcat(buff,tok);
if(cyco_mode)
	{
	fprintf(classname ? hhandle : stdout, "Model name: %s\n",buff);
	}

nexttoken();
if(strcmp("{",tok)) { fprintf(stderr,"'{' missing after model name.\n"); exit(0); }
nexttoken();
if(!strcmp("MP",tok)) 
	{
	/* nothing */
	}
	else if(!strcmp("MIP",tok)) 
	{
	compressed_ids=1;
	}
	else
	{ 
	fprintf(stderr,"'MP' not found.\n"); 
	exit(0); 
	}

nexttoken();
if(strcmp("{",tok)) { fprintf(stderr,"'{' missing after MP/MIP.\n"); exit(0); }

/*
 * port section...
 */
while(1)
{
struct port *p;
struct symbol *sm;

nexttoken();
if(!strcmp("}",tok)) break;
p=(struct port *)calloc(1,sizeof(struct port));
sm=symfind(tok, 1);

if(compressed_ids)
	{
	nexttoken();
	if(!strcmp("}",tok)) { fprintf(stderr,"Missing name after ID in MIP section.\n"); exit(0); };
	strcpy(sm->id=(char *)malloc(strlen(tok)+1),tok);
	}

p->port=sm;
sm->isport=p;

nexttoken();
if(!strcmp("IN",tok))
	{
	p->type=0;
	}
else
if(!strcmp("OUT",tok))
	{
	p->type=1;
	}
else
	{
	fprintf(stderr,"'%s' unknown port type for port %s.\n",tok,compressed_ids?sm->id:sm->name);
	exit(0);
	}

p->next=ports;
ports=p;

if(cyco_mode)
	{
	fprintf(classname ? hhandle : stdout, "%3s %s",tok,(!sm->id)?sm->name:sm->id);
	if((no_facs)||(!no_varreduct))
		{
		fprintf(classname ? hhandle : stdout, " -> t[%d]\n",sm->number);
		}
		else
		{
		fprintf(classname ? hhandle : stdout, "\n");
		}
	}

nexttoken();
if(strcmp(";",tok)) { fprintf(stderr,"Missing semicolon for port %s.\n",sm->name); exit(0); }
}

nexttoken();
if(strcmp("}",tok)) { fprintf(stderr,"Missing '}' for MX section finish.\n"); exit(0); }


/*
 * symbol section...
 */
nexttoken();
if(strcmp("MF",tok)) { fprintf(stderr,"Missing MF section.\n"); exit(0); }
nexttoken();
if(strcmp("{",tok)) { fprintf(stderr,"Missing '{' after MF section preamble.\n"); exit(0); }

while(1)
{
nexttoken();
if(!strcmp("FS",tok))
	{
	nexttoken();
	if(strcmp("{",tok)) { fprintf(stderr,"Missing '{' after FS preamble.\n"); exit(0); }
	while(1)
		{
		nexttoken();
		if(!strcmp("}",tok)) break;
		else		
		symfind(tok, 1);
		}
	}
else
if(!strcmp("FSA",tok))
	{
	struct symbol *s, *t;
	struct alias *a;

	nexttoken();
	if(strcmp("{",tok)) { fprintf(stderr,"Missing '{' after FSA preamble.\n"); exit(0); }
	nexttoken();
	if(strcmp("}",tok))
		{
		s=symfind(tok, 1);

		while(1)
			{
			nexttoken();
			if(!strcmp("}",tok))break;
			if(!alias_disable)
				{
				t=symfind(tok, 1);
				t->zapped=~0;
                                t->flags |= SYMF_IS_ALIAS;
				a=(struct alias *)calloc(1,sizeof(struct alias));
				a->sym=t;
				a->next=s->alias;
				s->alias=a;
				}
			}
		}
	}
else
if(!strcmp("FIS",tok))
	{
	nexttoken();
	if(strcmp("{",tok)) { fprintf(stderr,"Missing '{' after FIS preamble.\n"); exit(0); }
	while(1)
		{
		struct symbol *s;
		nexttoken();
		if(!strcmp("}",tok)) break;
		else		
		s=symfind(tok, 1);

		nexttoken();
		if(!strcmp("}",tok)) { fprintf(stderr,"Missing name after ID in FI section.\n"); exit(0); };
		strcpy(s->id=(char *)malloc(strlen(tok)+1),tok);
		compressed_ids=~0;
		}
	}
else
if(!strcmp("FISA",tok))
	{
	struct symbol *s, *t;
	struct alias  *a;

	nexttoken();
	if(strcmp("{",tok)) { fprintf(stderr,"Missing '{' after FISA preamble.\n"); exit(0); }
	nexttoken();
	if(strcmp("}",tok))
		{
		s=symfind(tok, 1);	
	
		nexttoken();
		if(!strcmp("}",tok)) { fprintf(stderr,"Missing name after ID in FI section.\n"); exit(0); };
		strcpy(s->id=(char *)malloc(strlen(tok)+1),tok);
		compressed_ids=~0;

		while(1)
			{
			nexttoken();
			if(!strcmp("}",tok))break;
			if(!alias_disable)
				{
				t=symfind(tok, 1);
				t->zapped=~0;
                                t->flags |= SYMF_IS_ALIAS;
				a=(struct alias *)calloc(1,sizeof(struct alias));
				a->sym=t;
				a->next=s->alias;
				s->alias=a;
				}
			}
		}
	}
else
if(!strcmp("FG",tok))
	{
	struct gate *g;
	struct symbol *gs;
	struct gateport *gp, *gpwalk;
	struct sensitivity *sens;
	int i;

	g=(struct gate *)calloc(1,sizeof(struct gate));
	numgates++;

	nexttoken();
	for(i=0;i<GATETYPES;i++)
		{
		if(!strcmp(gatenames[i],tok)) break;
		}
	if(i==GATETYPES)
		{
		fprintf(stderr,"Unknown gate type: %s.\n", tok);
		exit(0);
		}
	
	g->type=i;
	nexttoken();
	if(strcmp(".",tok))
		{
		fprintf(stderr,"Missing '.' in gate instance.\n");
		exit(0);
		}

	nexttoken();
	g->number=atoi(tok);

	nexttoken();
	if(strcmp("{",tok))
		{
		fprintf(stderr,"Missing '{' in gate instance.\n");
		exit(0);
		}

	nexttoken();
	for(i=0;i<GATETYPES;i++)
		{
		if(!strcmp(gatenames[i],tok)) break;
		}
	if(i!=g->type)
		{
		fprintf(stderr,"Gate types %s and %s do not match.\n",gatenames[g->type],tok);
		exit(0);
		}

	nexttoken();
	g->delay=atoi(tok);

	nexttoken();
	if(strcmp("I",tok))
		{
		fprintf(stderr,"Missing 'I' in gate instance.\n");
		exit(0);
		}

	nexttoken();
	g->incnt=atoi(tok);
	for(i=0;i<g->incnt;i++)
		{
		nexttoken();	/* add input ports here */
		gs=symfind(tok, 1);		

		gp=(struct gateport *)calloc(1,sizeof(struct gateport));
		gp->port=gs;

		if(!i)	/* construct IN gateports in order */
			{
			g->in=gp;
			gpwalk=gp;
			}
			else
			{
			gpwalk->next=gp;
			gpwalk=gp;
			}			

		sens=(struct sensitivity *)calloc(1,sizeof(struct sensitivity));
		sens->gate=g;
		sens->next=gs->togate;
		gs->togate=sens;
		(gs->incount)++;		
		}

	nexttoken();
	g->outcnt=atoi(tok);
	for(i=0;i<g->outcnt;i++)
		{
		nexttoken();	/* add output ports here */
		gs=symfind(tok, 1);

		gp=(struct gateport *)calloc(1,sizeof(struct gateport));
		gp->port=gs;


		if(!i)	/* construct OUT gateports in order */
			{
			g->out=gp;
			gpwalk=gp;
			}
			else
			{
			gpwalk->next=gp;
			gpwalk=gp;
			}			

		sens=(struct sensitivity *)calloc(1,sizeof(struct sensitivity));
		sens->gate=g;
		sens->next=gs->fromgate;
		gs->fromgate=sens;
		(gs->outcount)++;		
		}

	nexttoken();
	if(strcmp("}",tok))
		{
		fprintf(stderr,"Missing '}' in gate instance.\n");
		exit(0);
		}

	g->next=gatehead;
	gatehead=g;
	}
else
if(!strcmp("}",tok))
	{
	if(cyco_mode)
		{
		fprintf(classname ? hhandle : stdout, "Netlist end encountered.\n");
		}
	break;
	}
}

fclose(handle);

if(wireopt_mode)
	{
	wireopt(buff);
	return(0);
	}

if(cyco_mode)
	{
	const char *prec = no_tempsmash?"":"precoalesce ";
        fprintf(classname ? hhandle : stdout, "Number of %sgates: %d, number of %ssymbols: %d\n", prec, numgates, prec, numsyms);
        }

add_latches_to_queue();
add_in_ports_to_queue();
add_const_nets_to_queue();
breadth_first_search();

add_out_ports_to_queue();
if(!dangle_mode)add_dangling_nets_to_queue();
breadth_first_search_rvs();

sortdepth();
if(cone_mode) cone();    

if(cyco_mode) fprintf(classname ? hhandle : stdout, "*/\n\n");

#ifdef ALLOW_SORTSTATS
sortstats();
#endif

if(cyco_mode)
	{
	tempsmash(sortedgates, numgates);
	cstate();
	fprintf(classname ? hhandle : stdout, "\n\n/* End */\n");
	if(classname)
		{
		fprintf(hhandle, "\n#endif\n");
		fprintf(chandle, "\n\n/* End */\n");
		}
	}
else if (mvl_mode)
	{
	mvlstate();
	printf("END\n");
	}
else if (bdd_mode)
	{
	bddstate((bdd_levels<1)?1:bdd_levels);
	}
else if (abel_mode)
	{
	abelstate(buff, numgates, sortedgates, ports, symchain);
	}
else if (ivf_mode)
	{
	ivfstate(buff, numgates, sortedgates, ports, symchain);
	}

free(buff);
free(tok);
return(0);
}

