/* 
 * Copyright (c) Tony Bybell, Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/*
 * cyco.h 27/08/97ajb
 */
#ifndef CYCO_H
#define CYCO_H

#define CYCO_VERSION "v1.3.32 (w)1999-2002 BSI"

#undef ALLOW_SORTSTATS	/* we don't need stats normally! */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "splay.h"

#define MAX_NAME 1025   /* max length of symbol name */

enum GateType
{ NL_GATE, NL_OPTIMIZED, NOT, AND, OR, XOR, NAND, NOR, XNOR, LATCH, LATCH_N,
        LATCH_SR, LATCH_SR_N, NEGC, ADDC, SUBC,
        BUF, BUFIF0, PMOS, BUFIF1, NMOS, NOTIF0, NOTIF1, MUX,
        WIRE, TRI, WAND, TRIAND, WOR, TRIOR, TRI0, TRI1, SUPPLY0, SUPPLY1,

	GATETYPES       /* counts number of gate types */
};

extern char *gatenames[];

#ifdef DEBUG_FEEDBACK
#define DEBUG   printf
#endif

struct port
{
struct port *next;
struct symbol *port;
int type;		/* 0=in, 1=out */
};

struct sensitivity
{
struct sensitivity *next;
struct gate *gate;
};

struct alias
{
struct alias *next;
struct symbol *sym;
};


struct symbol
{
struct symbol *nextinsymchain; /* for linked list of all symbols */
struct symbol *prev;		/* reverse of the above.  for fac refs array */

char *name;		/* full name or compressed name */
char *id;		/* full name if above name is compressed else NULL */
int incount, outcount;

struct sensitivity *fromgate;
struct sensitivity *togate;

struct port *isport;
signed char frozen;
int gate_min, gate_max;

int number;	/* symbol gets converted to T### format for c file */
char zapped;	/* set when this symbol gets optimized out of the model */
char is_latch;  /* set when this symbol is the q pin of a latch */

char is_valid;	/* set when all input requirements are met */
char is_valid2;	/* set when all output requirements are met (rvs) */

struct symbol *back;		/* for wireopt..get to "base" symbol */
struct alias *alias;		/* for wireopt (generates aliases of zapped syms) */
int flags;
struct bitvec *bitvec;          /* NULL if symbol is not a vector element */
int index;                      /* index in the bit vector */
};


struct bitvec
{
char *name;
int size;		
int flags;
int range_hi, range_lo;
struct symbol **bits;
};

/* flags for struct bitvec.flags */
#define BVF_ANALIZED       0x01
#define BVF_BROKEN         0x02
#define BVF_DECL_DONE      0x04


/* values for struct symbol.flags */
#define SYMF_USED          0x001
#define SYMF_IS_LATCH_Q    0x002
#define SYMF_CONST_0       0x004
#define SYMF_CONST_1       0x008
#define SYMF_CONST_X       0x010
#define SYMF_INPUT         0x020
#define SYMF_OUTPUT        0x040
#define SYMF_MODEL_USE     0x080
#define SYMF_REFERENCED    0x100
#define SYMF_MODEL_OUTPUT  0x200
#define SYMF_IS_ALIAS      0x400
#define SYMF_IS_EXPRESSION 0x800


struct gateport 
{
struct gateport *next;
struct symbol *port;
struct gate *expr;	/* if this inport is collapsed for a "recursive" gate def in order to kill temporaries */
};

struct gate
{
struct gate *next;

int type;
int number;
int delay;

int cone;       /* cone number that the gate belongs to (starting at 1)... */

int incnt, outcnt;

struct gateport *in;
struct gateport *out;

char visited;	/* set & keep upon visiting.  use to check for unused gates */
char visited2;	/* rvs visitation */
char feedback;	/* definite feedback loop at this gate */
char flags;     /* #define GF_* below */
int depth;	/* how far this gate is from inputs/latch outs */
int depth2;	/* reverse from outputs for 2nd combinational pass */
};

/* values for struct gate.flags */
#define GF_MODEL_USE  0x01       /* gate extracted into high-level model */
#define GF_REMOVED    0x02       /* gate removed from netlist */
#define GF_EXPRESSION 0x04       /* gate removed from netlist but is pointed to as an "expression" gate from a gateport */


struct gatequeue
{
struct gatequeue *next;
struct gate *gate;
};

extern struct gatequeue *queueroot, *queuecurrent;

enum adder_type {A_add, A_sub, A_neg};

struct adder
{
enum adder_type type;        /* A_add (A + B), A_sub (A - B), A_neg (-A)*/
char *name;
int size;
struct adder_section *vec;
struct adder *next;
};


struct adder_section         /* used during adder extraction */
{
struct symbol *y;            /* output */
struct symbol *a;            /* A operand */
struct symbol *b;            /* B operand */
struct gate *g_carry;
struct gate *g_sum;
};

/* public functions defined in extract.c */
struct bitvec *extract_vector_range(struct symbol *sym);
struct adder *extract_adder(struct gate* seed);
void cleanup_netlist(struct gate **sortedgates,
                     int numgates,
                     struct port *ports);
const char *gp_name(struct gateport *gp);
const char *sym_name(struct symbol *sym);
const char *sym_name_real(struct symbol *sym);
int is_const(struct symbol *sym);
int is_ioport(struct symbol *sym);
int is_latch(struct gate *g);

/* functions defined in cyco.c */
int nexttoken(void);
struct symbol *symadd(const char *);
struct symbol *symfind(const char *, int add_if_not_found);

void quicksort(struct gate **, int, int);
int partition(struct gate **, int, int);

/* from abel.c */
void abelstate(const char *modelname,
               int numgates,
               struct gate **sortedgates,
               struct port *ports,
               struct symbol *symchain);

/* from ivf.c */
void ivfstate(const char *modelname,
              int numgates,
              struct gate **sortedgates,
              struct port *ports,
              struct symbol *symchain);

/* from tempsmash.c */
void tempsmash(struct gate **sortedgates, 
	       int numgates);

#endif
