/* 
 * Copyright (c) Tony Bybell 2002.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/* 
 * tempsmash.c removes all gates with FO=1
 */

#include "cyco.h"

extern int no_tempsmash;  /* command line option */
extern int no_varreduct;  /* command line option */
struct symbol *symnumberone;
extern int numsyms, numgates;

static int maxsym;

struct symnum_allocate
{
struct symnum_allocate *next;
int number;
};

static struct symnum_allocate *symnum_pool=NULL;

int symnum_alloc(void)
{
if(symnum_pool)
	{
	int rv = symnum_pool->number;
	struct symnum_allocate *stemp = symnum_pool;
	symnum_pool = symnum_pool->next;
	free(stemp);
	return(rv);
	}
	else
	{
	return(maxsym++);
	}
}

void symnum_free(int number)
{
struct symnum_allocate *stemp = malloc(sizeof(struct symnum_allocate));
stemp->next = symnum_pool;
stemp->number = number;
symnum_pool = stemp;
}

void symbacktrack_free(struct symbol *s, int i)
{
struct gate *g = s->fromgate->gate;
struct gateport *gp = g->in;

while(gp)
	{
	struct symbol *s2 = gp->port;
	if(s2->flags & SYMF_IS_EXPRESSION)
		{
		symbacktrack_free(s2, i);
		s2->frozen = -4;
		}
		else
		{
		if(s2->frozen>=0)
		if((s2->gate_max == i) && (!(s2->flags & SYMF_IS_EXPRESSION)))
			{
			symnum_free(s2->number);
			s2->frozen = -5;
			}
		}
	gp=gp->next;
	}
}


void tempsmash(struct gate **sortedgates, int numgates)
{
struct gate *g;
int i;
int combine_numgates=0;
int combine_numsyms=0;
struct symbol *s;
int cnt;

   if (!no_tempsmash)
	{
	fprintf(stderr, "-- INFO: starting tempsmash...\n");

      	for (i = 0; i < numgates; i += 1)
      		{
		struct gateport *gp;
		int skipgate;

        	g = sortedgates[i];
        	if (g->flags & (GF_REMOVED | GF_MODEL_USE))
        		continue;

		switch(g->type)
			{
			case LATCH:
			case LATCH_N:
			case LATCH_SR:
			case LATCH_SR_N:
			case PMOS:
			case NMOS:	skipgate = 1; break;
			
			default:	skipgate = 0; break;
			}

		if(skipgate) continue;

		gp = g->out;
		if(!gp->next)	/* single outputs only */
			{
			s = gp->port;
			if((s->incount == 1)&&(!s->isport))
				{
				switch(s->togate->gate->type)
					{
					case LATCH:
					case LATCH_N:
					case LATCH_SR:
					case LATCH_SR_N:
					case PMOS:
					case NMOS:	skipgate = 1; break;
					
					default:	skipgate = 0; break;
					}

				if(skipgate) continue;

				combine_numgates++;
				/* fprintf(stderr, "*** Net '%s' has a fanout of one\n", s->name); */
				gp = s->togate->gate->in;
				s->zapped = ~0;
				while(gp)
					{
					if(gp->port == s)
						{
						gp->expr = g;				  /* point back to previous gate */
						g->flags |= (GF_REMOVED | GF_EXPRESSION); /* mark gate as deleted */			
						s->flags |= SYMF_IS_EXPRESSION;
						break;					  
						}
					gp=gp->next;
					}
				}
			}
		}

	s=symnumberone;
	cnt=1;
	while(s)
		{
		if((!s->togate)&&(!s->isport))
			{
			combine_numsyms++;
			s->zapped = ~0;
			}
		else
		if(!s->zapped)
			{
			s->number = cnt++;
			}
		
		s=s->prev;
		}

	combine_numsyms += combine_numgates;

	fprintf(stderr,"-- INFO: coalesced %1d gates out of %1d (%1.1f%%)\n",
        	combine_numgates, numgates, 100.0*(double)combine_numgates/numgates);

	fprintf(stderr,"-- INFO: removed %1d symbols out of %1d (%1.1f%%)\n",
        	combine_numsyms, numsyms, 100.0*(double)combine_numsyms/numsyms);

	numsyms -= combine_numsyms;

	fprintf(stderr,"-- INFO: model size is %1d symbols / %1d gates\n",
        	numsyms, numgates-combine_numgates);
	}
		
if(!no_varreduct)
	{
	fprintf(stderr, "-- INFO: starting varreduct scheduling...\n");

	s=symnumberone;
	maxsym=1;
	while(s)
		{
		s->gate_max = -1;
		s->gate_min = numgates;

		if((!s->fromgate)||(s->isport))
			{
			if(!s->zapped)
				{
				s->number = maxsym++;
				s->frozen = -1;
				}
			}
		
		s=s->prev;
		}


      	for (i = 0; i < numgates; i += 1)
      		{
		struct gateport *gp;
		int freezegate;

        	g = sortedgates[i];
        	if ((g->flags & (GF_REMOVED | GF_MODEL_USE)) && (!(g->flags & GF_EXPRESSION)))
        		continue;

		switch(g->type)
			{
			case LATCH:
			case LATCH_N:
			case LATCH_SR:
			case LATCH_SR_N:
			case PMOS:
			case NMOS:	freezegate = 1; break;
			
			default:	freezegate = 0; break;
			}

		if(freezegate)
			{
			gp = g->in;
			while(gp)
				{
				s=gp->port;
				if(!s->frozen)
					{
					s->number = maxsym++;
					s->frozen = -2;
					}
				gp=gp->next;
				}
			gp = g->out;
			while(gp)
				{
				s=gp->port;
				if(!s->frozen)
					{
					s->number = maxsym++;
					s->frozen = -2;
					}
				gp=gp->next;
				}

			}
			else
			{
			gp = g->in;
			while(gp)
				{
				s=gp->port;
				if(!s->frozen)
					{
					if(s->gate_min > i) s->gate_min = i;
					if(s->gate_max < i) s->gate_max = i;
					}
				gp=gp->next;
				}
			gp = g->out;
			while(gp)
				{
				s=gp->port;
				if(!s->frozen)
					{
					if(s->gate_min > i) s->gate_min = i;
					if(s->gate_max < i) s->gate_max = i;
					}
				gp=gp->next;
				}
			}
		}

      	for (i = 0; i < numgates; i += 1)
      		{
		struct gateport *gp;
		struct gate *togate;
		int freezegate;

        	g = sortedgates[i];
        	if (!(g->flags & GF_EXPRESSION))
        		continue;

		togate = g;
		do
			{
			togate = togate -> out -> port -> togate -> gate;
			} while(togate->flags & GF_EXPRESSION);

		gp = g->in;
		while(gp)
			{	
			if (togate->out->port->gate_max > gp->port->gate_max) gp->port->gate_max = togate->out->port->gate_max;
			gp=gp->next;
			}
		}

      	for (i = 0; i < numgates; i += 1)
      		{
		struct gateport *gp;

        	g = sortedgates[i];
        	if (g->flags & (GF_REMOVED | GF_MODEL_USE))
        		continue;

		gp = g->in;
		while(gp)
			{
			s=gp->port;
			if(!s->zapped)
			if(s->frozen>=0)
				{
				if((s->gate_min == i) && (!s->frozen) && (!(s->flags & SYMF_IS_EXPRESSION)))
					{
					s->number = symnum_alloc();
					s->frozen = 1;
					}
				}
			gp=gp->next;
			}
		gp = g->out;
		while(gp)
			{
			s=gp->port;
			if(!s->zapped)
			if(s->frozen>=0)
				{
				if((s->gate_min == i) && (!s->frozen) && (!(s->flags & SYMF_IS_EXPRESSION)))
					{
					s->number = symnum_alloc();
					s->frozen = 1;
					}
				}
			gp=gp->next;
			}
		gp = g->in;
		while(gp)
			{
			s=gp->port;
			if(!s->zapped)
				{
				if(s->frozen>=0)
					{
					if((s->gate_max == i)&&(!(s->flags & SYMF_IS_EXPRESSION)))
						{
						symnum_free(s->number);
						s->frozen = -3;
						}
					}
				}
			else
			if(s->flags & SYMF_IS_EXPRESSION)
				{
				if(s->frozen>=0)
					{
					symbacktrack_free(s, i);
					s->frozen = -4;
					}
				}
			
			gp=gp->next;
			}
		gp = g->out;
		while(gp)
			{
			s=gp->port;
			if(!s->zapped)
			if(s->frozen>=0)
				{
				if(s->gate_max == i)
					{
					symnum_free(s->number);
					s->frozen = -3;
					}
				}
			gp=gp->next;
			}
		}
	fprintf(stderr, "-- INFO: scheduled %d symbols into %d (%1.1f%%)\n", numsyms, maxsym, 100.0*(double)maxsym/numsyms); 
	if(numsyms < maxsym)
		{
		fprintf(stderr, "-- INFO: reverting to original scheduler (%d symbols)\n", numsyms); 

		s=symnumberone;
		cnt=1;
		while(s)
			{
			if((!s->togate)&&(!s->isport))
				{
				combine_numsyms++;
				s->zapped = ~0;
				}
			else
			if(!s->zapped)
				{
				s->number = cnt++;
				}
			
			s=s->prev;
			}
		}
		else
		{
		numsyms = maxsym;
		}
      	}

}

