/* 
 * Copyright (c) Tony Bybell 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */


#include <string.h>
#include <stdlib.h>
#include "bindings.h"
#include "main.h"

short maxx, maxy;

struct line *textroot=NULL;
struct line *readend=NULL, *textend=NULL;
struct line *topline=NULL;
struct line *curline=NULL;
int current_line_base, current_line_row, current_line_col;
int total_lines;
int read_only=0;

void ed_clear_screen(void)
{
move(0,0);
clear();
refresh();
}

void find_current_line(void)
{
struct line *l;
int i;

l=topline;
for(i=0;i<current_line_row;i++)
	{
	l=l->next;
	}
curline=l;
}

void ed_status_bar(void)
{
int currow;
int i, x1, y1;

currow=current_line_base+current_line_row+1;
standout();
move(maxy-1,0);
if(read_only) printw("RO ");
printw("%3d%% of %d: %d.%d",currow*100/total_lines,
	total_lines,currow,curline->offset+current_line_col+1);
getyx(stdscr,y1,x1);
for(i=x1;i<=maxx;i++)
	{
	addch(' ');
	}
standend();
}

void ed_draw_line(short i, struct line *pos, short mx, short my, int offset)
{
short curx;
int j;

	{
	char *pnt;
	if((pnt=pos->str))
		{
		if(!(pos->offset=offset)) { move(i,(curx=j=0)); }
		else	                  { move(i,0); addch('$'); curx=1; j=offset+1; pnt+=j; }

		for(;j<pos->len;j++)
			{
			char ch;
			ch=*(pnt++);

			if((curx==mx)&&(j!=(pos->len-1)))
				{
				addch('$');
				break;
				}
			else
				{
				if(ch<32) /* works b/c it's signed */
					{
					ch=(ch=='\t')?' ':'.';
					}
				addch(ch);
				curx++;
				}
			}	
		}

	if(curx!=mx)
	while(curx<=mx)
		{
		addch(' ');
		curx++;
		}
	}
}

void ed_refresh_screen(void)
{
struct line *pos;
short mx,my, i, curx;
int x, y, j;

pos=topline;
getyx(stdscr,y,x);

my=maxy-1;
mx=maxx-1;

move(0,0);
clear();
 
for(i=0;((i<my)&&(pos));i++)
	{
	ed_draw_line(i, pos, mx, my, 0);
	pos=pos->next;
	}

ed_status_bar();
move(y,x);
refresh();
}

struct line *ed_read_file(char *name, struct line *inspos)
{
FILE *handle;
struct line *lineroot=NULL, *linecurr=NULL, *l;

errno=0;

if((handle=fopen(name,"rb")))
{
for(;;)
{
char buf[16385];
char *pnt, *pnt2;
int j, curx;

if(feof(handle)) break;
buf[0]=0;
fgets(buf,16384,handle);
l=(struct line *)calloc(1,sizeof(struct line));

curx=0;
l->len=strlen(pnt=buf);
if(l->len) l->len--;
for(j=0;j<l->len;j++)
	{
	curx=((*(pnt++))=='\t')?((curx+8)&(~7)):(curx+1);
	}

pnt=buf;
pnt2=l->str=(char *)calloc(1,curx+1);
curx=0;
if(l->len)
	{
	for(j=0;j<l->len;j++)
		{
		char ch;

		ch=*(pnt++);
		if(ch=='\t')
			{
			int ocx, delta, k;

			ocx=curx;
			curx=(curx+8)&(~7);
			
			delta=curx-ocx;
			for(k=0;k<delta;k++)
				{
				*(pnt2++)='\t';			
				}
			}
			else
			{
			curx++;
			*(pnt2++)=ch;			
			}
		}
	l->len=curx;
	}

if(!lineroot)
	{
	lineroot=linecurr=l;
	}
	else
	{
	l->prev=linecurr;
	linecurr->next=l;
	linecurr=l;
	}
total_lines++;
}
fclose(handle);
}

errno=0;

if(!lineroot)
	{
	lineroot=linecurr=(struct line *)calloc(1,sizeof(struct line));
	total_lines++;
	}

readend=linecurr;
return(lineroot);
}


void ed_sigstop(int dummy)
{
signal(SIGTSTP, ed_sigstop);
}

void ed_sigint(int dummy)
{
signal(SIGINT, ed_sigint);
}


void ed_die(int dummy)
{
ed_clear_screen();
signal(SIGINT, SIG_IGN);
mvcur(0,COLS-1, LINES-1, 0);
endwin();
exit(0);
}

int main(int argc, char **argv)
{
if(argc<2)
	{
	printf("Need to specify a filename!\n");
	exit(1);
	}

initscr();
getmaxyx(stdscr,maxy,maxx);

read_only=0;

scrollok(stdscr,FALSE);
noecho();
signal(SIGINT, ed_die);
signal(SIGTSTP, ed_sigstop);
/* signal(SIGINT, ed_sigint); */

total_lines=0;
curline=topline=textroot=ed_read_file(argv[1],NULL);
textend=readend;

current_line_base=current_line_row=current_line_col=0;

ed_refresh_screen();
mainloop();

/* sleep(7); */
ed_die(0);
return(0);
}

