%{

/* 
 * Copyright (c) Tony Bybell 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include <stdio.h>
#include "symbol.h"

extern int yylineno;
int tempnum=0;

struct quad *quadhead=NULL, *quadcurr=NULL;
struct quad **estackhead=NULL, **estack=NULL;

struct statement *statehead=NULL, *statecurr=NULL;
struct statement **sstackhead=NULL, **sstack=NULL;

struct symbol *gentemp(void)
{
char buff[32];
sprintf(buff,"%%T%d",++tempnum);

return(symfind(buff));
}

void printquad(struct quad *q)
{
struct symbol *qs1, *qs2, *qd;
int op;

qs1=q->op1;
qs2=q->op2;
qd=q->res;
op=q->oper;

if((op==OP_EQ)||(op==OP_NEG))
DEBUG("Quad: %6s %6s %6s %6s\n",
	qs1->name, "------", qd->name, operatorstrings[op]);
else
DEBUG("Quad: %6s %6s %6s %6s\n",
	qs1->name, qs2->name, qd->name, operatorstrings[op]);
}

void genquad(struct symbol *s1, struct symbol *s2, struct symbol *dest, int op)
{
struct symbol *qs1, *qs2, *qd;
struct quad *q;

qs1=(struct symbol *)s1;
qs2=(struct symbol *)s2;
qd =(struct symbol *)dest;

q=(struct quad *)calloc(1,sizeof(struct quad));
q->op1=qs1;
q->op2=qs2;
q->res=qd;
q->oper=op;

if(quadhead)
	{
	quadcurr->next=q;
	quadcurr=q;
	}
	else
	{
	quadhead=quadcurr=q;
	}
}

%}

%union {
	int i;
        void *pnt;
}

%token GENERATE BEG END TO LP RP 
%token <pnt> LABEL 
%token NL 
%token <pnt> TEXT 
%token <i>   NUMBER 
%token SEMI MACRO MACROEND 
%token <i>   HEX
%token LSH RSH ADD SUB MPY DIV MOD AND OR XOR NOT NNOT 
%token LT GT LET GET EQV NEQ AAND OOR EQ
%token EVAL IF ENDIF ELSE WHILE WEND FOR NEXT COMMA STRING
%right EQ

%type <pnt> expression fetchedlabel number num genbylist
%type <pnt> val0 val1 val2 val3 val4 val5 val6 val7 val8 val9 val10 val11 stringlist STRING
%type <i>   numhex

%%

main:	gentex2 ;

gentex2: gentex | gentex gentex2 ;

gentex:	gen | genstr | tex | macro | eval | ifstmt | whilestmt | forstmt ;

eval:	EVAL LP expression RP SEMI	
	{
	struct quad *q;
	struct statement *s;

	q=*(--estack);
	s=(struct statement *)calloc(1,sizeof(struct statement));
	s->type=ST_EVAL;
	s->u.eval=q;	

#ifdef DEBUG_PARSE
	DEBUG("--\n");		
	DEBUG("EVAL quads for %s:\n", ((struct symbol *)$3)->name);
	while(q)
		{
		printquad(q);
		q=q->next;
		}
	DEBUG("--\n");		
#endif

	if(statehead)
		{
		statecurr->next=s;
		statecurr=s;
		}
		else
		{
		statehead=statecurr=s;
		}
	}

tex:	TEXT
	{
	struct statement *s;

	DEBUG("TEXT: %s\n",(char *)$1);

	s=(struct statement *)calloc(1,sizeof(struct statement));
	s->type=ST_TEXT;
	s->u.text=(char *)($1);	

	if(statehead)
		{
		statecurr->next=s;
		statecurr=s;
		}
		else
		{
		statehead=statecurr=s;
		}
	}

gen:	 genhead gentex2 gentail | genhead gentail ;

genhead: GENERATE LABEL LP expression TO expression RP SEMI
	 {
	 struct quad *q1, *q2;
	 struct statement *s;

	 DEBUG("Generate %s(%s to %s)\n",
	 ((struct symbol *)$2)->name,
	 ((struct symbol *)$4)->name,
	 ((struct symbol *)$6)->name);

	q2=*(--estack);	/* fetch expression sequences.. */
	q1=*(--estack);

	s=(struct statement *)calloc(1,sizeof(struct statement));
	s->type=ST_GEN;
	s->u.gen=(struct generate *)calloc(1,sizeof(struct generate));
	s->u.gen->sym=(struct symbol *)($2);
	s->u.gen->from=q1;
	s->u.gen->to  =q2;

	if(statehead)
		{
		statecurr->next=s;
		statecurr=s;
		}
		else
		{
		statehead=statecurr=s;
		}

	*(sstack++)=statehead;	/* save current statement pos */
	*(sstack++)=statecurr;
	statehead=NULL;		/* start list over...will get attached */
	statecurr=NULL;		/* the the next field of s later */

#ifdef DEBUG_PARSE
	DEBUG("--\n");		
	DEBUG("Q1 quads:\n");
	while(q1)
		{
		printquad(q1);
		q1=q1->next;
		}

	DEBUG("Q2 quads:\n");
	while(q2)
		{
		printquad(q2);
		q2=q2->next;
		}
	DEBUG("--\n");		
#endif
	 }

gentail: END SEMI
	{
	struct statement *oldsh;

	oldsh=statehead;

	statecurr=*(--sstack);
	statehead=*(--sstack);
	statecurr->u.gen->next=oldsh;

	DEBUG("Generate tail\n");
	}


genstr:	 genstrhead gentex2 genstrtail | genstrhead genstrtail ;

genstrhead: GENERATE genbylist SEMI
	{
	struct statement *s;

	s=(struct statement *)calloc(1,sizeof(struct statement));
	s->type=ST_GST;
	s->u.gst=$2;

	if(statehead)
		{
		statecurr->next=s;
		statecurr=s;
		}
		else
		{
		statehead=statecurr=s;
		}

	*(sstack++)=statehead;	/* save current statement pos */
	*(sstack++)=statecurr;
	statehead=NULL;		/* start list over...will get attached */
	statecurr=NULL;		/* the the next field of s later */
	}

genstrtail: END SEMI
	{
	struct statement *oldsh;

	oldsh=statehead;

	statecurr=*(--sstack);
	statehead=*(--sstack);
	statecurr->u.gst->next=oldsh;

	DEBUG("Generate string tail\n");
	}


genbylist: LABEL LP stringlist RP COMMA genbylist
	{
	struct generate_string *gst;
	gst=(struct generate_string *)calloc(1,sizeof(struct generate_string));
	gst->sym=(struct symbol *)($1);
	gst->head=(struct gen_str *)($3);
	gst->gnext=$6;
	$$=gst;
	}
	| LABEL LP stringlist RP
	{
	struct generate_string *gst;
	gst=(struct generate_string *)calloc(1,sizeof(struct generate_string));
	gst->sym=(struct symbol *)($1);
	gst->head=(struct gen_str *)($3);
	$$=gst;
	}

stringlist: STRING COMMA stringlist
	{
	struct gen_str *g;

	g=(struct gen_str *)calloc(1,sizeof(struct gen_str));
	g->str=$1;
	g->next=(struct gen_str *)$3;
	$$=g;
	}
	| STRING
	{
	struct gen_str *g;

	g=(struct gen_str *)calloc(1,sizeof(struct gen_str));
	g->str=$1;
	$$=g;
	}

ifstmt:	 ifhead gentex2 else gentex2 elseend |
	 ifhead gentex2 iftail | 
	 ifhead gentex2 else elseend |
	 ifhead else gentex2 elseend |
	 ifhead else elseend ;


ifhead:  IF LP expression RP SEMI
	 {
	 struct quad *q1;
	 struct statement *s;

	 DEBUG("If (%s)\n",
	 ((struct symbol *)$3)->name);

	q1=*(--estack);	/* fetch expression sequences.. */

	s=(struct statement *)calloc(1,sizeof(struct statement));
	s->type=ST_IF;
	s->u.ifstmt=(struct ifstmt *)calloc(1,sizeof(struct ifstmt));
	s->u.ifstmt->cond=q1;

	if(statehead)
		{
		statecurr->next=s;
		statecurr=s;
		}
		else
		{
		statehead=statecurr=s;
		}

	*(sstack++)=statehead;	/* save current statement pos */
	*(sstack++)=statecurr;
	statehead=NULL;		/* start list over...will get attached */
	statecurr=NULL;		/* the the next field of s later */

#ifdef DEBUG_PARSE
	DEBUG("--\n");		
	DEBUG("Q1 quads:\n");
	while(q1)
		{
		printquad(q1);
		q1=q1->next;
		}

	DEBUG("--\n");		
#endif
	 }


iftail: ENDIF
	{
	struct statement *oldsh;

	oldsh=statehead;

	statecurr=*(--sstack);
	statehead=*(--sstack);
	statecurr->u.ifstmt->next=oldsh;

	DEBUG("If tail\n");
	}

else:   ELSE
	{
	struct statement *oldsh;

	oldsh=statehead;

	statecurr=*(--sstack);
	statehead=*(--sstack);
	statecurr->u.ifstmt->next=oldsh;

	*(sstack++)=statehead;	/* save current statement pos */
	*(sstack++)=statecurr;
	statehead=NULL;		/* start list over...will get attached */
	statecurr=NULL;		/* the the next field of s later */

	DEBUG("If tail (else head)\n");
	}

elseend:ENDIF
	{
	struct statement *oldsh;

	oldsh=statehead;

	statecurr=*(--sstack);
	statehead=*(--sstack);
	statecurr->u.ifstmt->elsestmt=oldsh;

	DEBUG("else tail\n");
	}


whilestmt: whilehead gentex2 whiletail |
	whilehead whiletail ;

whilehead: WHILE LP expression RP SEMI
	 {
	 struct quad *q1;
	 struct statement *s;

	 DEBUG("While (%s)\n",
	 ((struct symbol *)$3)->name); 

	q1=*(--estack);	/* fetch expression sequences.. */

	s=(struct statement *)calloc(1,sizeof(struct statement));
	s->type=ST_WHILE;
	s->u.whilestmt=(struct whilestmt *)calloc(1,sizeof(struct whilestmt));
	s->u.whilestmt->cond=q1;

	if(statehead)
		{
		statecurr->next=s;
		statecurr=s;
		}
		else
		{
		statehead=statecurr=s;
		}

	*(sstack++)=statehead;	/* save current statement pos */
	*(sstack++)=statecurr;
	statehead=NULL;		/* start list over...will get attached */
	statecurr=NULL;		/* the the next field of s later */

#ifdef DEBUG_PARSE
	DEBUG("--\n");		
	DEBUG("Q1 quads:\n");
	while(q1)
		{
		printquad(q1);
		q1=q1->next;
		}

	DEBUG("--\n");		
#endif
	 }


whiletail: WEND
	{
	struct statement *oldsh;

	oldsh=statehead;

	statecurr=*(--sstack);
	statehead=*(--sstack);
	statecurr->u.whilestmt->next=oldsh;

	DEBUG("While tail\n");
	}


forstmt: forhead gentex2 fortail |
	 forhead fortail ;

forhead: FOR LP expression COMMA expression COMMA expression RP SEMI
	 {
	 struct quad *q1, *q2, *q3;
	 struct statement *s;

	 DEBUG("For (%s, %s, %s)\n",
	 ((struct symbol *)$3)->name,
	 ((struct symbol *)$5)->name,
	 ((struct symbol *)$7)->name);

	q3=*(--estack);	/* fetch expression sequences.. */
	q2=*(--estack);	
	q1=*(--estack);

	s=(struct statement *)calloc(1,sizeof(struct statement));
	s->type=ST_FOR;
	s->u.forstmt=(struct forstmt *)calloc(1,sizeof(struct forstmt));
	s->u.forstmt->start=q1;
	s->u.forstmt->cond =q2;
	s->u.forstmt->inc  =q3;

	if(statehead)
		{
		statecurr->next=s;
		statecurr=s;
		}
		else
		{
		statehead=statecurr=s;
		}

	*(sstack++)=statehead;	/* save current statement pos */
	*(sstack++)=statecurr;
	statehead=NULL;		/* start list over...will get attached */
	statecurr=NULL;		/* the the next field of s later */

#ifdef DEBUG_PARSE
	DEBUG("--\n");		
	DEBUG("Q1 quads:\n");
	while(q1)
		{
		printquad(q1);
		q1=q1->next;
		}

	DEBUG("Q2 quads:\n");
	while(q2)
		{
		printquad(q2);
		q2=q2->next;
		}

	DEBUG("Q3 quads:\n");
	while(q3)
		{
		printquad(q3);
		q3=q3->next;
		}
	DEBUG("--\n");		
#endif
	 }

fortail: NEXT
	{
	struct statement *oldsh;

	oldsh=statehead;

	statecurr=*(--sstack);
	statehead=*(--sstack);
	statecurr->u.forstmt->next=oldsh;

	DEBUG("For tail (next was encountered)\n");
	}


macro:	MACRO expression MACROEND 
	{
	struct quad *q;
	struct statement *s;

	q=*(--estack);
	s=(struct statement *)calloc(1,sizeof(struct statement));
	s->type=ST_MACRO;
	s->u.mac=q;	

#ifdef DEBUG_PARSE
	DEBUG("--\n");		
	DEBUG("MACRO quads for %s:\n", ((struct symbol *)$2)->name);
	while(q)
		{
		printquad(q);
		q=q->next;
		}
#endif

	if(statehead)
		{
		statecurr->next=s;
		statecurr=s;
		}
		else
		{
		statehead=statecurr=s;
		}
	}


fetchedlabel:	LABEL
		{
		/* symbol pointer is passed up in $$ */
		}

numhex:		NUMBER | HEX
		{
		/* symbol pointer is passed up in $$ */
		}

num:		numhex
		{
		struct symbol *s;
		char numbuff[32];	/* more than enough */
		sprintf(numbuff,"%d",$1);
		(s=symfind(numbuff))->value=$1;		
		s->isconstant=~0;  /* to keep from redefining consts! */
		$$=s;
		}

number:         num | fetchedlabel ;

val0:           LP val11 RP 
			{
			$$=$2; /* nothing necessary to do! */
			} 
		| SUB val0 
			{
			genquad($2,0,$$=gentemp(), OP_NEG);
			} 
		| ADD val0
			{
			genquad($2,0,$$=gentemp(), OP_EQ);
			}
		| NOT val0
			{
			genquad($2,0,$$=gentemp(), OP_NOT);
			}
		| NNOT val0
			{
			genquad($2,0,$$=gentemp(), OP_NNOT);
			}
		| number ;

val1:           val1 MPY val0 
			{
			genquad($1,$3,$$=gentemp(), OP_MUL);
			} 
		| val1 DIV val0 
			{
			genquad($1,$3,$$=gentemp(), OP_DIV);
			} 
		| val1 MOD val0 
			{
			genquad($1,$3,$$=gentemp(), OP_MOD);
			} 
		| val0 ;

val2:           val2 ADD val1 
			{
			genquad($1,$3,$$=gentemp(), OP_ADD);
			} 
		| val2 SUB val1 
			{
			genquad($1,$3,$$=gentemp(), OP_SUB);
			} 
		| val1 ;

val3:           val3 LSH val2 
			{
			genquad($1,$3,$$=gentemp(), OP_LSH);
			} 
		| val3 RSH val2 
			{
			genquad($1,$3,$$=gentemp(), OP_RSH);
			} 
		| val2 ;

val4:		val4 LT val3
			{
			genquad($1,$3,$$=gentemp(), OP_LT);
			}
		| val4 GT val3
			{
			genquad($1,$3,$$=gentemp(), OP_GT);
			}
		| val4 LET val3
			{
			genquad($1,$3,$$=gentemp(), OP_LET);
			}
		| val4 GET val3
			{
			genquad($1,$3,$$=gentemp(), OP_GET);
			}
		| val3 ;


val5:		val5 EQV val4
			{
			genquad($1,$3,$$=gentemp(), OP_EQV);
			}
		| val5 NEQ val4
			{
			genquad($1,$3,$$=gentemp(), OP_NEQ);
			}
		| val4 ;


val6:		val6 AND val5
			{
			genquad($1,$3,$$=gentemp(), OP_AND);
			} 
		| val5 ;


val7:		val7 XOR val6
			{
			genquad($1,$3,$$=gentemp(), OP_XOR);
			} 
		| val6 ;


val8:		val8 OR val7
			{
			genquad($1,$3,$$=gentemp(), OP_OR);
			} 
		| val7 ;

val9:		val9 AAND val8
			{
			genquad($1,$3,$$=gentemp(), OP_AAND);
			} 
		| val8 ;

val10:		val10 OOR val9
			{
			genquad($1,$3,$$=gentemp(), OP_OOR);
			} 
		| val9 ;

val11:		val10 EQ val11
			{
			/* 
			 * note: this doesn't complain about invalid
			 * lvalue stuff like 2+a=4...
			 */
			struct symbol *s;
			s=(struct symbol *)$1;
			if(s->isconstant) /* assign to a const?!? */
				{	  /* don't allow it!      */
				genquad($3,0,$$=gentemp(), OP_EQ);
				}
				else
				{
				genquad($3,0,$1, OP_EQ);
				}
			} 
		| val10 ;

expression:     val11
                {
		struct symbol *s;

		genquad($1,0,$$=gentemp(), OP_EQ);
			/* so there is always an init quad to run! */

		s=(struct symbol *)$$;
		DEBUG("Top of expression: %s\n",s->name);
		*(estack++)=quadhead;
		quadhead=quadcurr=NULL;	/* new expression starting.. */
                }

%%
