/* 
 * Copyright (c) Tony Bybell 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include "symbol.h"

struct symbol **sym=NULL;
int numsyms=0;

char *operatorstrings[]=
{"eq", "neg", "add", "sub", "mul", "div", "mod", "lsh", "rsh",
 "and", "xor", "or", "not", "nnot", 
 "lt", "gt", "let", "get", "eqv", "neq", "aand", "oor"
};


/*
 * handle debugging stuff...
 */
#ifndef DEBUG_PARSE
void DEBUG(char *dummy, ...) { }
#endif

#ifndef DEBUG_SYM
void DEBUG2(char *dummy, ...) { }
#endif


/*
 * Generic hash function for symbol names...
 */
int hash(char *s)
{
char *p;
unsigned int h=0, g;
for(p=s;*p;p++)
        {
        h=(h<<4)+(*p);
        if((g=h&0xf0000000))
                {
                h=h^(g>>24);
                h=h^g;
                }   
        }
return(h%SYMPRIME);
}


/*
 * add symbol to table.  no duplicate checking
 * is necessary as this is only called by symfind.
 */
struct symbol *symadd(char *name, int hv)
{
struct symbol *s;

numsyms++;

s=(struct symbol *)calloc(1,sizeof(struct symbol));
strcpy(s->name=(char *)malloc(strlen(name)+1),name);
s->next=sym[hv];
sym[hv]=s;

return(s);
}

/*
 * find a slot already in the table...
 */
struct symbol *symfind(char *s)
{
int hv;
struct symbol *temp;

hv=hash(s);
if(!(temp=sym[hv])) return(symadd(s,hv)); /* no table made, add */

while(temp)
        {
        if(!strcmp(temp->name,s))
                {
                return(temp); /* in table already */    
                }
        if(!temp->next) break;
        temp=temp->next;
        }

return(symadd(s,hv)); /* not found, add */
}

