/* 
 * Copyright (c) Tony Bybell 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/*
 * symbol.h 25/09/97ajb
 */
#ifndef SYMBOL_H
#define SYMBOL_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef DEBUG_PARSE
#define DEBUG   printf
#endif

#ifndef DEBUG_PARSE
#define DEBUG   blackhole_parse
void    DEBUG(char *, ...);
#endif


#define SYMPRIME 211
#define FATAL_ERROR 666

struct symbol
{
struct symbol *next;
char *name;

char *string;	/* null means isn't string, else use string value */
int isconstant;	/* to detect against constant redefs */
int value;
int assigned;
};

enum OperatorType
{ OP_EQ, OP_NEG, OP_ADD, OP_SUB, OP_MUL, OP_DIV, OP_MOD, OP_LSH, OP_RSH,
  OP_AND, OP_XOR, OP_OR, OP_NOT, OP_NNOT, 
  OP_LT, OP_GT, OP_LET, OP_GET, OP_EQV, OP_NEQ, OP_AAND, OP_OOR
};

enum StatementType
{ ST_TEXT, ST_MACRO, ST_GEN, ST_GST, ST_EVAL, ST_IF, ST_WHILE, ST_FOR };

struct statement
{
int type;
union 	{
	char *text;
	struct quad *mac;
	struct quad *eval;
	struct generate *gen;
	struct generate_string *gst;
	struct ifstmt *ifstmt;
	struct forstmt *forstmt;
	struct whilestmt *whilestmt;
	} u;

struct statement *next;
};


struct generate
{
struct symbol *sym;
struct quad *from;
struct quad *to;
struct statement *next;
};

struct gen_str
{
char *str;
struct gen_str *next;
};

struct generate_string
{
struct generate_string *gnext;
struct symbol *sym;
struct gen_str *head;
struct statement *next;
};

struct forstmt
{
struct quad *start;
struct quad *cond;
struct quad *inc;
struct statement *next;
};

struct ifstmt
{
struct quad *cond;
struct statement *next;
struct statement *elsestmt;
};

struct whilestmt
{
struct quad *cond;
struct statement *next;
};


struct quad		/* macros are a series of quads */
{
struct quad *next;

struct symbol *op1;
struct symbol *op2;
struct symbol *res;
int oper;
};


struct symbol *symfind(char *);
void printquad(struct quad *);
void exec_stmt(struct statement *);
int exec_quad(struct quad *);
void exec_gen(struct generate *);
void exec_gst(struct generate_string *);
void exec_if(struct ifstmt *);
void exec_while(struct whilestmt *);
void exec_for(struct forstmt *);

extern char *operatorstrings[];

extern char *yytext;
extern int yylineno;
extern struct symbol **sym;
extern int numsyms;

extern struct quad *quadhead, *quadcurr;
extern struct quad **estackhead, **estack;

extern struct statement *statehead, *statecurr;
extern struct statement **sstackhead, **sstack;

#endif

