/* 
 * Copyright (c) Tony Bybell, Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/*
 * flatten.c
 * 19/08/97ajb
 */

#include "symbol.h"
#include "flatten.h"
#include "y.tab.h"

int mode_pica_vsim_heritage=0; /* set to ~0 by the -pica command line flag */
int usecompressed_ids=0;  /* set to ~0 by -c command line flag */
int noivf=0;	          /* set to ~0 by -n command line flag */

struct module *topmodule=NULL, *matchmodule=NULL;
struct flatmodule *flatmodule=NULL;
int portmismatches=0;
int notgs=0, andgs=0, orgs=0, xorgs=0, xnorgs=0, nandgs=0, norgs=0, nullgs=0;
int latchgs=0, wiregs=0, wandgs=0, worgs=0, muxs=0;
int trigs=0, triandgs=0, triorgs=0, tri0gs=0, tri1gs=0, supply0gs=0, supply1gs=0, bufgs=0;
int bufif0gs=0, bufif1gs=0, notif0gs=0, notif1gs=0, nmosgs=0, pmosgs=0;

int negcgs=0, addcgs=0, subcgs=0;

FILE *ivffile=NULL;

/*
 * Find matching module for selective flatten...
 */
void findmatchingmodule(struct module *node, char *str)
{
if((!node)||(matchmodule)) return; /* early abort if found already */
if(node->left) findmatchingmodule(node->left,str);
if(!strcmp(node->name,str)) {matchmodule=node; return;}
if(node->right) findmatchingmodule(node->right,str);
}


/*
 * Find the first top level module for unguided flatten...
 */
void findtopmodule2(struct module *node)
{
if((!node)||(topmodule)) return; /* early abort if found already */
if(node->left) findtopmodule2(node->left);
if((!node->referencecount)&&(node->comps)) {topmodule=node; return;}
if(node->right) findtopmodule2(node->right);
}

void findtopmodule(struct module *node)
{
if((!node->left)&&(!node->right)) { topmodule=node; return; }
else findtopmodule2(node);
}

/*
 * Recurse through the syntax tree and flatten as you walk along...
 */
void flatten2entry(struct module *thismod, char *oldprefix, char *suffix)
{
char fname[300];

if(!noivf)
	{
	strcpy(fname,thismod->name);
	strcat(fname,".ivf");
	}
	else
	{
	strcpy(fname,"/dev/null");	/* quick and dirty ivf suppression */
	}

if((ivffile=fopen(fname,"w")))
	{
	flatten2(thismod, oldprefix, suffix);
	dumpflatsym();
	dequeuestring();
	fprintf(ivffile,"}\n");
	fclose(ivffile);
	}
	else
	{
	mesg("** Could not open output file \"%s\" for output.\n\n",fname);
	exit(255);
	}
}

void flatten2(struct module *thismod, char *oldprefix, char *suffix)
{
char *prefix;
int prefixlength;
struct component *comp;
struct gate *gates;
struct wire *wires;
struct reg *reg;
int gatenum;
char line[1025];

if(!strlen(suffix))
	{
	struct gateport *gp;
	int i;

	fprintf(ivffile,"MX %s.logic{\n",thismod->name);
	fprintf(ivffile,(!usecompressed_ids)?"MP{":"MIP{");
	gp=thismod->moduleports;

	while(gp)
		{
		for(i=gp->port->lorange;i<=gp->port->hirange;i++)
			{
			if(!usecompressed_ids)
				{
				fprintf(ivffile,"%s/%s/%d", thismod->name, gp->port->name, i);
				}
				else
				{
				int len;
				char *buff;
				struct symslot *s;
				len=strlen(thismod->name)+strlen(gp->port->name)+32;
				buff=(char *)ver_alloca(len);
				sprintf(buff,"%s/%s/%d", thismod->name, gp->port->name,i);				
				s=addflatsym(buff);
				fprintf(ivffile,"%s %s", s->id, s->name);	
				}

			if(gp->port->inaccess) { fprintf(ivffile," IN;"); }
			else
			if(gp->port->outaccess) { fprintf(ivffile," OUT;"); }
			else { fprintf(ivffile," UNK;"); }
			}

		gp=gp->next;
		}	

	fprintf(ivffile,"}\n}\n\nMF{\n");
	}

prefix=(char *)malloc(strlen(oldprefix)+strlen(suffix)+2);
strcpy(prefix,oldprefix);
strcat(prefix,(const char *)(suffix));
strcat(prefix,"/");
prefixlength=strlen(prefix)+1;

if((reg=thismod->reg))
	{
	while(reg)
	{
	char *str;

	flatmodule->numregs+=(!reg->not_a_latch);	/* so the count at the end statistics is correct */

	str=(char *)malloc(prefixlength+strlen(reg->port->name));
	strcpy(str,prefix);
	strcat(str,reg->port->name);
	addflatsym(str);
	free(str);

	/* behave.c actually instantiates the gate for the latch so we don't have to do it here */

	reg=reg->next;
	}
	}


if((wires=thismod->wires))
	{
	while(wires)
	{
	struct wire *w;
	char *str;

	flatmodule->numwires++;

	w=(struct wire *)calloc(1,sizeof(struct wire));
	w->delayr = wires->delayr;
	w->delayf = wires->delayf;
	w->delayz = wires->delayz;

	if(!flatmodule->wires)
        	{
		flatmodule->wires=w;
		}
		else
		{
		w->next=flatmodule->wires;
		flatmodule->wires=w;
        	}

	if((wires->port1)&&(wires->port2))
	{
	str=(char *)malloc(prefixlength+strlen(wires->port1->name));
	strcpy(str,prefix);
	strcat(str,wires->port1->name);
	w->port1=addflatsym(str);
	free(str);

	str=(char *)malloc(prefixlength+strlen(wires->port2->name));
	strcpy(str,prefix);
	strcat(str,wires->port2->name);
	w->port2=addflatsym(str);
	free(str);

	if(!usecompressed_ids)
		{
		sprintf(line,"FG NL_GATE.%d{NL_GATE %d I 1 %s 1 %s }\n",++nullgs,
			MAX3(w->delayr,w->delayf,w->delayz)*10,w->port2->name, w->port1->name);
		}
		else
		{
		sprintf(line,"FG NL_GATE.%d{NL_GATE %d I 1 %s 1 %s }\n",++nullgs,
			MAX3(w->delayr,w->delayf,w->delayz)*10,w->port2->id, w->port1->id);
		}

	queuestring(line);
	}
	else
	{
	warn("** Internal error in wire flattening!\n");
	}

	wires=wires->next;
	}
	}

if((gates=thismod->gates))
	{
	while(gates)
		{
		struct gate *g;
		char gstr[12];  /* only used in debug... */

		flatmodule->numgates++;

		g=(struct gate *)calloc(1,sizeof(struct gate));
		g->delayr = gates->delayr;
		g->delayf = gates->delayf;
		g->delayz = gates->delayz;
		g->type   = gates->type;

		if(!flatmodule->gates)
        		{
		        flatmodule->gates=g;
		        }
		        else
		        {
		        g->next=flatmodule->gates;
		        flatmodule->gates=g;
        		}
		
		switch(g->type)
        		{
			case NEGC:	strcpy(gstr,"NEGC");
					gatenum=++negcgs; break;
			case ADDC:	strcpy(gstr,"ADDC");
					gatenum=++addcgs; break;
			case SUBC:	strcpy(gstr,"SUBC");
					gatenum=++subcgs; break;

			case WIRE:	strcpy(gstr,"WIRE");
					gatenum=++wiregs; break;
			case TRI:	strcpy(gstr,"TRI");
					gatenum=++trigs; break;
			case WAND:	strcpy(gstr,"WAND");
					gatenum=++wandgs; break;
			case TRIAND:	strcpy(gstr,"TRIAND");
					gatenum=++triandgs; break;
			case WOR:	strcpy(gstr,"WOR");
					gatenum=++worgs; break;
			case TRIOR:	strcpy(gstr,"TRIOR");
					gatenum=++triorgs; break;
			case TRI0:	strcpy(gstr,"TRI0");
					gatenum=++tri0gs; break;
			case TRI1:	strcpy(gstr,"TRI1");
					gatenum=++tri1gs; break;
			case SUPPLY0:	strcpy(gstr,"SUPPLY0");
					gatenum=++supply0gs; break;
			case SUPPLY1:	strcpy(gstr,"SUPPLY1");
					gatenum=++supply1gs; break;

        		case NMOS:      strcpy(gstr,"NMOS");
					gatenum=++nmosgs; break;
        		case PMOS:      strcpy(gstr,"PMOS");
					gatenum=++pmosgs; break;

        		case BUF:       strcpy(gstr,"BUF");
					gatenum=++bufgs; break;
        		case BUFIF0:    strcpy(gstr,"BUFIF0");
					gatenum=++bufif0gs; break;
        		case BUFIF1:    strcpy(gstr,"BUFIF1");
					gatenum=++bufif1gs; break;

        		case NOT:       strcpy(gstr,"NOT");
					gatenum=++notgs; break;
        		case NOTIF0:    strcpy(gstr,"NOTIF0");
					gatenum=++notif0gs; break;
        		case NOTIF1:    strcpy(gstr,"NOTIF1");
					gatenum=++notif1gs; break;

        		case AND:       strcpy(gstr,"AND");
					gatenum=++andgs; break;
        		case OR:        strcpy(gstr,"OR");   
					gatenum=++orgs; break;
        		case XOR:       strcpy(gstr,"XOR");
					gatenum=++xorgs; break;
        		case NAND:      strcpy(gstr,"NAND");
					gatenum=++nandgs; break;
        		case NOR:       strcpy(gstr,"NOR");
					gatenum=++norgs; break;
        		case XNOR:      strcpy(gstr,"XNOR"); 
					gatenum=++xnorgs; break; 
                        case MUX:       strcpy(gstr,"MUX");
                                        gatenum=++muxs; break;
        		case LATCH:     strcpy(gstr,"LATCH");
					gatenum=++latchgs; break; 
        		case LATCH_N:   strcpy(gstr,"LATCH_N");
					gatenum=++latchgs; break; 
        		case LATCH_SR:  strcpy(gstr,"LATCH_SR");
					gatenum=++latchgs; break; 
        		case LATCH_SR_N:strcpy(gstr,"LATCH_SR_N");
					gatenum=++latchgs; break; 

        		case CONST0:
			case CONST1:
			case CONSTX:
			case CONSTZ:	strcpy(gstr,"NL_GATE"); 
					gatenum=++nullgs; break;
        		default:        strcpy(gstr,"ERROR");
					gatenum=0; break;
        		}

		sprintf(line,"FG %s.%d{%s %d I ",gstr,gatenum,gstr,MAX3(g->delayr,g->delayf,g->delayz)*10);
		queuestring(line);
		g->port=flattengate(gates->port,prefix,prefixlength,g->type); 

		gates=gates->next;
		}
	}

if((comp=thismod->comp))
	{
	while(comp)
		{
		int alo, ahi, isarr, arinc, array, arcnt, ardepth;
		int offset;

		if(comp->pnums)	/* patch in param module name */
			{
			char mbuf[2048];
			struct pnumlist *p2=comp->pnums;
			int len;
			int cnt=0;
			struct module *m;

			sprintf(mbuf, "%s#", comp->type->name);

                        len=strlen(mbuf);
                        while(p2)
                                {
                                char buf[128];
                                if(cnt++)
                                        {
					if(p2->valid)
						{
	                                        sprintf(buf,"_%d", p2->value);
						}
						else
						{
						strcpy(buf, "_");
						}
                                        strcpy(mbuf+len, buf);
                                        len+=strlen(buf);
                                        }
                                        else
                                        {
					if(p2->valid)
						{
	                                        sprintf(buf,"%d", p2->value);
	                                        strcpy(mbuf+len, buf);
	                                        len+=strlen(buf);
						}
						else
						{
						/* nothing */
						}
                                        }
                                p2=p2->next;
                                }

			m=simple_modfind(modroot, mbuf);
			if(m) 
				{
				comp->type = m;
				comp->pnums = NULL; /* no longer needed here */
				}
			}


		alo=comp->arraystart;
		ahi=comp->arrayfin;
		isarr=comp->isarray;
		if(alo<=ahi) { arinc=1; ardepth=ahi-alo+1;} 
			else {arinc=-1; ardepth=alo-ahi+1;}
		arcnt=0;

for(array=alo;;array+=arinc)	/* array loop */
	{
		if(comp->type->visited)
			{
			mesg("Module recursion error in module (%s) via module (%s).\n",comp->type->name,thismod->name);
			exit(255);
			}

/*********************************************************************************/
/* FIXME */
if(comp->pnums)
	{
	mesg("Skipping component: %s with parameters (%d, ...)\n", comp->type->name, comp->pnums->value);
	if(array!=ahi) break; else continue;
	}
/* FIXME */
/*********************************************************************************/



{
struct gateport *gpm, *g;
struct metavec *meta;
char *prefix2;
char *ds1, *ds2;
int prefix2len;
struct wire *w;
int numports=0;

meta=comp->port;
g=gpm=comp->type->moduleports;

while(g)
	{
	numports++;
	g=g->next;
	}

if(numports!=meta->length)
{
mesg("** Port Count Mismatch!  "
     "   Prefix: (%s), component: (%s), type: (%s), line: (%d).\n",
     prefix,comp->name,comp->type->name,
     comp->linenumber);
portmismatches++; yyerrors++;
break;
}
else
{
int i;

if(!isarr)
{
prefix2=(char *)malloc(prefix2len=prefixlength+strlen(comp->name)+1);
strcpy(prefix2,prefix);
strcat(prefix2+prefixlength-1,comp->name);
strcpy(prefix2+prefix2len-2,"/");
}
else
{
prefix2len=prefixlength+strlen(comp->name)+1;
prefix2=(char *)malloc(prefix2len+16);
strcpy(prefix2,prefix);
strcat(prefix2+prefixlength-1,comp->name);
sprintf(prefix2+prefix2len-2,"[%d]/",array);
prefix2len=strlen(prefix2)+1;
}

g=gpm;
for(i=0;i<numports;i++)
	{
        if (g->port->outaccess)
        {  /* check that outport is not connected to a reg */
           struct symslot *sym = meta->vec[i]->bits[0];
           struct reg *reg = thismod->reg;
           for (; reg != 0; reg = reg->next)
              if (reg->port == sym)
              {
                 char s[200], *p;
                 strcpy(s,sym->name);
                 p = strrchr(s,'/');
                 if (p != 0) *p = 0;
                 mesg("** Component '%s' in module '%s' cannot drive reg '%s' in line %d\n",
                      comp->name, thismod->name, s, comp->linenumber);
                 portmismatches += 1;
                 break;
              }
        }
	offset=(g->port->hirange-g->port->lorange+1)*arcnt;

	if((ardepth*(g->port->hirange-g->port->lorange+1))!=(meta->vec[i]->length))
		{
		if(!arcnt)
		{
		if(meta->vec[i]->bits[0]->flags & SF_UNCONNECTED)
			{
			mesg("** Unconnected Port (%s) "
	                     "Prefix: (%s), component: (%s), type: (%s), line: (%d).\n",
			     g->port->name,
	                     prefix,comp->name,comp->type->name,
	                     comp->linenumber);
			}
			else
			{
			mesg("** Port Size Mismatch!\n"
	                     "Prefix: (%s), component: (%s), type: (%s), line: (%d).\n",
	                     prefix,comp->name,comp->type->name,
	                     comp->linenumber);

			mesg("\tOffending port (#%d) is (%d) bits wide,\n"
	                     "\t it conflicts with definition of (%s) of (%d) bits wide\n",
	                     numports-i,
	                     meta->vec[i]->length,g->port->name,
	                     ardepth*(g->port->hirange-g->port->lorange+1));
			mesg("\t across %d instances.\n",ardepth);
	                portmismatches++; yyerrors++;
			}
		}
		}
		else
		{
		struct bitvec *b;
                int j, ds2len;
                        
                ds2=(char *)malloc(prefix2len+strlen(g->port->name)+16);
                        strcpy(ds2,prefix2);
                        strcpy(ds2+prefix2len-1,g->port->name);
                        ds2len=strlen(ds2);
                
		b=meta->vec[i];
		for(j=0;j<((b->length)/ardepth);j++)
			{
			ds1=(char *)malloc(prefixlength+
				strlen(b->bits[j+offset]->name)+4);
			strcpy(ds1,prefix);
			strcpy(ds1+prefixlength-1,b->bits[j+offset]->name);	

			if(g->port->dir==-1)
			sprintf(ds2+ds2len,"/%d",j+g->port->lorange);
			else
			sprintf(ds2+ds2len,"/%d",g->port->hirange-j);

			flatmodule->numwires++;
			w=(struct wire *)calloc(1,sizeof(struct wire));
			
			if(!flatmodule->wires)
				{
				flatmodule->wires=w;
				}
				else
				{
				w->next=flatmodule->wires;
				flatmodule->wires=w;
				}

			w->port1=addflatsym(ds1);
			w->port2=addflatsym(ds2);

			if(g->port->inaccess)
				{
				if(!usecompressed_ids)
					{
					sprintf(line,"FG NL_GATE.%d{NL_GATE %d I 1 %s 1 %s }\n",++nullgs,
					        MAX3(w->delayr,w->delayf,w->delayz)*10,w->port1->name, w->port2->name);
					}
					else
					{
					sprintf(line,"FG NL_GATE.%d{NL_GATE %d I 1 %s 1 %s }\n",++nullgs,
					        MAX3(w->delayr,w->delayf,w->delayz)*10,w->port1->id, w->port2->id);
					}
				queuestring(line);
				}
			if(g->port->outaccess)
				{
				if(!usecompressed_ids)
					{
					sprintf(line,"FG NL_GATE.%d{NL_GATE %d I 1 %s 1 %s }\n",++nullgs,
					        MAX3(w->delayr,w->delayf,w->delayz)*10,w->port2->name, w->port1->name);
					}
					else
					{
					sprintf(line,"FG NL_GATE.%d{NL_GATE %d I 1 %s 1 %s }\n",++nullgs,
					        MAX3(w->delayr,w->delayf,w->delayz)*10,w->port2->id, w->port1->id);
					}

				queuestring(line);
				}
			free(ds1);
			}
		free(ds2);
		}
	g=g->next;
	}


/* end of block */
free(prefix2);
}
}
if(!isarr)
	{
	comp->type->visited=~0;
	flatten2(comp->type,prefix,comp->name);
	comp->type->visited=0;
	}
	else
	{
	char *nam;
	nam=(char *)calloc(1,strlen(comp->name)+16);
	sprintf(nam,"%s[%d]",comp->name,array);

	comp->type->visited=0;
	flatten2(comp->type,prefix,nam);
	comp->type->visited=0;

	free(nam);
	}

	if(array==ahi) break;
	arcnt++;	/* increment the array counter */
	} /* of the array instance loop */
		comp=comp->next;
		}
	}

}


/*
 * Generic hash function for symbol names...this is a ver that uses more
 * buckets since the hierarchy has been flattened and there are many
 * more symbols...
 */
int hash2(char *s)
{
char *p;
unsigned int h=0, g;
for(p=s;*p;p++)
        {
        h=(h<<4)+(*p);
        if((g=h&0xf0000000))
                {
                h=h^(g>>24);
                h=h^g;
                }
        }
return(h%FLATPRIME);
}


/*
 * Simply look for a symbol's presence in the flattened table...
 */
struct symslot *symfind2(char *s)
{
int hv;
struct symslot *temp;

hv=hash2(s);
if(!(temp=flatmodule->symbol[hv])) return(NULL); /* no table made */

while(temp)
        {
        if(!strcmp(temp->name,s))
                {
                return(temp); /* in table already */    
                }
        if(!temp->next) break;
        temp=temp->next;
        }

return(NULL); /* not found */
}


/*
 * Add symbol to flattened module's symbol table...
 */
struct symslot *addflatsym(char *s)
{
int i, hv;
struct symslot *temp;
char *snew;

static char vcd_id[8]={'@',0,0,0,0,0,0,0};
static int vcd_len=1;

hv=hash2(s);

if(!(temp=flatmodule->symbol[hv]))
        {
        flatmodule->symbol[hv]=temp=
                (struct symslot *)calloc(1,sizeof(struct symslot));
#ifdef DEBUG_FLATSYM
        printf("Adding %s to symbol table at $%x.\n",s,temp);
#endif

	strcpy(temp->id,vcd_id);
	for(i=0;i<8;i++)
		{
		if(!vcd_id[i]) break;

		if(vcd_id[i]!='z')
			{
			vcd_id[i]++;
			break;
			}
			else
			{
			vcd_id[i]='@';
			if(!vcd_id[i+1])
				{
				vcd_len++;
				vcd_id[i+1]='@';
				break;
				}
			}
		}

        snew=(char *)malloc(strlen(s)+1);
        strcpy(snew,s);
        temp->name=snew;
	flatmodule->numsymbols++;
        return(temp);   
        }

while(temp)
        {
        if(!strcmp(temp->name,s))
                {
                return(temp); /* in table already */    
                }
        if(!temp->next) break;
        temp=temp->next;
        }

temp->next=(struct symslot *)calloc(1,sizeof(struct symslot));
#ifdef DEBUG_FLATSYM
        printf("Adding %s to symbol table at $%x.\n",s,temp->next);
#endif

	strcpy(temp->next->id,vcd_id);
	for(i=0;i<8;i++)
		{
		if(!vcd_id[i]) break;

		if(vcd_id[i]!='z')
			{
			vcd_id[i]++;
			break;
			}
			else
			{
			vcd_id[i]='@';
			if(!vcd_id[i+1])
				{
				vcd_len++;
				vcd_id[i+1]='@';
				break;
				}
			}
		}

snew=(char *)malloc(strlen(s)+1);
strcpy(snew,s);
temp->next->name=snew;
flatmodule->numsymbols++;
return(temp->next);   
}


/*
 * Dump out the flattened symbol table...
 */
void dumpflatsym(void)
{
struct symslot *temp;
int hv;

if(flatmodule)
	{
	for(hv=0;hv<FLATPRIME;hv++)
		{
		if((temp=flatmodule->symbol[hv]))
			{
			fprintf(ivffile,(!usecompressed_ids)?"FS{":"FIS{");
			while(temp)
				{
				if(!usecompressed_ids)
					{
					fprintf(ivffile,"%s ",temp->name);
					}
					else
					{
					fprintf(ivffile,"%s %s ",temp->id, temp->name);
					}
				temp=temp->next;
				}
			fprintf(ivffile,"}\n");
			}
		}

	}
}

/*
 * Invert the old gateport to the original parsed order and include 
 * in the new gate... 
 */
struct gateport *flattengate(struct gateport *gp, char *prefix, int plen,int gtype)
{
struct gateport *gnew=NULL;
struct gateport *g;
char *str;
int incount=0;
char line[1025];

if(!gp)	/* this should never happen! */
	{
	mesg("\nInternal error: missing gateport in (%s).\n",prefix);
	exit(255);
	}

while(gp)
	{
	g=(struct gateport *)calloc(1,sizeof(struct gateport));
	g->next=gnew;
	gnew=g;	

	str=(char *)malloc(plen+strlen(gp->port->name)); /* plen really +1! */
	strcpy(str,prefix);
	strcat(str,gp->port->name);
	gnew->port=addflatsym(str);
	free(str);

	gp=gp->next;
	}

gp=gnew->next;	
while(gp) { incount++; gp=gp->next; }

if(incount)	/* normal gate */
	{
	if	((mode_pica_vsim_heritage)&&
		((gtype==TRI0)||(gtype==TRI1)||
		(gtype==WIRE)|| (gtype==TRI) ||
		(gtype==WAND)||
		(gtype==WOR) ||
		(gtype==NMOS)|| (gtype==PMOS) ||
		(gtype==BUFIF0)||(gtype==BUFIF1)||
		(gtype==NOTIF0)||(gtype==NOTIF1)||
		(gtype==SUPPLY0)||(gtype==SUPPLY1)))
		{	/* force eval on 1st dl command */
		sprintf(line,"%d $z ",incount+1);
		queuestring(line);
		}
		else
		{
		sprintf(line,"%d ",incount);
		queuestring(line);
		}

	gp=gnew->next;	
	while(gp)
		{
		if(!usecompressed_ids)
			{
			sprintf(line,"%s ",gp->port->name); 
			}
			else
			{
			sprintf(line,"%s ",gp->port->id); 
			}
		queuestring(line);
		gp=gp->next;
		}
	}
	else	/* is a const gate! */
	{
	if(gtype==CONST0) 
		{
		sprintf(line,"1 $0 "); 
		queuestring(line);
		}
	else
	if(gtype==CONST1) 
		{
		sprintf(line,"1 $1 "); 
		queuestring(line);
		}
	else
	if(gtype==CONSTX) 		/* requires additions to vsim! */
		{
		sprintf(line,"1 $x "); 
		queuestring(line);
		}
	else
	if(gtype==CONSTZ) 
		{
		sprintf(line,"1 $z "); 
		queuestring(line);
		}
	else
	if((gtype==TRI0)||(gtype==SUPPLY0))
		{
		sprintf(line,"1 $0 "); 
		queuestring(line);
		}
	else
	if((gtype==TRI1)||(gtype==SUPPLY1))
		{
		sprintf(line,"1 $1 "); 
		queuestring(line);
		}
	else
		{				/* everyone else */
		sprintf(line,"1 $z "); 
		queuestring(line);
		}
		
	}

if(!usecompressed_ids)
	{
	sprintf(line,"1 %s }\n",gnew->port->name);
	}
	else
	{
	sprintf(line,"1 %s }\n",gnew->port->id);
	}
queuestring(line);

return(gnew);
}


/*
 * Queue up the FG definitions...
 */
struct qstr *fgqueue=NULL, *fgprev=NULL;

void queuestring(char *str)
{
char *m;
struct qstr *q;

m=(char *)malloc(strlen(str)+1);
strcpy(m,str);

q=(struct qstr *)calloc(1,sizeof(struct qstr));
q->str=m;

if(!fgqueue)
	{
	fgprev=fgqueue=q;
	}
	else
	{
	fgprev->next=q;
	fgprev=q;
	}
}


/*
 * Empty the queue en masse...
 */
void dequeuestring(void)
{
struct qstr *q, *qold;

q=fgqueue;
while(q)
	{
	fprintf(ivffile,"%s",q->str);
	free(q->str);
	qold=q;

	q=q->next;
	free(qold);
	}
}
