/* 
 * Copyright (c) Tony Bybell 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

/*
 * symbol.c
 * 19/08/97ajb
 */

#include "symbol.h"
#include "y.tab.h"

#include <stdarg.h>             /* variable number of args used in mesg() */

int skipnorefs=0;

struct module *modroot=NULL;
struct module *currentmod=NULL;
int signaltype=0;
int isgate=0;	
int gtype=0;
int assigninprogress=0;
struct gateport *gptr=NULL, *goutport=NULL;

/*
 * handle debugging stuff...
 */
#ifndef DEBUG_PARSE
void DEBUG(char *dummy, ...) { }
#endif

#ifndef DEBUG_SYM
void DEBUG2(char *dummy, ...) { }
#endif


/*
 * Add a component to a module...
 */
struct component *addcomponent(char *name, struct module *type,
				struct metavec *metaport)
{
struct component *cptr, *cptr2;
char *n;

        if(!currentmod->comp)
                {
                cptr=currentmod->comp=(struct component *)
                                calloc(1,sizeof(struct component));
                }
                else
                {
                cptr=(struct component *)calloc(1,sizeof(struct component));
                cptr->next=currentmod->comp;
                currentmod->comp=cptr;
                }

n=(char *)malloc(strlen(name)+1);
strcpy(n,name);
cptr->name=n;
cptr->type=type;
cptr->port=metaport;

#ifdef DEBUG_DUMPGATE
printf("Component %s of type %s added\n",name,type->name);
#endif

cptr2=cptr->next;
while(cptr2)
	{
	if(!strcmp(cptr->name,cptr2->name))
		{
		mesg("** Duplicate component name conflict for '%s' in lines %d and %d.\n",cptr->name,cptr2->linenumber,componentline);		
		warn("   Exiting.\n\n");
		exit(0);
		}
	cptr2=cptr2->next;
	}

return(cptr);
}        



/*
 * Add a register bit to a module...
 */
struct reg *addreg(struct symslot *r)
{
struct reg *rptr;

        if(!currentmod->reg)
                {
                rptr=currentmod->reg=(struct reg *)
                                calloc(1,sizeof(struct reg));
                }
                else
                {
                rptr=(struct reg *)calloc(1,sizeof(struct reg));
                rptr->next=currentmod->reg;
                currentmod->reg=rptr;
                }
rptr->port=r;                
currentmod->numregs++;
r->flags |= SF_REGISTER;

#ifdef DEBUG_DUMPGATE
printf("Reg %s added\n",r->name);
#endif

return(rptr);
}        


/*
 * Add assign'd wire to a module...
 */
struct wire *addwire(struct symslot *w1, struct symslot *w2)
{
struct wire *wptr;

        if(!currentmod->wires)
                {
                wptr=currentmod->wires=(struct wire *)
                                calloc(1,sizeof(struct wire));
                }
                else
                {
                wptr=(struct wire *)calloc(1,sizeof(struct wire));
                wptr->next=currentmod->wires;
                currentmod->wires=wptr;
                }
wptr->port1=w1;                
wptr->port2=w2;

wptr->delayr=gdelayr; /* give wire at end of assign the    */
wptr->delayf=gdelayf; /* whole assign gate's value.        */
wptr->delayz=gdelayz;
currentmod->numwires++;

#ifdef DEBUG_DUMPGATE
printf("Wire Rise (%d), Fall (%d) Z (%d) %s %s\n",wptr->delayr,wptr->delayf,wptr->delayz,w1->name, w2->name); 
#endif

return(wptr);
}        


/*
 * Simply look for a param's presence in the table...
 */
struct param *paramfind(char *s)
{
int hv;
struct param *temp;

hv=hash(s);
if(!(temp=currentmod->params[hv])) return(NULL); /* no table made */

while(temp)
        {
        if(!strcmp(temp->name,s))
                {
                return(temp); /* in table already */    
                }
        if(!temp->next) break;
        temp=temp->next;
        }

return(NULL); /* not found */
}

/*
 * Add a parameter to the table..
 */
struct param *paramadd(char *s)
{
int hv;
struct param *p;

DEBUG2("Adding param: %s\n", s);

hv=hash(s);
p=(struct param *)calloc(1,sizeof(struct param));
p->next=currentmod->params[hv];
strcpy(p->name=(char *)calloc(1+strlen(s), sizeof(char)),s);
currentmod->params[hv]=p;

return(p);
}

/*
 * give cooked param based on substitution
 */
int param_subst(struct param *p)
{
int which = p->which;
struct pnumlist *pl = parameter_module_in_effect->p;
int cnt=0;
while(pl)
	{
	if(pl->valid)
		if(cnt==which)
			{
			return(pl->value);
			}

	cnt++;
	pl=pl->next;
	}

return(p->value);
}



/*
 * Simply look for a symbol's presence in the table...
 */
struct symslot *symfind(char *s)
{
int hv;
struct symslot *temp;

hv=hash(s);
if(!(temp=currentmod->symbol[hv])) return(NULL); /* no table made */

while(temp)
        {
        if(!strcmp(temp->name,s))
                {
                return(temp); /* in table already */    
                }
        if(!temp->next) break;
        temp=temp->next;
        }

return(NULL); /* not found */
}


/*
 * Add symbol to module's symbol table...
 */
struct symslot *addsym2(char *s, int bitnum)
{
int hv, slen;
struct symslot *temp;
char *snew, *stemp;
struct wirechain *w;

if((!ismodulehdr)&&(!iswirestmt))
	{
	slen=strlen(s)+16;	/* enough for string + suffix */
	stemp=(char *)malloc(slen);
	sprintf(stemp,"%s/%d",s, (bitnum!=-1)?bitnum:0);
	}
	else
	{
	slen=strlen(s)+1;	
	stemp=(char *)malloc(slen);
	sprintf(stemp,"%s",s);
	}

hv=hash(stemp);

if(!(temp=currentmod->symbol[hv]))
	{
	DEBUG2("Adding %s to symbol table.\n",stemp);
	if(ismodulehdr) currentmod->numports++;else if (!iswirestmt) currentmod->numsignals++;
	currentmod->symbol[hv]=temp=
		(struct symslot *)calloc(1,sizeof(struct symslot));
	snew=(char *)malloc(strlen(stemp)+1);
	strcpy(snew,stemp);
	temp->name=snew;
	temp->bitnum=bitnum;
	free(stemp);

	if((!ismodulehdr)&&(!isinputstmt))			/* this is to catch */
		{						/* implied wire defs too */
        	w=(struct wirechain *)calloc(1,sizeof(struct wirechain));
        	w->sym=temp;
        	w->next=currentmod->wiresigs;
        	currentmod->wiresigs=w;
		}

	return(temp);	
	}

while(temp)
	{
	if(!strcmp(temp->name,stemp))
		{
		free(stemp);
		return(temp); /* in table already */	
		}
	if(!temp->next) break;
	temp=temp->next;
	}

DEBUG2("Adding %s to symbol table.\n",stemp);
if(ismodulehdr) currentmod->numports++;else if (!iswirestmt) currentmod->numsignals++;
temp->next=(struct symslot *)calloc(1,sizeof(struct symslot));
snew=(char *)malloc(strlen(stemp)+1);
strcpy(snew,stemp);
temp->next->name=snew;
temp->next->bitnum=bitnum;

	if((!ismodulehdr)&&(!isinputstmt))			/* implied wire def */
		{						/* catch as above */
        	w=(struct wirechain *)calloc(1,sizeof(struct wirechain));
        	w->sym=temp->next;
        	w->next=currentmod->wiresigs;
        	currentmod->wiresigs=w;
		}


free(stemp);
return(temp->next);	
}


/*
 * save/retrieve port and gate status for assigns nested inside of gate primitives..
 * the save variables should remain local so don't extern them--there's no need to.
 */
int l_isgate, l_gtype, l_gdelayr, l_gdelayf, l_assigninprogress;
struct gateport *l_gptr, *l_goutport;

void saveports(void)
{
l_isgate=isgate;
l_gtype=gtype;
l_gdelayr=gdelayr;
l_gdelayf=gdelayf;
l_assigninprogress=assigninprogress;
l_gptr=gptr;
l_goutport=goutport;
}

void retrports(void)
{
isgate=l_isgate;
gtype=l_gtype;
gdelayr=l_gdelayr;
gdelayf=l_gdelayf;
assigninprogress=l_assigninprogress;
gptr=l_gptr;
goutport=l_goutport;
}


/*
 * Add port for a gate leg...
 */
void addport(struct symslot *ret)
{
        if(!gptr)
                {
		goutport=gptr=(struct gateport *)
				calloc(1,sizeof(struct gateport));
                gptr->port=ret;
		
                }   
                else
                {
                struct gateport *temp;
                
                temp=(struct gateport *)calloc(1,sizeof(struct gateport));
                temp->next=gptr;
                temp->port=ret;
                gptr=temp;
                }
}


/*
 * Add symbol (via call to ~2) and generate gate ports if necessary...
 */
struct symslot *addsym(char *s, int bitnum)
{
struct symslot *ret;

ret=addsym2(s,bitnum);
if((isgate)||(ismodulehdr))
	{
	addport(ret);
	}

return(ret);
}


/*
 * Add gate to module...note that addgate() calls this function.  this
 * function (addgate2()) is the old addgate() function.  the reason for
 * the split is to allow multiple output not and buf gates..
 */
void addgate2(void)
{
struct gate *ret;

#ifdef DEBUG_DUMPGATE
	char gstr[8];		/* only used in debug... */
	struct gateport *gdeb;	/* ditto */
#endif

ret=(struct gate *)calloc(1,sizeof(struct gate));

if((!assigninprogress)&&(gtype!=CONST0)&&(gtype!=CONST1)&&
		(gtype!=CONSTX)&&(gtype!=CONSTZ))
	{
	ret->delayr=gdelayr;		/* make wire absorb whole delay */
	ret->delayf=gdelayf;		/* in an assign statement.      */
	ret->delayz=gdelayz;
	ret->str0=str0;
	ret->str1=str1;
	}

ret->type=gtype;
ret->port=gptr;
ret->outport=goutport;
if (goutport != 0 && goutport->port != 0)
   /* later we may want to traverse the gate network. For that we need
    * to be able to find which gate drives a signal, so we need a link
    * from the struct symslot to the struct gate. Note that this has
    * nothing to do with vhdl or something ...
    */
   goutport->port->vhdlgate = ret;
goutport=gptr=NULL;

if(!currentmod->gates)
	{
	currentmod->gates=ret;
	}
	else
	{
	ret->next=currentmod->gates;
	currentmod->gates=ret;
	}

#ifdef DEBUG_DUMPGATE
switch(ret->type)
	{
	case BUF:	strcpy(gstr,"buf");  break;
	case BUFIF0:	strcpy(gstr,"bufif0");  break;
	case BUFIF1:	strcpy(gstr,"bufif1");  break;
	case NMOS:	strcpy(gstr,"nmos");  break;
	case PMOS:	strcpy(gstr,"pmos");  break;
	case NOT:	strcpy(gstr,"not");  break;
	case NOTIF0:	strcpy(gstr,"notif0");  break;
	case NOTIF1:	strcpy(gstr,"notif1");  break;
	case AND:	strcpy(gstr,"and");  break;
	case OR:	strcpy(gstr,"or");   break;
	case XOR:	strcpy(gstr,"xor");  break;
	case NAND:	strcpy(gstr,"nand"); break;
	case NOR:	strcpy(gstr,"nor");  break;
	case XNOR:	strcpy(gstr,"xnor"); break;
	case CONST0:	strcpy(gstr,"const0"); break;
	case CONST1:	strcpy(gstr,"const1"); break;
	case CONSTX:	strcpy(gstr,"constx"); break;
	case CONSTZ:	strcpy(gstr,"constz"); break;
	case LATCH:	strcpy(gstr,"latch"); break;
	default:	strcpy(gstr,"OTHER");break;
	}

gdeb=ret->port;
printf("%-6s #(%d,%d,%d) (",gstr,ret->delayr,ret->delayf,ret->delayz);
while(gdeb)
	{
	printf("%s",gdeb->port->name);
	if(!gdeb->next) { printf(")\n"); break; }
		else
		{
		printf(","); gdeb=gdeb->next;
		}
	}
#endif

currentmod->numgates++;
}

void addgate(void)	/* layer that maintains the buf + not gates */
{
if((gtype==NOT)||(gtype==BUF))
	{
	struct gateport *gp, *gold, *inport=NULL;
	
	gold=gp=gptr;
	inport=gold;
	gp=gp->next;

	while(gp)
		{	
		gptr=NULL;
		addport(gp->port);
		addport(inport->port);
		addgate2();
		gp=gp->next;
		}

	while(gold)	/* dealloc gold chain.. */
		{
		gp=gold;
		gold=gold->next;
		free(gp);
		}
	}
	else
	{
	addgate2();
	}
}



/*
 * Create temporary signal for use with assign statements...
 */ 
struct symslot *astemp;
int tempnum=0;

struct symslot *createtemp(void)
{
char astempstr[256];

sprintf(astempstr,"!T%d",tempnum++);
return(astemp=addsym(astempstr,0));
}


/*
 * Generic hash function for symbol names...
 */
int hash(char *s)
{
char *p;
unsigned int h=0, g;
for(p=s;*p;p++)
	{
	h=(h<<4)+(*p);
	if((g=h&0xf0000000))
		{
		h=h^(g>>24);
		h=h^g;
		}
	}
return(h%SYMPRIME);
}


/*
 * Binary search in tree for module name...
 */
struct module *modfind(struct module *modroot, char *str,
                         struct module **previous, int *lastwalk)
{
struct module *node;
int searchdirection;

node=modroot;

while(node)
        {       
        *previous=node;
        searchdirection=strcmp(str,node->name);
        if(!searchdirection) return(node);

        if(searchdirection<0) 
                {
                node=node->left;
                *lastwalk=0;
                }
                else
                {
                node=node->right;       
                *lastwalk=1;
                }
        }
return(NULL);
}

/*
 * search w/o weird extra stuff for modadd
 */
struct module *simple_modfind(struct module *node,char *str)
{
int searchdirection;

while(node)
        {
	searchdirection=strcmp(str,node->name);
	if(!searchdirection) return(node);

	if(searchdirection<0)
		{
		node=node->left;
		}
		else
		{
		node=node->right;
		}
	}

return(NULL);
}

/*
 * Add a node to the tree if not present
 */
struct module *modadd(struct module **modroot,char *str)
{
struct module *insertpoint, *newentry, *foundentry;

int lastwalk,strsize;

if((foundentry=modfind(*modroot,str, &insertpoint, &lastwalk)))
{
#ifdef DEBUG_MOD
printf("%s found!\n",str);
#endif
return(foundentry);
}
else
{
#ifdef DEBUG_MOD
printf("%s added!\n",str);
#endif
}

newentry=(struct module *)calloc(1,sizeof(struct module));

if(!strlen) strsize=1; else strsize=strlen(str)+1;
newentry->name=(char *)malloc(strsize);
strcpy(newentry->name,str);

if(*modroot)
{
if(lastwalk) insertpoint->right=newentry;
else         insertpoint->left =newentry;
}
else
*modroot=newentry;
return(newentry);
}


/*
 * Check I/O ports in input and output statements...
 */
void ioportwatch(char *idptr)
{
struct symslot *t;

if((isinputstmt)||(isoutputstmt))
	{
	if(!(t=symfind(idptr)))
		{
                mesg("** %s not an I/O port in '%s' on line %d.\n",idptr,yyfilename, yylineno);
                }
                else
                {
                t->portdefined++;
                t->lorange=lorange;
                t->hirange=hirange;     
		t->dir=rangedir;

                if(isinputstmt)
                	{
                        t->inaccess=1;
			currentmod->numinputports++;
                        }
                        else
                        {
                        t->outaccess=1;
			currentmod->numoutputports++;
                        }
		if(t->portdefined>1) mesg("** Multiple definition of port %s in '%s' on line %d.\n",idptr,yyfilename,yylineno);
                }
	}
}


/* generate a `not connected' signal.
 */
struct symslot *new_unconnected(void)
{
   struct symslot *ss;
   static int unconnect_seq = 0;
   char s[50];
   sprintf(s,"Internal_Not_Connected__%.4d", unconnect_seq++);
   ss=addsym(s,0);
   ss->flags |= SF_UNCONNECTED;
   return (ss);
}


/*
 * Dump out the module information...
 */
int nummodules, numtoplev, nummissingmods;
int modules_printed;

void moddump2(struct module *node)
{
static char defined[2]={' ','+'};
static char toplevel[2]={' ','T'};

if(!node) return;
if(node->left)  moddump2(node->left);

if((!node->referencecount)&&(node->comps)) numtoplev++;
if(!(node->defined)) nummissingmods++;

if(node->numinputports+node->numoutputports==node->numports)
	{
	int istop = (!node->referencecount)&&(node->comps);

	if(!((!node->referencecount)&&(!istop)&&(skipnorefs)))
		{
		fprintf(stderr,"%c%c%-20s",
			toplevel[istop],
			(node->isexpanded)?'*':defined[(node->defined)!=0],
		        node->name);
		if(strlen(node->name)>20) fprintf(stderr, "\n  \\------------------>");
		fprintf(stderr," %-4d %-4d %-4d  %-5d %-5d %-5d %-5d %-5d %-5d\n",
			node->numinputports, node->numoutputports,node->numports,
			node->referencecount,node->comps,node->numsignals,
			node->numgates, node->numwires, node->numregs);
		modules_printed++;
		}
	}
	else
	{
	modules_printed++;
	fprintf(stderr,"%c%c%-20s",
		toplevel[(!node->referencecount)&&(node->comps)],
		(node->isexpanded)?'*':defined[(node->defined)!=0],
	        node->name);
	if(strlen(node->name)>20) fprintf(stderr, "\n  \\------------------>");
	fprintf(stderr,"[%-4d %-4d %-4d] %-5d %-5d %-5d %-5d %-5d %-5d  PORT WARNING\n",
		node->numinputports, node->numoutputports,node->numports,
		node->referencecount,node->comps,node->numsignals,
		node->numgates, node->numwires, node->numregs);
	}

nummodules++;

if(node->right) moddump2(node->right);
}

void moddump(struct module *root)
{
char plural[2][2]={"s",""};
int nm;

modules_printed=nummodules=numtoplev=nummissingmods=0;

warn("     Module Name        IP   OP   TP   Refs  Comps Sigs  Gates Wires Regs\n"\
     "---------------------- ---- ---- ----  ----- ----- ----- ----- ----- -----\n");
moddump2(root);

nm=skipnorefs ? modules_printed : nummodules;
warn("\nEncountered %d module%s, %d top level, %d missing.\n\n",
     nm, plural[nm==1], numtoplev, nummissingmods);
if(nummissingmods)
	{
	warn("Analysis terminated, missing modules were encountered.\n\n");
	exit(0);
	}
}


/*
 * This group of functions was formerly in ver.y.
 * This is a better place for them (doesn't clutter grammar)
 */

void supplytest(struct symslot *s)
{
int w;

if(s->bitnum==-1) return;	/* keeps port defs from getting attached */

w=s->signaltype;
DEBUGW("Supplytest: %s\n",s->name);

switch(w)
	{
	case SUPPLY0:		/* power supplies */
	case SUPPLY1:
	case TRI:		/* so we can model the capacitance on the wire--force to be a discrete gate */
	case TRI0:
	case TRI1:
	case TRIOR:
	case TRIAND:
		isgate=TRUE;
	        gptr=NULL;      
		gtype=w;
		assigninprogress=0;
		gdelayr=s->delayr; gdelayf=s->delayf; gdelayz=s->delayz;
		addport(s);	/* input becomes $0/$1/$z depending on context */
	        isgate=FALSE;
	       	addgate();
		/* fallthrough */
	default:
		break;
	}
}


void wandfix(struct module *cu)	/* recursively from modroot */
{
if(!cu) return;

if(cu->left) wandfix(cu->left);

{
struct component *comp;
comp=cu->comp;
while(comp)
	{
	sortport(comp);
	comp=comp->next;
	}
}

DEBUGW("Wandfix: %x %x %s\n",modroot,cu,cu->name);
wandfix2(cu); 

if(cu->right) wandfix(cu->right);

}

void wandfix2(struct module *cu)
{
struct wirechain *w;
struct gate *g;
struct wire *wi;
struct component *cp;

currentmod=cu;

DEBUGW("Module: %s\n",cu->name);

cp=cu->comp;
while(cp)
	{
	struct module *m;
	struct gateport *gp;
	struct metavec *v;
	int nports;
	int i, j;

	DEBUGW("\tComponent: %s\n",cp->name);
	if (!(m=cp->type)) { cp=cp->next; continue; }
	if (!(m->numoutputports)) { cp=cp->next; continue; }
	if (!(v=cp->port)) { cp=cp->next; continue; }
	if ((nports=m->numports)!=(v->length)) { cp=cp->next; continue; }

	i=0;
	gp=m->moduleports;

	while(gp)
		{
		if(gp->port->outaccess)
			{
			struct gpchain *gpc;
			struct symslot *s;

			DEBUGW("\t\tOut ports: %s %s...\n",
				gp->port->name, v->vec[i]->bits[0]->name);

			for(j=0;j<v->vec[i]->length;j++)
				{
			        gpc=(struct gpchain *)calloc(1,sizeof(struct gpchain));
        			gpc->fix=&(v->vec[i]->bits[j]);
        
        			s=v->vec[i]->bits[j];  
        			gpc->next=s->wchain;
        			s->wchain=gpc;   
				}            
			}
		i++;
		gp=gp->next;
		}

	cp=cp->next;
	}


g=cu->gates;
while(g)
	{
	struct symslot *s;
	struct gpchain *gpc;

	gpc=(struct gpchain *)calloc(1,sizeof(struct gpchain));
	gpc->fix=&(g->outport->port);

	s=g->outport->port;
	gpc->next=s->wchain;
	s->wchain=gpc;

	g=g->next;
	}

wi=cu->wires;
while(wi)	
	{
	struct symslot *s;
	struct gpchain *gpc;

	gpc=(struct gpchain *)calloc(1,sizeof(struct gpchain));
	gpc->fix=&(wi->port1);

	s=wi->port1;
	gpc->next=s->wchain;
	s->wchain=gpc;

	wi=wi->next;
	}

w=cu->wiresigs;

DEBUGW("W is: %x\n",w);

while(w)
	{
	struct gpchain *gpc;
	int symlen,i,gpcnt=0;
	struct symslot **t;
	char *buf;

	if(!w->sym->signaltype) w->sym->signaltype=WIRE;
	DEBUGW("\t%s (%d) : ",w->sym->name,w->sym->signaltype);

	gpc=w->sym->wchain;

	while(gpc)
		{
		DEBUGW("(%x : %s) ",gpc->fix,(*gpc->fix)->name);
		gpcnt++;

		gpc=gpc->next;
		}
	DEBUGW("\n");

	gpc=w->sym->wchain;
	if(!gpc) { supplytest(w->sym); w=w->next; continue; } /* unreffed out */

	if((!gpc->next)&&(w->sym->delayr==0)&&(w->sym->delayf==0)
		&&(w->sym->delayz==0)&&
		(w->sym->signaltype!=TRI)&&
		(w->sym->signaltype!=TRIOR)&&
		(w->sym->signaltype!=TRIAND)&&
		(w->sym->signaltype!=TRI0)&&(w->sym->signaltype!=TRI1)&&
		(w->sym->signaltype!=SUPPLY0)&&(w->sym->signaltype!=SUPPLY1)) 
			{ w=w->next; continue; } /* wfix not needed */
	DEBUGW("^^ Must fix with %d temps!\n",gpcnt); 

	symlen=strlen(w->sym->name);
	buf=(char *)malloc(symlen+3);		
	strcpy(buf,w->sym->name);
	strcpy(buf+symlen,"!W");

	t=(struct symslot **)calloc(gpcnt,sizeof(struct symslot *));
	for(i=0;i<gpcnt;i++)
		{
		t[i]=addsym(buf,i);
		DEBUGW("\tAdded: %s\n",t[i]->name);
		}

	        isgate=TRUE;
        	gptr=NULL;      
		gdelayr=w->sym->delayr; gdelayf=w->sym->delayf; gdelayz=w->sym->delayz;
		gtype=w->sym->signaltype;
		assigninprogress=0;

		DEBUGW("\tAdded: gate with output %s.\n",w->sym->name);
		addport(w->sym);
	        for(i=0;i<gpcnt;i++)
        	        {
                	addport(t[i]);
                	}

	        isgate=FALSE;
        	addgate();

	i=0;
	while(gpc)
		{
		*(gpc->fix)=t[i++];
		gpc=gpc->next;
		}

	free(buf);
	free(t);

	w=w->next;		
	}
}

void vecheck(char *nam)
{
struct symslot *t;

if((t=symfind(nam)))
	{
	if(t->flags & SF_VECTORED)
		{
		char line[512];

		sprintf(line,"Scalar operation on vector '%s'",nam);
		yyerror(line);
		}
	}
}

struct symslot *wdelay(struct symslot *t)
{
if((signaltype)&&(!(t->signaltype)))
	{	
	t->signaltype=signaltype;
	t->delayr=wdelayr;
	t->delayf=wdelayf;
	t->delayz=wdelayz;
	t->str0=str0;
	t->str1=str1;
	if (wvectored)
           t->flags |= SF_VECTORED;
        else
           t->flags &= ~SF_VECTORED;
	}
return(t);
}

void wirevecadd(char *lhs)
{
struct symslot *t;
int iws;

	iws=iswirestmt;
	iswirestmt=~0;
	t=wdelay(addsym(lhs,-1));
	t->lorange=wlo;
	t->hirange=whi;
	t->dir=wdir;
	iswirestmt=iws;
}

void expandgate(struct gatename *g, struct metavec *m)
{
int i, j, numports, numgates, error_issued = FALSE;

numports=m->length;
if(g->lo<g->hi) { numgates=g->hi-g->lo+1; } else { numgates=g->lo-g->hi+1; }

for(i=0;i<numports;i++)
	{
	if(m->vec[i]->length!=numgates)
		{
		mesg("** Array gate port instance error (port %d) in '%s' on line %d.\n",i+1,yyfilename, yylineno);
		warn("\t Expected bit width of %d, found %d.\n\n"
                     ,numgates,(m->vec[i]->length));
		return;
		}
	}

for(j=0;j<numgates;j++)
	{
	isgate=TRUE;
	gptr=NULL;

	for(i=0;i<numports;i++)
		{
		addport(m->vec[i]->bits[j]);
                if ((i == 0) && !error_issued)
                   { /* check that the output port has not been declared REG */
                   struct reg *r = currentmod->reg;
                   for (; r != 0; r = r->next)
                      if (r->port == m->vec[i]->bits[j])
                         {
                         mesg("** Gate output connects to a register in '%s' on line: %d\n",
                              yyfilename, yylineno);
                         error_issued = TRUE;
                         }
                   }
		}

	isgate=FALSE;
	addgate();
	}

for(i=0;i<numports;i++)
	{
	free(m->vec[i]);
	}
free(m);

}

struct bitvec *unaryreduce(struct bitvec *b, int redtype, char *str)
		{
		int i;
		char err[80];

		if(b->length<2)
			{
			sprintf(err,"Attempted %s unary reduction of a single bit",str);
			yyerror(err);
			}
			else
			{
			saveports();

			DEBUG("Assign (unary): (%s) = (%s %s)...\n",
			createtemp()->name,str,b->bits[0]->name); 
			isgate=1; gptr=NULL;
			gtype=redtype;
			addport(astemp);

			for(i=0;i<b->length;i++)
			{
			addport(b->bits[i]);
			} 

			assigninprogress=~0;
			addgate();
			assigninprogress=0;
			isgate=0; gptr=NULL;

			free(b);

			b=(struct bitvec *)calloc(1,sizeof(struct bitvec));
			b->bits[0]=astemp;
			b->length=1;

			retrports();
			}
		return(b);
		}


void freevex(struct bitvec **b)	/* reclaim memory from dead bitvectors */
{
struct bitvec *p;

while((p=*(b--)))
	{
	free(p);
	}
}


int portmatch(struct bitvec *a, struct bitvec *b)
{
int i;

if(a->length!=b->length)
	{
	mesg("** Bit widths do not match in '%s' on line %d,\n",yyfilename, yylineno);
	warn("\tvector 1 is length %d:",b->length);
		for(i=0;i<b->length;i++)
			{
			warn(" (%s)",b->bits[i]->name);
			}
	warn("\n\tvector 2 is length %d:",a->length);
		for(i=0;i<a->length;i++)
			{
			warn(" (%s)",a->bits[i]->name);
			}
	warn(".\n");
	return(~0);
	}
	else
	{
	return(0);
	}
}

void swaprange(void)
{
int temp;

if(lorange>hirange) 
	{
	temp=hirange;
	hirange=lorange;
	lorange=temp;
	rangedir=-1;	/* means ranges are swapped */
	}
	else
	{
	rangedir=1;	/* means they're not..goes fwd */
	}
}



/*
 * sort ports if they're named...and do sanity checking accordingly..
 */
int sortport(struct component *comp)
{
int porterrors=0;

if(comp->port)
if(comp->port->length)
if(comp->port->vec[0])
if(comp->port->vec[0]->name)
	{
	int i, j;	
	struct gateport *gp;
	struct bitvec *temp;

	if(comp->type->numports!=comp->port->length)
		{
		mesg("** Named Port Count Mismatch!"
                     "\tComponent: (%s), type: (%s), line: (%d).\n",
                     comp->name,comp->type->name,
                     comp->linenumber);
                portmismatches++;
		return(0);
		}

	gp=comp->type->moduleports;
	j=0;
	while(gp)
		{
		for(i=0;i<comp->port->length;i++)
			{
			if(!strcmp(gp->port->name,comp->port->vec[i]->name))
				{	 /* a match! */
				if(i!=j) /* don't swap if same port! */
					{
					temp=comp->port->vec[i];
					comp->port->vec[i]=comp->port->vec[j];
					comp->port->vec[j]=temp;
					}
				break; /* from i loop... */
				}
			}
		j++; gp=gp->next;
		}

	gp=comp->type->moduleports;
	for(i=0;i<comp->port->length;i++)
		{
		if(strcmp(comp->port->vec[i]->name,gp->port->name))
			{
			mesg("** Named Port Mismatch!\n"
                             "\tComponent: (%s), type: (%s), line: (%d).\n"
                             "\tInstance is missing port: (%s).\n",
                             comp->name,comp->type->name,
                             comp->linenumber,gp->port->name);
                        portmismatches++;
			porterrors++;
			}
		gp=gp->next;
		}
	}

return(!porterrors);
}


/*
 * Print a message on stderr and increment the yerror counter.
 */
void mesg(const char *format, ...)
{
   va_list ap;
   va_start(ap, format);
   vfprintf(stderr, format, ap);
   yyerrors += 1;
}

/*
 * Print a message on stderr but do NOT increment yyerrors
 */
void warn(const char *format, ...)
{
   va_list ap;
   va_start(ap, format);
   vfprintf(stderr, format, ap);
}

