/* 
 * Copyright (c) Tony Bybell, Paul Stravers 1999.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

%{

/*
 * ver.y 
 * 19/08/97ajb
 */

#include "alloca.h"
#include "symbol.h"
#include "behave.h"
#define YYMAXDEPTH (1024*1024)

struct linestack
{
struct linestack *next;
int line;
};

struct linestack *linestack=NULL;

void pushline(void)
{
struct linestack *l;

l=(struct linestack *)malloc(sizeof(struct linestack));
l->line=yylineno;
l->next=linestack;

linestack=l;
}

int popline(void)
{
struct linestack *l;
int lineval;

if(linestack)
	{
	lineval=(l=linestack)->line;
	linestack=linestack->next;
	free(l);
	}
	else
	{
	lineval=0;	/* should never happen */
	}

return(lineval);
}


struct symslot *tslot;
struct module  *compptr;

int pline;	/* parameter line # */

int oldlo, oldhi, olddir;
int lorange,hirange, rangedir, lhslo, lhshi, ldir, rhslo, rhshi, rdir;
int expval, gdelayr=0, gdelayf=0, gdelayz=0;
int         wdelayr=0, wdelayf=0, wdelayz=0;
int str0=0, str1=0;	/* strength (left,right) */

int wlo=0, whi=0, wdir=0; /* wire lo, hi, dir */
int wvectored=0;

char comptype[2048+32], compname[2048], modname[2048], lhs[2048], rhs[2048];
char comptype_prefix[2048];

int ismodulehdr=0, isinputstmt=0, isoutputstmt=0, isregstmt=0, iswirestmt=0;
int iscomponent=0;
int componentline=0;
int modtempnum=0;
int arraystart=0, arrayfin=0, isarray=0;
int pcount=0;	/* number of params in current module */

int default_param_width, param_width, param_base;
char *param_str;
struct param_comp_chain *param_comp_chain=NULL;
struct param_comp_chain *parameter_module_in_effect=NULL;

struct bitvec **orbase=NULL, **orstack=NULL;
struct bitvec **norbase=NULL, **norstack=NULL;
struct bitvec **andbase=NULL, **andstack=NULL;
struct bitvec **nandbase=NULL, **nandstack=NULL;
struct bitvec **xorbase=NULL, **xorstack=NULL;
struct bitvec **xnorbase=NULL, **xnorstack=NULL;

%}

%union {
	int i;
	void *pnt;
}

%token MODULE ENDMODULE PARAMETER 
%token INPUT OUTPUT REG LP RP LBRACK RBRACK COMMA 
%token COLON SEMI 
%token <i>   NUMBER 
%token <pnt> IDENTIFIER 
%token ASSIGN EQ AND OR NOT XOR NAND NOR XNOR
%token NMOS PMOS
%token BUF BUFIF0 BUFIF1 NOTIF0 NOTIF1
%token CONST0 CONST1 CONSTX CONSTZ POUND BITNOT BITAND BITOR
%token LATCH LATCH_N LATCH_SR LATCH_SR_N LB RB 
%token <pnt> DBASE BBASE OBASE HBASE
%token <i>   WIRE TRI TRI1 SUPPLY0 WAND TRIAND TRI0 SUPPLY1 WOR TRIOR HIGHZ0 HIGHZ1
%token <i>   STRONG0 STRONG1 PULL0 PULL1 WEAK0 WEAK1
%token LSH RSH ADD SUB MPY DIV MOD AND OR XOR NOT NNOT 
%token LT GT LET GET EQV NEQ AAND OOR
%token BITNAND BITNOR BITXNOR
%token VECTORED SCALARED DOT
%token UNTERMCOMMENT 
%token ALWAYS BEGIN_TOKEN END IF ELSE 
%token <i>   CASE CASEX CASEZ 
%token ENDCASE DEFAULT
%token EVENT POSEDGE NEGEDGE MUX 
%token <pnt> PRAGMA
%token NEGC ADDC SUBC

%type <pnt> pitema concexp asexpression asexpression2 asexp6 asexp6a asexp6b asexp4
%type <pnt> asexp4a asexp4b asexp5 asexp5a asexp05 asexp05a asexp1 asexp1a asexp1p asexp1x
%type <pnt> asexp2 asexp2a asexp2x asexp3 asexp3a asexp20
%type <pnt> unaryreduction simpleid modsignals genidcomp compsignals compsignals1
%type <pnt> component2 assign_sym
%type <pnt> bitlist compsignals2 bits bits2 scalar bitbase scalarorvec
%type <pnt> event event_list clock opt_range reg_lvalue switch_expr lvalue_list
%type <pnt> opt_event statement assign_stmt group_stmt if_stmt case_stmt null_stmt
%type <pnt> value_list selector choice choice_list condition opt_else opt_stmt_list
%type <pnt> gate1x gate2x gate3x gate3plusx gate5x namedgate lvalue_list2
%type <pnt> plist plist2 pnums

%type <i>   opt_number pos_neg_edge case_xz number
%type <i>   pitemb pitemc expression n0 n1 n2 n0a nettype eqv_neq
%type <i>   strength0 strength1 opt_pragma1
%type <i>   val0 val1 val2 val3 val4 val5 val6 val7 val8 val9 val10

%%

world:		program 
		| /* empty */ {mesg("** No modules found\n");} ;

program:	module
		| module program ;

module:		modulehdr ENDMODULE 
		{
		DEBUG("End of Module (%s)\n\n",modname);
		currentmod->numparams=pcount;
		psub=FALSE;
		} ;

modulehdr:	modulehdr2 iowbody | modulehdr2 ;

modulehdr2:	modulehdr2a {
		DEBUG("End of Module Header\n");
		currentmod->moduleports=gptr; gptr=NULL;
		ismodulehdr=0; } ;

modulehdr2a:	modulehdr3 LP modsignals RP semis |
		modulehdr3 LP RP semis |
		modulehdr3 semis ;

param:		parama paramb ;

parama:		PARAMETER opt_param_siz { psub=FALSE; pline=yylineno;} ;
paramb: 	pitems semis { psub=TRUE; } ;

pitem:		pitema pitemb 
		{ 
		char *s;
		int v;

		s=(char *)$1;
		v=$2;

		if(paramfind(s))
			{
			mesg("** '%s' parameter redefinition on line %d.\n",s,pline);
			}
			else
			{
                        if (default_param_width != -1)
                              { /* store value as a DBASE bit string */
                              struct param *param = paramadd(s);
                              char str10[30];
                              sprintf(str10, "%d", v);
                              param->width = default_param_width;
                              param->base  = 10;
                              param->str   = (char*)calloc(1,strlen(str10)+1);
			      param->which = pcount++;
                              strcpy(param->str,str10);
                              }
			      else
			      {
                              /* store value as an EXPRESSION */
		              struct param *param = paramadd(s);
                              param->value=v;
			      param->which=pcount++;
                              }
			}
		}
		| pitema pitemc 
		{ 
		char *s;

		s=(char *)$1;

		if ((default_param_width != -1)
                    && (default_param_width != param_width))
			{
			mesg("** conflicting vector sizes for parameter '%s' in line: %d\n",
                             s, pline);
			}
		if(paramfind(s))
			{
			mesg("** '%s' parameter redefinition on line %d.\n",s,pline);
			}
			else
			{
                        struct param *param = paramadd(s);
                        param->width = param_width;
                        param->base  = param_base;
                        param->str   = param_str;
			param->which = pcount++;
			}
		}
                | pitema pitemd
		{ 
		char *s;

		s=(char *)$1;

		if ((default_param_width != -1)
                    && (default_param_width != param_width))
			{
			mesg("** conflicting vector sizes for parameter '%s' in line: %d\n",
                             s, pline);
			}
		if(paramfind(s))
			{
			mesg("** '%s' parameter redefinition on line %d.\n",s,pline);
			}
			else
			{
                        struct param *param = paramadd(s);
                        param->width = param_width;
                        param->base  = param_base;
                        param->str   = param_str;
			param->which = pcount++;
			}
		}		;

pitema:		IDENTIFIER EQ { psub=TRUE; $$ = $1;} ;

pitemb:		expression { psub=FALSE; } ;

pitemc:		NUMBER BBASE
                {param_width = $1; param_base = 2; param_str = (char*)$2;}
                | NUMBER OBASE
                {param_width = $1; param_base = 8; param_str = (char*)$2;}
                | NUMBER DBASE
                {param_width = $1; param_base = 10; param_str = (char*)$2;}
                | NUMBER HBASE
                {param_width = $1; param_base = 16; param_str = (char*)$2;}
                ;

pitemd:         bits
                {
                   struct bitvec *v = (struct bitvec *)$1;
                   static char str_buf[1001];
                   int i, t;
                   if (v->length > 1000)
                   {
                      mesg("** parameter size truncated to 1000, '%s' line %d\n",
                           yyfilename, yylineno);
                      v->length = 1000;
                   }
                   for (i = 0; i < v->length; i += 1)
                      if ((t = get_driver_type(v->bits[i])) == CONST0)
                         str_buf[i] = '0';
                      else if (t == CONST1)
                         str_buf[i] = '1';
                      else if ((t == CONSTX) || (t == CONSTZ))
                         str_buf[i] = 'x';
                      else
                      {
                         mesg("** parameter value must be constant, '%s' line %d\n",
                              yyfilename, yylineno);
                         str_buf[i] = '0';
                      }
                   str_buf[i] = 0;
                   param_width = v->length;
                   param_base = 2;
                   param_str = str_buf;
                }

pitems:		pitem | pitem COMMA pitems ;

opt_param_siz:  /* empty */ { default_param_width = -1; }
                | LBRACK expression COLON expression RBRACK
                {
                   int w = $2 - $4;
                   default_param_width = (w < 0) ? (1 - w) : (1 + w);
                };


modsignals:	simpleid | simpleid COMMA modsignals ;

modulehdr3:	MODULE IDENTIFIER { strcpy(modname,(char *)($2));
		if(parameter_module_in_effect)
			{
			DEBUG("Module Header (%s)\n",parameter_module_in_effect->fullname);
			(currentmod=modadd(&modroot,parameter_module_in_effect->fullname))->defined++;
			}
			else
			{
			DEBUG("Module Header (%s)\n",modname);
			(currentmod=modadd(&modroot,modname))->defined++;
			}
		ismodulehdr=~0; gptr=NULL; modtempnum=tempnum=0;
		pcount=0;
		};

iowbody:	iowbodya | iowbodya iowbody;
iowbodya:	inputdec semis
                | outputdec semis
                | gates semis
                | param
                | always;

semis:		semi2 | semis semi2;

semi2:		SEMI
		| error
		{
		resynced=1;	/* don't elaborate design after this! */
		}

inputdec:	inputdec1 inputdec2;
inputdec1:	INPUT {lorange=hirange=0; isinputstmt=~0;
				DEBUG("Input Statement\n");}; 
inputdec2:	rangedsignals { isinputstmt=0;
				DEBUG("End of Input Statement\n");};

outputdec:	outputdec1 outputdec2;
outputdec1:	OUTPUT {lorange=hirange=0; isoutputstmt=~0; 
				DEBUG("Output Statement\n");}; 
outputdec2:	rangedsignals { isoutputstmt=0;
				DEBUG("End of Output Statement\n");};


n0:		WIRE | TRI | TRI1 | SUPPLY0 | WAND | TRIAND | TRI0 | 
		SUPPLY1 | WOR | TRIOR ;

nvs:		VECTORED 
		{ wvectored=~0; }
		| SCALARED { wvectored=0; };

n0a:		n0 nvs {$$=$1;} | n0 {wvectored=0;} ;

n1:		n0a drivestrength { $$=$1; } | n0a ;
n2:		n1 rangesig { $$=$1; wlo=lorange; whi=hirange; wdir=rangedir; } | 
		n1 { lorange=hirange=0; }
nettype:	n2 delay { $$=$1; wdelayr=gdelayr; wdelayf=gdelayf; wdelayz=gdelayz; 
		gdelayr=gdelayf=gdelayz=0; /* for assign */ } 
		| n2 ;

wiredec:	wiredec1 wiredec2 
		| wiredec1 wassigns { 
		iswirestmt=signaltype=wdelayr=wdelayf=wdelayz=wlo=whi=wdir=wvectored=0;
                DEBUG("End of Wire Statement\n");};

wiredec1:	nettype { signaltype=$1;
		DEBUG("Wire Statement\n");}; 

wiredec2:	signals2 { iswirestmt=signaltype=wdelayr=wdelayf=wdelayz=0;
		wlo=whi=wdir;
		DEBUG("End of Wire Statement\n");};

wassigns:	wassign0 | wassign0 COMMA wassigns ;
wassign0:	assign1 | assign1n ;

regdec:		regdec1 regdec2;
regdec1:	REG {lorange=hirange=0; isregstmt=~0;
		DEBUG("Reg Statement\n"); }; 

regdec2:	rangedsignals { isregstmt=0;
		DEBUG("End of Reg Statement\n");};


atype:		atype2 {gdelayr=gdelayf=gdelayz=0; } ;
atype2:		ASSIGN ;

assign:		atype assigns | 
		atype delay assigns ;
assigns:	assign0 | assign0 COMMA assigns ;
assign0:	assign1 | assign1a | assign1n | assign1an | 
		assign2conc | assign3 | assign4;

idl:		IDENTIFIER {strcpy(lhs,(char *)($1));
		if(signaltype) wirevecadd(lhs);
		} ;


assign4:	concexp EQ expression
		{
		struct bitvec *lvalue;
		int t, shifted;


		lvalue=(struct bitvec *)$1;
		shifted=expval;
		for(t=(lvalue->length-1);t>=0;t--)
			{
                        isgate=1; gptr=NULL;
                        gtype=(shifted&1)?CONST1:CONST0;
			addport(lvalue->bits[t]);
                        wdelay(lvalue->bits[t]);  
                        addgate();
                        isgate=0; gptr=NULL;
                        shifted>>=1;
			}

		free((void *)lvalue);
		}
		|concexp EQ asexpression2
		{
		struct bitvec *lvalue, *rvalue;

		lvalue=(struct bitvec *)$1;
		rvalue=(struct bitvec *)$3;

		if(lvalue->length==rvalue->length)
			{
			int t;
			for(t=0;t<lvalue->length;t++)
				{
				addwire(lvalue->bits[t],rvalue->bits[t]);
				}
			}
			else
			{
			mesg("** Attempt to assign a vector of length %d to one of length %d,"
                             " '%s' line %d.\n",rvalue->length, lvalue->length,yyfilename, yylineno);
			}

		free((void *)lvalue);
		free((void *)rvalue);
		}


assign1:	idl EQ asexpression2
		{ 
		struct bitvec *b;
		struct symslot *t;
		int i,len,len2;

		b=(struct bitvec *)$3;

		lhslo=lhshi=rhslo=rhshi=0;

		if(!(t=symfind(lhs)))
			{
			if(!signaltype)
				{
				if(b->length!=1)
				{
				mesg("** Attempt to assign a vector of length %d to a bit,"
                                     " '%s' line %d.\n",b->length,yyfilename,yylineno);
				}
				else
				{
				t=wdelay(addsym(lhs,0));
				addwire(t,b->bits[0]);
				DEBUG("Assign: (%s) = (%s)\n",t->name,b->bits[0]->name);
				}
				}
				else

                {
                int ldelta, rdelta, t, i;

                for(i=wlo;i<=whi;i++)wdelay(addsym(lhs,i));
                        
                DEBUG("Wire Assign: (%s[%d:%d]) = {bitvec}\n",
                        lhs,lorange,hirange);
        
                ldelta=whi-wlo+1; rdelta=b->length;
                 
                if(ldelta!=rdelta)
                        {
                        mesg("** Bit count error in assign statement,\n\t'%s' line %d: %s (%d) and a bitvec of width (%d).\n",yyfilename, yylineno, lhs, ldelta, rdelta);
                        }
                        else
                        {
                        if(rangedir==-1)
                                for(t=0;t<ldelta;t++)
                                        {
                                        addwire(addsym(lhs,wlo+t),b->bits[t]);
                                        }
                                else /* reverse bit order of above */
                                for(t=0;t<ldelta;t++)
                                        {
					addwire(addsym(lhs,wlo+t),b->bits[rdelta-t-1]);
                                        }
                        }
                }



			}
			else
			{
			len=t->hirange-t->lorange+1;
			len2=b->length;

			if(len!=len2)
                                {
				if(len!=1)
	                                mesg("** Attempt to assign a vector of length %d to"
                                             "\n\ta vector of length %d in '%s' line %d.\n",len2,len,yyfilename, yylineno);
					else
					mesg("** Attempt to assign a vector of length %d to a bit,"
                                             " '%s' line %d.\n",len2,yyfilename,yylineno);

                                }   
				else
				{
				if(t->dir==1)
				for(i=0;i<len;i++)
					{
					addwire(wdelay(addsym(lhs,i+t->lorange)),b->bits[len-i-1]);
					}
				else
				for(i=0;i<len;i++)
					{
					addwire(wdelay(addsym(lhs,i+t->lorange)),b->bits[i]);
					}
				}
			}
		} ;

assign1n:	idl EQ expression
		{ 
		struct symslot *t;

		t=symfind(lhs);

		if((!t)&&(!signaltype))
			{
			lhslo=lhshi=rhslo=rhshi=0;
			isgate=1; gptr=NULL;
			gtype=(expval!=0)?CONST1:CONST0;
			tslot=wdelay(addsym(lhs,0));
			addgate();
			isgate=0; gptr=NULL;
			DEBUG("Assign: (%s) = (%d)\n",tslot->name,expval); 
			}
		else
			{
			int i, shifted;
			int xwdir, xwlo, xwhi;

			if(signaltype)
				{
				xwlo=wlo;xwhi=whi;xwdir=wdir;
				}
				else
				{
				xwlo=t->lorange;
				xwhi=t->hirange;
				xwdir=t->dir;
				}

			shifted=expval;
			if(xwdir==1) /* forward */
			for(i=xwlo;i<=xwhi;i++)
				{
				isgate=1; gptr=NULL;
				gtype=(shifted&1)?CONST1:CONST0;
				wdelay(addsym(lhs,i));
				addgate();
				isgate=0; gptr=NULL;
				shifted>>=1;
				}
			else	/* backward */
			for(i=xwhi;i>=xwlo;i--)
				{
				isgate=1; gptr=NULL;
				gtype=(shifted&1)?CONST1:CONST0;
				wdelay(addsym(lhs,i));
				addgate();
				isgate=0; gptr=NULL;
				shifted>>=1;
				}
			}
		} ;

assign1a:	idl subscript EQ asexpression
		{ 
		struct symslot *t;
		t=(struct symslot *)$4;

		vecheck(lhs);

		rhslo=rhshi=0; lhslo=lhshi=lorange;
		tslot=addsym(lhs,lhslo);
		DEBUG("Assign: (%s) = (%s)\n", tslot->name, t->name);
		addwire(tslot,t);
		} ;

assign1an:	idl subscript EQ expression 
		{ rhslo=rhshi=0; lhslo=lhshi=lorange;
		vecheck(lhs);
		isgate=1; gptr=NULL;
		gtype=(expval!=0)?CONST1:CONST0;
		tslot=addsym(lhs,lhslo);
		addgate();
		isgate=0; gptr=NULL;
		DEBUG("Assign: (%s) = (%d)\n",tslot->name,expval); } ;


assign3:	idl range EQ expression {
			lhslo=lorange; lhshi=hirange; ldir=rangedir;
			vecheck(lhs);
			{
			int i, shifted;
			shifted=expval;
			if(ldir==1) /* forward */
			for(i=lhslo;i<=lhshi;i++)
				{
				isgate=1; gptr=NULL;
				gtype=(shifted&1)?CONST1:CONST0;
				addsym(lhs,i);
				addgate();
				isgate=0; gptr=NULL;
				shifted>>=1;
				}
			else	/* backward */
			for(i=lhshi;i>=lhslo;i--)
				{
				isgate=1; gptr=NULL;
				gtype=(shifted&1)?CONST1:CONST0;
				addsym(lhs,i);
				addgate();
				isgate=0; gptr=NULL;
				shifted>>=1;
				}
			}
                DEBUG("Assign: (%s[%d:%d]) = %d\n",
			lhs,lhslo,lhshi,expval); } ;

asexpression:	asexpression2
		{
		struct bitvec *b;
		struct symslot *t;

		b=(struct bitvec *)$1;
		t=b->bits[0];

		if(b->length!=1)
			{
			mesg("** Attempt to assign a vector to a single bit in '%s' line %d.\n",yyfilename, yylineno);
			}

		free(b);

		$$=t;
		}

asexpression2:	asexp6 ;

asexp6:         asexp6 MUX asexp6 { pushline(); } COLON asexp6
                {
                   struct bitvec *sel = (struct bitvec *)$1;
                   struct bitvec *tru = (struct bitvec *)$3; /* sel == TRUE */
                   struct bitvec *fal = (struct bitvec *)$6; /* sel == FALSE */
                   struct symslot *sel_sig = sel->bits[0];
		   int lineval=popline();
	
                   if (sel->length > 1)
                      /* must reduce this to a single boolean: */
                      sel_sig = generate_gate(OR, sel, sel->length);
                   if (tru->length != fal->length)
                   {
                      mesg("** Operands for ?: operator must have equal lengths in line: %d\n", lineval);
		      free(sel); free(fal);
                      $$ = $3; /* have to choose something... */
                   }
                   else
                   {
                      struct bitvec *mux_out = new_bitvec(tru->length);
                      int j;
                      for (j = 0; j < tru->length; j += 1)
                         mux_out->bits[j] = generate_gate3(MUX, sel_sig,
                                                           tru->bits[j],
                                                           fal->bits[j]);
		      free(sel); free(tru); free(fal);
                      $$ = mux_out;
                   }
                }
                | asexp6a;


asexp6a:        asexp6a OOR asexp6b
                {
		struct bitvec *lhs = (struct bitvec *)$1;
		struct bitvec *rhs = (struct bitvec *)$3;
                struct symslot *reduced_sig;
		struct bitvec *t;
		int i, j;

		t=new_bitvec(lhs->length + rhs->length);
		for(i=0;i<lhs->length;i++)
			{
			t->bits[i]=lhs->bits[i];
			}
		for(j=0;j<rhs->length;j++,i++)
			{
			t->bits[i]=rhs->bits[j];
			}
		free(lhs); free(rhs);

		reduced_sig=generate_gate(OR, t, t->length);
		free(t);

           	t=(struct bitvec *)calloc(1,sizeof(struct bitvec));
                t->length=1;
                t->bits[0]=reduced_sig;

		$$=t;
		}
                | asexp6b;


asexp6b:        asexp6b AAND asexp4
                {
		struct bitvec *lhs = (struct bitvec *)$1;
		struct bitvec *rhs = (struct bitvec *)$3;
                struct symslot *reduced_sig;
		struct bitvec *t = new_bitvec(2);	/* each leg of the AND */

		if(lhs->length > 1)
			{
			reduced_sig=generate_gate(OR, lhs, lhs->length);
			t->bits[0]=reduced_sig;
			}
			else
			{
			t->bits[0]=lhs->bits[0];
			}
		free(lhs);

		if(rhs->length > 1)
			{
			reduced_sig=generate_gate(OR, rhs, rhs->length);
			t->bits[1]=reduced_sig;
			}
			else
			{
			t->bits[1]=rhs->bits[0];
			}
		free(rhs);

		reduced_sig=generate_gate(AND, t, t->length);
		free(t);

           	t=(struct bitvec *)calloc(1,sizeof(struct bitvec));
                t->length=1;
                t->bits[0]=reduced_sig;

		$$=t;
		}
                | asexp4;


asexp4:		asexp4a { 
		struct bitvec *t, *p;
		struct bitvec **os;
		int i, fail=0;

		saveports();
		
		t=*(orstack--);
		p=*(orstack);
		os=orstack;

		if(!p)
			{
			orstack--;
			$$=t;
			}
			else
			{
			while((p=*(orstack--)))
				{	/* sanity check the ports */
				fail=fail|portmatch(t,p);
				}

			if(!fail)
			for(i=0;i<t->length;i++)
			{
			orstack=os;

			createtemp();

			isgate=1; gptr=NULL;
			gtype=OR;
			addport(astemp);
			addport(t->bits[i]);
			DEBUG("Assign: (%s) = %s OR\n",
				astemp->name,t->bits[i]->name);

			while((p=*(orstack--)))
				{
				addport(p->bits[i]);
				DEBUG("\t%s\n",p->bits[i]->name);
				}
			assigninprogress=~0;
			addgate();
			assigninprogress=0;
			isgate=0; gptr=NULL;
			t->bits[i]=astemp;
			}

			freevex(os);
			$$=t;
			}

		retrports();
		};

asexp4a:	asexp4a BITOR asexp3a { 
			struct bitvec *u;
			u=(struct bitvec *)$3;
			*(++orstack)=u;
			} |
		asexp4b ; 

asexp4b:	asexp3a 
		{	/* start the gate with null prefix.. */
		struct bitvec *t;
		t=(struct bitvec *)$1; 
		*(++orstack)=NULL;
		*(++orstack)=t;
		}
		;

asexp3a:	asexp3 { 
		struct bitvec *t, *p;
		struct bitvec **ns;
		int i, fail=0;

		saveports();
		
		t=*(norstack--);
		p=*(norstack);
		ns=norstack;

		if(!p)
			{
			norstack--;
			$$=t;
			}
			else
			{
                        while((p=*(norstack--)))
                                {       /* sanity check the ports */
                                fail=fail|portmatch(t,p);
                                }

                        if(!fail)
			for(i=0;i<t->length;i++)
			{
			norstack=ns;

			createtemp();

			isgate=1; gptr=NULL;
			gtype=NOR;
			addport(astemp);
			addport(t->bits[i]);
			DEBUG("Assign: (%s) = %s NOR\n",
				astemp->name,t->bits[i]->name);

			while((p=*(norstack--)))
				{
				addport(p->bits[i]);
				DEBUG("\t%s\n",p->bits[i]->name);
				}
			assigninprogress=~0;
			addgate();
			assigninprogress=0;
			isgate=0; gptr=NULL;
			t->bits[i]=astemp;
			}

			freevex(ns);
			$$=t;
			}

		retrports();
		};

asexp3:		asexp3 BITNOR asexp20 { 
			struct bitvec *u;
			u=(struct bitvec *)$3;
			*(++norstack)=u;
			} |
		asexp2a ; 

asexp2a:	asexp20 
		{	/* start the gate with null prefix.. */
		struct bitvec *t;
		t=(struct bitvec *)$1; 
		*(++norstack)=NULL;
		*(++norstack)=t;
		}
		;
	

asexp20:	asexp20x { 
		struct bitvec *t, *p;
		struct bitvec **as;
		int i, fail=0;

		saveports();
		
		t=*(andstack--);
		p=*(andstack);
		as=andstack;

		if(!p)
			{
			andstack--;
			$$=t;
			}
			else
			{
                        while((p=*(andstack--)))
                                {       /* sanity check the ports */
                                fail=fail|portmatch(t,p);
                                }

                        if(!fail)
			for(i=0;i<t->length;i++)
			{
			andstack=as;

			createtemp();

			isgate=1; gptr=NULL;
			gtype=AND;
			addport(astemp);
			addport(t->bits[i]);
			DEBUG("Assign: (%s) = %s AND\n",
				astemp->name,t->bits[i]->name);

			while((p=*(andstack--)))
				{
				addport(p->bits[i]);
				DEBUG("\t%s\n",p->bits[i]->name);
				}
			assigninprogress=~0;
			addgate();
			assigninprogress=0;
			isgate=0; gptr=NULL;
			t->bits[i]=astemp;
			}

			freevex(as);
			$$=t;
			}

		retrports();
		};

asexp20x:	asexp20x BITAND asexp2 { 
			struct bitvec *u;
			u=(struct bitvec *)$3;
			*(++andstack)=u;
			} |
		asexp10a ; 

asexp10a:	asexp2 
		{	/* start the gate with null prefix.. */
		struct bitvec *t;
		t=(struct bitvec *)$1; 
		*(++andstack)=NULL;
		*(++andstack)=t;
		}
		;


asexp2:		asexp2x { 
		struct bitvec *t, *p;
		struct bitvec **ns;
		int i, fail=0;

		saveports();
		
		t=*(nandstack--);
		p=*(nandstack);
		ns=nandstack;

		if(!p)
			{
			nandstack--;
			$$=t;
			}
			else
			{
                        while((p=*(nandstack--)))
                                {       /* sanity check the ports */
                                fail=fail|portmatch(t,p);
                                }

                        if(!fail)
			for(i=0;i<t->length;i++)
			{
			nandstack=ns;

			createtemp();

			isgate=1; gptr=NULL;
			gtype=NAND;
			addport(astemp);
			addport(t->bits[i]);
			DEBUG("Assign: (%s) = %s NAND\n",
				astemp->name,t->bits[i]->name);

			while((p=*(nandstack--)))
				{
				addport(p->bits[i]);
				DEBUG("\t%s\n",p->bits[i]->name);
				}
			assigninprogress=~0;
			addgate();
			assigninprogress=0;
			isgate=0; gptr=NULL;

			t->bits[i]=astemp;
			}

			freevex(ns);
			$$=t;
			}

		retrports();
		};

asexp2x:	asexp2x BITNAND asexp1 { 
			struct bitvec *u;
			u=(struct bitvec *)$3;
			*(++nandstack)=u;
			} |
		asexp1a ; 

asexp1a:	asexp1 
		{	/* start the gate with null prefix.. */
		struct bitvec *t;
		t=(struct bitvec *)$1; 
		*(++nandstack)=NULL;
		*(++nandstack)=t;
		}
		;


asexp1:		asexp1x { 
		struct bitvec *t, *p;
		struct bitvec **xs;
		int i, fail=0;

		saveports();
		
		t=*(xorstack--);
		p=*(xorstack);
		xs=xorstack;

		if(!p)
			{
			xorstack--;
			$$=t;
			}
			else
			{
                        while((p=*(xorstack--)))
                                {       /* sanity check the ports */
                                fail=fail|portmatch(t,p);
                                }

                        if(!fail)
			for(i=0;i<t->length;i++)
			{
			xorstack=xs;

			createtemp();

			isgate=1; gptr=NULL;
			gtype=XOR;
			addport(astemp);
			addport(t->bits[i]);
			DEBUG("Assign: (%s) = %s XOR\n",
				astemp->name,t->bits[i]->name);

			while((p=*(xorstack--)))
				{
				addport(p->bits[i]);
				DEBUG("\t%s\n",p->bits[i]->name);
				}
			assigninprogress=~0;
			addgate();
			assigninprogress=0;
			isgate=0; gptr=NULL;

			t->bits[i]=astemp;
			}

			freevex(xs);
			$$=t;
			}

		retrports();
		};

asexp1x:	asexp1x XOR asexp1p { 
			struct bitvec *u;
			u=(struct bitvec *)$3;
			*(++xorstack)=u;
			} |
		asexp05a ; 

asexp05a:	asexp1p
		{	/* start the gate with null prefix.. */
		struct bitvec *t;
		t=(struct bitvec *)$1; 
		*(++xorstack)=NULL;
		*(++xorstack)=t;
		}
		;


asexp1p:	asexp1xp { 
		struct bitvec *t, *p;
		struct bitvec **xn;
		int i, fail=0;

		saveports();
		
		t=*(xnorstack--);
		p=*(xnorstack);
		xn=xnorstack;

		if(!p)
			{
			xnorstack--;
			$$=t;
			}
			else
			{
                        while((p=*(xnorstack--)))
                                {       /* sanity check the ports */
                                fail=fail|portmatch(t,p);
                                }

                        if(!fail)
			for(i=0;i<t->length;i++)
			{
			xnorstack=xn;

			createtemp();

			isgate=1; gptr=NULL;
			gtype=XNOR;
			addport(astemp);
			addport(t->bits[i]);
			DEBUG("Assign: (%s) = %s XNOR\n",
				astemp->name,t->bits[i]->name);

			while((p=*(xnorstack--)))
				{
				addport(p->bits[i]);
				DEBUG("\t%s\n",p->bits[i]->name);
				}
			assigninprogress=~0;
			addgate();
			assigninprogress=0;
			isgate=0; gptr=NULL;

			t->bits[i]=astemp;
			}
			
			freevex(xn);
			$$=t;
			}

		retrports();
		};

asexp1xp:	asexp1xp BITXNOR asexp5 { 
			struct bitvec *u;
			u=(struct bitvec *)$3;
			*(++xnorstack)=u;
			} |
		asexp05ap ; 

asexp05ap:	asexp5 
		{	/* start the gate with null prefix.. */
		struct bitvec *t;
		t=(struct bitvec *)$1; 
		*(++xnorstack)=NULL;
		*(++xnorstack)=t;
		}
		;

asexp5:         asexp5 eqv_neq asexp5a
                {
                   struct bitvec *b1 = (struct bitvec *)$1;
                   struct bitvec *b2 = (struct bitvec *)$3;
                   if (b1->length != b2->length)
                   {
                      mesg("** Operator %s requires equal operand sizes in '%s' line: %d\n",
                           $2 ? "!=" : "==", yyfilename, yylineno);
                      $$ = $1;  /* need to return something ... */
                   }
                   else
                      $$ = compare_vectors(b1, b2, (int)$2);
                }
                | asexp5 eqv_neq number
                {
                   struct bitvec *b1 = (struct bitvec *)$1;
                   struct bitvec *b2 = number_to_bitvec($<i>3, b1->length);
                   $$ = compare_vectors(b1, b2, (int)$2);
                }
                | asexp5a;


eqv_neq:        EQV   { $$ = 0; }
                | NEQ { $$ = 1; };


asexp5a:	asexp5a ADD asexp05
		{
                struct bitvec *lhs = (struct bitvec *)$1;
                struct bitvec *rhs = (struct bitvec *)$3;
                $$ = generate_adder(lhs, rhs, '+', XOR, AND, XOR, ADDC);
		}
		| asexp5a ADD NUMBER
		{
                struct bitvec *lhs = (struct bitvec *)$1;
                struct bitvec *rhs = number_to_bitvec($3, lhs->length);
                $$ = generate_adder(lhs, rhs, '+', XOR, AND, XOR, ADDC);
		}
		| asexp5a SUB asexp05
		{
                struct bitvec *lhs = (struct bitvec *)$1;
                struct bitvec *rhs = (struct bitvec *)$3;
                $$ = generate_adder(lhs, rhs, '-', XOR, SUBC, XNOR, SUBC);
		}
		| asexp5a SUB NUMBER
		{
                struct bitvec *lhs = (struct bitvec *)$1;
                struct bitvec *rhs = number_to_bitvec($3, lhs->length);
                $$ = generate_adder(lhs, rhs, '-', XOR, SUBC, XNOR, SUBC);
		}
		| asexp05

asexp05:
		LP asexp6 RP { 
			$$=$2;
			} | 
		BITNOT asexp05
			{
			struct bitvec *b;
			int i;

			b=(struct bitvec *)$2;

			saveports();

			for(i=0;i<b->length;i++)
			{
			DEBUG("Assign: (%s) = (NOT %s)\n",
			createtemp()->name,b->bits[i]->name); 

			isgate=1; gptr=NULL;
			gtype=NOT;
			addport(astemp);
			addport(b->bits[i]);
			assigninprogress=~0;
			addgate();
			assigninprogress=0;
			b->bits[i]=astemp;
			isgate=0; gptr=NULL;
			}

			$$=b; 

			retrports();
			} |
		NNOT asexp05
			{
			struct bitvec *b;

			b=(struct bitvec *)$2;

                   	if (b->length > 1)
				{
				struct symslot *reduced_sig;

                      		reduced_sig = generate_gate(NOR, b, b->length);
				free(b);

                		b=(struct bitvec *)calloc(1,sizeof(struct bitvec));
                		b->length=1;
                		b->bits[0]=reduced_sig;

				$$=b;
				}
				else
				{
				saveports();

				DEBUG("Assign: (%s) = (NOT %s)\n",
				createtemp()->name,b->bits[0]->name); 

					isgate=1; gptr=NULL;
				gtype=NOT;
				addport(astemp);
				addport(b->bits[0]);
				assigninprogress=~0;
				addgate();
				assigninprogress=0;
				b->bits[0]=astemp;
				isgate=0; gptr=NULL;

				$$=b; 

				retrports();
				}
			} | unaryreduction | bits;


unaryreduction:	BITOR asexp05
		{
		$$=unaryreduce((struct bitvec *)$2,OR,"OR");
		} 
		| BITAND asexp05
		{
		$$=unaryreduce((struct bitvec *)$2,AND,"AND");
		} 
		| BITNOR asexp05
		{
		$$=unaryreduce((struct bitvec *)$2,NOR,"NOR");
		} 
		| BITNAND asexp05
		{
		$$=unaryreduce((struct bitvec *)$2,NAND,"NAND");
		} 
		| XOR asexp05
		{
		$$=unaryreduce((struct bitvec *)$2,XOR,"XOR");
		} 
		| BITXNOR asexp05
		{
		$$=unaryreduce((struct bitvec *)$2,XNOR,"XNOR");
		} 
		| ADD asexp05
		{
		$$=$2;
		}
		| SUB asexp05
		{
		struct bitvec *neg, *rhs;
		struct symslot *carry;
		int i, length;

		rhs=(struct bitvec *)$2;
		length=rhs->length;

		if(length==1)
			{
			$$=$2;	/* neg of a single bit is that same bit */
			}
			else
			{
			neg=new_bitvec(rhs->length);

			neg->bits[length-1]= rhs->bits[length-1];
			carry=generate_gate1(NOT,rhs->bits[length-1]);

			for(i=length-2;i>=0;i--)
				{
				neg->bits[i]=generate_gate2(XNOR, rhs->bits[i], carry);

				if(i)
					{
					carry=generate_gate2(NEGC,rhs->bits[i], carry);
					}
				}			

			free(rhs);
			$$=neg;
			}
		}
		;

rangedsignals:	rangesig signals | signals ;
signals:	simpleid
			{
			char *idptr;
			int i,t;
			struct symslot *s;

			idptr=(char *)$1;

			ioportwatch(idptr);
			t=iswirestmt; 
			if((isregstmt)||(iswirestmt))
				{
				iswirestmt=~0;
				s=wdelay(addsym(idptr,-1));
				s->lorange=lorange;
				s->hirange=hirange;
				s->dir=rangedir;
				}
			iswirestmt=0;
			for(i=lorange;i<=hirange;i++)
				{			
				s=wdelay(addsym(idptr,i));
				if(isregstmt) addreg(s);
				}
			iswirestmt=t;
			}
		| simpleid 
			{
			char *idptr;
			int i,t;
			struct symslot *s;

			idptr=(char *)$1;

			ioportwatch(idptr);
                        t=iswirestmt;
			if((isregstmt)||(iswirestmt))
				{
				iswirestmt=~0;
				s=wdelay(addsym(idptr,-1));
				s->lorange=lorange;
				s->hirange=hirange;
				s->dir=rangedir;
				}
			iswirestmt=0;
			for(i=lorange;i<=hirange;i++)
				{
				s=wdelay(addsym(idptr,i));
				if(isregstmt) addreg(s);
				}
                        iswirestmt=t;
			}
			COMMA signals ;


signals2:	simpleid
			{
			char *idptr;
			int i,t;
			struct symslot *s;

			idptr=(char *)$1;

			ioportwatch(idptr);

			t=iswirestmt=~0; 

				iswirestmt=~0;
				s=wdelay(addsym(idptr,-1));
				s->lorange=lorange;
				s->hirange=hirange;
				s->dir=rangedir;

			iswirestmt=0;
			for(i=lorange;i<=hirange;i++)
				{			
				s=wdelay(addsym(idptr,i));
				if(isregstmt) addreg(s);
				}
			iswirestmt=t;
			}
		| simpleid 
			{
			char *idptr;
			int i,t;
			struct symslot *s;

			idptr=(char *)$1;

			ioportwatch(idptr);
                        t=iswirestmt=~0;

				iswirestmt=~0;
				s=wdelay(addsym(idptr,-1));
				s->lorange=lorange;
				s->hirange=hirange;
				s->dir=rangedir;

			iswirestmt=0;
			for(i=lorange;i<=hirange;i++)
				{
				s=wdelay(addsym(idptr,i));
				if(isregstmt) addreg(s);
				}
                        iswirestmt=t;
			}
			COMMA signals2 ;


rangesig:	LBRACK expression COLON expression RBRACK
		      	{ lorange=$4; hirange=$2; swaprange();
			  DEBUG("range: [%d:%d]\n",$2,$4); } ;

compsignals:	compsignals1 | compsignals2 ;

compsignals2:	DOT IDENTIFIER LP genidcomp RP
		{
		struct metavec *m;
		m=(struct metavec *)calloc(1,sizeof(struct metavec));

		m->length=1;
		m->vec[0]=(struct bitvec *)$4;
		m->vec[0]->name=(char *)$2;

		$$=m;
		}
		| compsignals2 COMMA DOT IDENTIFIER LP genidcomp RP 
		{
		struct metavec *m, *n;
		int i,len;

		n=(struct metavec *)$1;
		len=n->length;

		m=(struct metavec *)calloc(1,sizeof(struct metavec)+
			(len+1)*sizeof(struct bitvec *));
		m->length=len+1;
		m->vec[0]=(struct bitvec *)$6;
		m->vec[0]->name=(char *)$4;

		for(i=0;i<len;i++)		
			{
			m->vec[i+1]=n->vec[i];		
			}
		free(n);

		$$=m;
		}
		;

compsignals1:	genidcomp
		{
		struct metavec *m;
		m=(struct metavec *)calloc(1,sizeof(struct metavec));

		m->length=1;
		m->vec[0]=(struct bitvec *)$1;

		$$=m;
		}
		| compsignals1 COMMA genidcomp 
		{
		struct metavec *m, *n;
		int i,len;

		n=(struct metavec *)$1;
		len=n->length;

		m=(struct metavec *)calloc(1,sizeof(struct metavec)+
			(len+1)*sizeof(struct bitvec *));
		m->length=len+1;
		m->vec[0]=(struct bitvec *)$3;

		for(i=0;i<len;i++)		
			{
			m->vec[i+1]=n->vec[i];		
			}
		free(n);

		$$=m;
		}
		;

genidcomp:	/* empty */
                {
                   struct bitvec *nc = new_bitvec(1);
                   nc->bits[0] = new_unconnected();
                   $$ = nc;
                }
                | asexpression2 {$$ = $1; };

range:		LBRACK expression COLON expression RBRACK
		      	{ oldlo=lorange; oldhi=hirange; olddir=rangedir;
			  lorange=$4; hirange=$2; swaprange(); } ;
subscript:	LBRACK expression RBRACK 
			{ oldlo=lorange; oldhi=hirange; olddir=rangedir;
			  lorange=hirange=$2; swaprange(); } ;

simpleid:	IDENTIFIER { 
		DEBUG("genid: (%s)\n",(char *)($1)); 
		if((ismodulehdr)||(iswirestmt)) addsym((char *)($1),-1);
		$$=$1;
		} 

gate1x:		asexpression2
		{
		struct metavec *m;

		m=(struct metavec *)calloc(1,sizeof(struct metavec));
		m->length=1;
		m->vec[0]=(struct bitvec *)$1;
		$$=m;
		};

gate2x:		asexpression2 COMMA asexpression2 
		{
		struct metavec *m;

		m=(struct metavec *)calloc(1,sizeof(struct metavec)
				+1*sizeof(struct bitvec *));
		m->length=2;
		m->vec[0]=(struct bitvec *)$1;
		m->vec[1]=(struct bitvec *)$3;
		$$=m;
		};


gate3x:         asexpression2 COMMA asexpression2 COMMA asexpression2
		{
		struct metavec *m;

		m=(struct metavec *)calloc(1,sizeof(struct metavec)
				+2*sizeof(struct bitvec *));
		m->length=3;
		m->vec[0]=(struct bitvec *)$1;
		m->vec[1]=(struct bitvec *)$3;
		m->vec[2]=(struct bitvec *)$5;
		$$=m;
		};

gate5x:         asexpression2 COMMA asexpression2 COMMA asexpression2 COMMA asexpression2 COMMA asexpression2
		{
		struct metavec *m;

		m=(struct metavec *)calloc(1,sizeof(struct metavec)
				+4*sizeof(struct bitvec *));
		m->length=5;
		m->vec[0]=(struct bitvec *)$1;
		m->vec[1]=(struct bitvec *)$3;
		m->vec[2]=(struct bitvec *)$5;
		m->vec[3]=(struct bitvec *)$7;
		m->vec[4]=(struct bitvec *)$9;
		$$=m;
		};

gate3plusx:     gate3x 
		| asexpression2 COMMA gate3plusx 
		{
		struct metavec *m, *m2;
		int i;

		m2=(struct metavec *)$3;

		m=(struct metavec *)calloc(1,sizeof(struct metavec)
				+(m2->length)*sizeof(struct bitvec *));
		m->length=(m2->length)+1;
		m->vec[0]=(struct bitvec *)$1;

		for(i=0;i<m2->length;i++)
			{
			m->vec[i+1]=m2->vec[i];
			}

		free(m2);
		$$=m;
		};


gatesep:	COMMA {DEBUG("Comma Operator\n"); } ;

namedgate:	LP 
		{
		struct gatename *g;
		g=(struct gatename *)calloc(1,sizeof(struct gatename));
		$$=g;
		}
		| IDENTIFIER LP
		{
		struct gatename *g;
		g=(struct gatename *)calloc(1,sizeof(struct gatename));
		g->name=(char *)$1;
		$$=g;
		}
		| IDENTIFIER LBRACK expression RBRACK LP
		{
		struct gatename *g;
		g=(struct gatename *)calloc(1,sizeof(struct gatename));
		g->name=(char *)$1;
		g->lo=g->hi=$3;
		$$=g;
		}
		| IDENTIFIER LBRACK expression COLON expression RBRACK LP
		{
		struct gatename *g;
		g=(struct gatename *)calloc(1,sizeof(struct gatename));
		g->name=(char *)$1;
		g->lo=$3;
		g->hi=$5;
		$$=g;
		};

gate1:		namedgate gate1x RP 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		}
		| namedgate gate1x RP gatesep gate1 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		};

gate2:		namedgate gate2x RP 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		}
		| namedgate gate2x RP gatesep gate2 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		};

gate3:		namedgate gate3x RP 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		}
		| namedgate gate3x RP gatesep gate3 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		};


gate3plus:	namedgate gate3plusx RP 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		}
		| namedgate gate3plusx RP gatesep gate3plus 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		};


gate5:		namedgate gate5x RP 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		}
		| namedgate gate5x RP gatesep gate5 
		{ 
		expandgate((struct gatename *)$1, (struct metavec *)$2);
		};

delay:		POUND expression {DEBUG("delay: (%d)\n",gdelayr=gdelayf=gdelayz=$2);}
		| POUND LP expression COMMA expression RP
			{DEBUG("delay: (%d,%d)\n",gdelayr=$3,gdelayf=gdelayz=$5);} ;
		| POUND LP expression COMMA expression COMMA expression RP
			{ DEBUG("delay: (%d,%d,%d)\n",gdelayr=$3,gdelayf=$5,gdelayz=$7);} ;

const0gate:	const0gate2 {DEBUG("End of Const0 Gate\n"); } ;
const1gate:	const1gate2 {DEBUG("End of Const1 Gate\n"); } ;
constxgate:	constxgate2 {DEBUG("End of ConstX Gate\n"); } ;
constzgate:	constzgate2 {DEBUG("End of ConstZ Gate\n"); } ;

notgate:	notgate2 {DEBUG("End of Not Gate\n"); } ;
bufgate:	bufgate2 {DEBUG("End of Buf Gate\n"); } ;

nmosgate:	nmosgate2 {DEBUG("End of NMOS Gate\n"); } ;
pmosgate:	pmosgate2 {DEBUG("End of PMOS Gate\n"); } ;

bufif0gate:	bufif0gate2 {DEBUG("End of Bufif0 Gate\n"); } ;
bufif1gate:	bufif1gate2 {DEBUG("End of Bufif1 Gate\n"); } ;

notif0gate:	notif0gate2 {DEBUG("End of Notif0 Gate\n"); } ;
notif1gate:	notif1gate2 {DEBUG("End of Notif1 Gate\n"); } ;

xorgate:	xorgate2 {DEBUG("End of Xor Gate\n"); } ;
xnorgate:	xnorgate2 {DEBUG("End of Xnor Gate\n"); } ;
andgate:	andgate2 {DEBUG("End of And Gate\n"); } ;
nandgate:	nandgate2 {DEBUG("End of Nand Gate\n"); } ;
orgate:		orgate2 {DEBUG("End of Or Gate\n"); } ;
norgate:	norgate2 {DEBUG("End of Nor Gate\n"); } ;
latchgate:	latchgate2 {DEBUG("End of Latch Gate\n"); } ;
latchgate_n:	latchgate_n2 {DEBUG("End of Latch_N Gate\n"); } ;
latchgate_sr:	latchgate_sr2 {DEBUG("End of Latch_SR Gate\n"); } ;
latchgate_sr_n:	latchgate_sr_n2 {DEBUG("End of Latch_SR_N Gate\n"); } ;

const0dec:	CONST0 {DEBUG("Const0 Gate\n");   gtype=CONST0; gdelayr=gdelayf=gdelayz=0; } ;
const1dec:	CONST1 {DEBUG("Const1 Gate\n");   gtype=CONST1; gdelayr=gdelayf=gdelayz=0; } ;
constxdec:	CONSTX {DEBUG("ConstX Gate\n");   gtype=CONSTX; gdelayr=gdelayf=gdelayz=0; } ;
constzdec:	CONSTZ {DEBUG("ConstZ Gate\n");   gtype=CONSTZ; gdelayr=gdelayf=gdelayz=0; } ;

notdec:		NOT {DEBUG("Not Gate\n");   gtype=NOT; gdelayr=gdelayf=gdelayz=0; } ;
notif0dec:	NOTIF0 {DEBUG("Notif0 Gate\n");   gtype=NOTIF0;
			gdelayr=gdelayf=gdelayz=0; } ;
notif1dec:	NOTIF1 {DEBUG("Notif1 Gate\n");   gtype=NOTIF1;
			gdelayr=gdelayf=gdelayz=0; } ;

nmosdec:	NMOS {DEBUG("NMOS Gate\n");   gtype=NMOS;
			gdelayr=gdelayf=gdelayz=0; } ;
pmosdec:	PMOS {DEBUG("PMOS Gate\n");   gtype=PMOS;
			gdelayr=gdelayf=gdelayz=0; } ;

bufdec:		BUF {DEBUG("Buf Gate\n");   gtype=BUF;
			gdelayr=gdelayf=gdelayz=0; } ;
bufif0dec:	BUFIF0 {DEBUG("Bufif0 Gate\n");   gtype=BUFIF0;
			gdelayr=gdelayf=gdelayz=0; } ;
bufif1dec:	BUFIF1 {DEBUG("Bufif1 Gate\n");   gtype=BUFIF1;
			gdelayr=gdelayf=gdelayz=0; } ;

xordec:		XOR {DEBUG("Xor Gate\n");   gtype=XOR; gdelayr=gdelayf=gdelayz=0; } ;
xnordec:	XNOR {DEBUG("Xnor Gate\n"); gtype=XNOR; gdelayr=gdelayf=gdelayz=0; } ;
anddec:		AND {DEBUG("And Gate\n");   gtype=AND; gdelayr=gdelayf=gdelayz=0; } ;
nanddec:	NAND {DEBUG("Nand Gate\n"); gtype=NAND; gdelayr=gdelayf=gdelayz=0; } ;
ordec:		OR {DEBUG("Or Gate\n");     gtype=OR; gdelayr=gdelayf=gdelayz=0; } ;
nordec:		NOR {DEBUG("Nor Gate\n");   gtype=NOR; gdelayr=gdelayf=gdelayz=0; } ;
latchdec:	LATCH {DEBUG("Latch Gate\n"); gtype=LATCH; gdelayr=gdelayf=gdelayz=0; } ;
latchdec_n:	LATCH_N {DEBUG("Latch_N Gate\n"); gtype=LATCH_N; gdelayr=gdelayf=gdelayz=0; } ;
latchdec_sr:	LATCH_SR {DEBUG("Latch_SR Gate\n"); gtype=LATCH_SR; gdelayr=gdelayf=gdelayz=0; } ;
latchdec_sr_n:	LATCH_SR_N {DEBUG("Latch_SR_N Gate\n"); gtype=LATCH_SR_N; gdelayr=gdelayf=gdelayz=0; } ;

strength0:	SUPPLY0 | STRONG0 | PULL0 | WEAK0 ;
strength1:	SUPPLY1 | STRONG1 | PULL1 | WEAK1 ;

drivestrength:	LP strength0 COMMA strength1 RP { str0=$2; str1=$4; }
		| LP strength1 COMMA strength0 RP { str0=$2; str1=$4; }
		| LP strength0 COMMA HIGHZ1 RP { str0=$2; str1=HIGHZ1; }
		| LP strength1 COMMA HIGHZ0 RP { str0=$2; str1=HIGHZ0; }
		| LP HIGHZ1 COMMA strength0 RP { str0=HIGHZ1; str1=$4; }
		| LP HIGHZ0 COMMA strength1 RP { str0=HIGHZ0; str1=$4; } ;

drivedelay:	drivestrength delay | delay ;

g1:		drivedelay gate1 | gate1;
g2:		g2a | g3p ;
g2a:		drivedelay gate2 | gate2;
g3:		drivedelay gate3 | gate3;
g3p:		drivedelay gate3plus | gate3plus;
g5:		drivedelay gate5 | gate5;

const0gate2:	const0dec g1;
const1gate2:	const1dec g1;
constxgate2:	constxdec g1;
constzgate2:	constzdec g1;

notgate2:	notdec g2;
notif0gate2:	notif0dec g3;
notif1gate2:	notif1dec g3;

nmosgate2:	nmosdec g3;
pmosgate2:	pmosdec g3;

bufgate2:	bufdec g2;
bufif0gate2:	bufif0dec g3;
bufif1gate2:	bufif1dec g3;

xorgate2:	xordec g3p;
xnorgate2:	xnordec g3p;
andgate2:	anddec g3p;
nandgate2:	nanddec g3p;
orgate2:	ordec g3p;
norgate2:	nordec g3p;
latchgate2:	latchdec g2a | latchdec g3;
latchgate_n2:	latchdec_n g2a | latchdec_n g3;
latchgate_sr2:	latchdec_sr g5;
latchgate_sr_n2:latchdec_sr_n g5;

gates:		latchgate | latchgate_n | latchgate_sr | 
		latchgate_sr_n | assign |
		wiredec | regdec | bufgate | 
		nmosgate | pmosgate |
		bufif0gate | bufif1gate |
		notif0gate | notif1gate |
		notgate | xorgate | xnorgate |
		andgate | orgate | nandgate |
		norgate | const0gate | const1gate |
		constxgate | constzgate | component ;

component:	idid comps ;

comps:		component1 | component1 COMMA comps;

component1:	plist component2 { 
		struct metavec *m;
		struct component *c;
		struct pnumlist *p = $1;
	
		m=(struct metavec *)$2;

		(c=addcomponent(compname,compptr,m))->linenumber=componentline;

		c->arraystart=arraystart;
		c->arrayfin=arrayfin;
		c->isarray=isarray;
		c->pnums = p;

		iscomponent=0; gptr=NULL;
		currentmod->comps++;
		DEBUG("End of Component (%s)\n",compname); } ;

plist:		plist2
		{
		struct pnumlist *p = $1;

		if(p)
			{
			int cnt=0;
			int len;
			struct pnumlist *p2=p;
			struct param_comp_chain *pcc=calloc(1, sizeof(struct param_comp_chain));

			sprintf(comptype, "%s#", comptype_prefix);
			len=strlen(comptype);
			while(p2)
				{
				char buf[128];
				if(cnt++)
					{
					if(p2->valid)
						{
						sprintf(buf,"_%d", p2->value);
						}
						else
						{
						strcpy(buf,"_");
						}
					strcpy(comptype+len, buf);
					len+=strlen(buf);
					}
					else
					{
					if(p2->valid)
						{
						sprintf(buf,"%d", p2->value);
						strcpy(comptype+len, buf);
						len+=strlen(buf);
						}
						else
						{
						/* nothing */
						}
					}
				p2=p2->next;
				}			
			
			pcc->next = param_comp_chain;
			pcc->p = p;
			strcpy(pcc->basename = (char *)malloc(strlen(comptype_prefix)+1), comptype_prefix);
			strcpy(pcc->fullname = (char *)malloc(strlen(comptype)+1), comptype);
			param_comp_chain = pcc;
			}

		strcpy(comptype, comptype_prefix);	/* we use the regular old value for the name to keep from importing multiple times! */
		}

plist2:		POUND LP pnums RP
		{
		$$=$3;
		}
		| /* nothing */
		{
		$$=NULL;
		}

pnums:		expression
		{
		struct pnumlist *p=calloc(1,sizeof(struct pnumlist));
		p->value = $1;
		p->valid = ~0;

		$$=p;
		}
		| expression COMMA pnums
		{
		struct pnumlist *p=calloc(1,sizeof(struct pnumlist));
		p->value = $1;
		p->valid = ~0;
		p->next = $3;

		$$=p;
		}
		| /* nothing */
		{
		struct pnumlist *p=calloc(1,sizeof(struct pnumlist));
		$$=p;
		}
		| COMMA pnums
		{
		struct pnumlist *p=calloc(1,sizeof(struct pnumlist));
		p->next = $2;

		$$=p;
		}

component2:	IDENTIFIER LBRACK expression COLON expression RBRACK LP compsignals RP
		{
		strcpy(compname,(char *)$1);

                DEBUG("Component (%s[%d:%d]) of type (%s)\n",compname,$3,$5,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=$3;
		arrayfin=$5;
		isarray=~0;

                componentline=yylineno;
                iscomponent=~0; gptr=NULL;

		$$=$8;
		}
		| IDENTIFIER LBRACK expression COLON expression RBRACK LP RP
		{
		strcpy(compname,(char *)$1);

                DEBUG("Component (%s[%d:%d]) of type (%s)\n",compname,$3,$5,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=$3;
		arrayfin=$5;
		isarray=~0;

                componentline=yylineno;
                iscomponent=~0; gptr=NULL;

		$$=calloc(1,sizeof(struct metavec));
		}
		| IDENTIFIER LBRACK expression COLON expression RBRACK
		{
		strcpy(compname,(char *)$1);

                DEBUG("Component (%s[%d:%d]) of type (%s)\n",compname,$3,$5,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=$3;
		arrayfin=$5;
		isarray=~0;

                componentline=yylineno;
                iscomponent=~0; gptr=NULL;

		$$=calloc(1,sizeof(struct metavec));
		}
		| IDENTIFIER LBRACK expression RBRACK LP compsignals RP
		{
		strcpy(compname,(char *)$1);
		sprintf(compname+strlen(compname),"[%d]",$3);

                DEBUG("Component (%s) of type (%s)\n",compname,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=arrayfin=isarray=0;

                componentline=yylineno;   
                iscomponent=~0; gptr=NULL;

		$$=$6;		
		}
		| IDENTIFIER LBRACK expression RBRACK LP RP
		{
		strcpy(compname,(char *)$1);
		sprintf(compname+strlen(compname),"[%d]",$3);

                DEBUG("Component (%s) of type (%s)\n",compname,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=arrayfin=isarray=0;

                componentline=yylineno;
                iscomponent=~0; gptr=NULL;

		$$=calloc(1,sizeof(struct metavec));
		}
		| IDENTIFIER LBRACK expression RBRACK
		{
		strcpy(compname,(char *)$1);
		sprintf(compname+strlen(compname),"[%d]",$3);

                DEBUG("Component (%s) of type (%s)\n",compname,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=arrayfin=isarray=0;

                componentline=yylineno;
                iscomponent=~0; gptr=NULL;

		$$=calloc(1,sizeof(struct metavec));
		}
		| IDENTIFIER LP compsignals RP 
		{
		strcpy(compname,(char *)$1);

                DEBUG("Component (%s) of type (%s)\n",compname,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=arrayfin=isarray=0;

                componentline=yylineno;   
                iscomponent=~0; gptr=NULL;

		$$=$3;		
		}
		| IDENTIFIER LP RP 
		{
                strcpy(compname,(char *)$1);

                DEBUG("Component (%s) of type (%s)\n",compname,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=arrayfin=isarray=0;

                componentline=yylineno;
                iscomponent=~0; gptr=NULL;

		$$=calloc(1,sizeof(struct metavec)); 
		} 
		| IDENTIFIER
		{
                strcpy(compname,(char *)$1);

                DEBUG("Component (%s) of type (%s)\n",compname,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=arrayfin=isarray=0;

                componentline=yylineno;
                iscomponent=~0; gptr=NULL;

		$$=calloc(1,sizeof(struct metavec));
		}
		| 
		LP compsignals RP 
		{
                warn("-- WARNING: Missing module name in component of\n\ttype (%s), '%s' line %d.\n\n",comptype, yyfilename, yylineno);
		sprintf(compname,"!M%d",modtempnum++);

                DEBUG("Component (%s) of type (%s)\n",compname,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=arrayfin=isarray=0;

                componentline=yylineno;   
                iscomponent=~0; gptr=NULL;

		$$=$2;		
		}
		| LP RP 
		{
                warn("-- WARNING: Missing module name in component of\n\ttype (%s), '%s' line %d.\n\n",comptype, yyfilename, yylineno);
		sprintf(compname,"!M%d",modtempnum++);

                DEBUG("Component (%s) of type (%s)\n",compname,comptype);
                (compptr=modadd(&modroot,comptype))->referencecount++;

		arraystart=arrayfin=isarray=0;

                componentline=yylineno;
                iscomponent=~0; gptr=NULL;

		$$=calloc(1,sizeof(struct metavec)); 
		} 
		;

idid:		IDENTIFIER { strcpy(comptype_prefix,(char *)($1));} ;


number:         NUMBER ;

val0:           LP val10 RP 
			{
			$$=$2; /* nothing necessary to do! */
			} 
		| SUB val0 
			{
			$$=(-$2);
			} 
		| ADD val0
			{
			$$=($2);
			}
		| NOT val0
			{
			$$=(~$2);
			}
		| NNOT val0
			{
			$$=(!$2);
			}
		| number ;

val1:           val1 MPY val0 
			{
			$$=($1*$3);
			} 
		| val1 DIV val0 
			{
			if($3)
				{
				$$=($1/$3);
				}
				else
				{
				yyerror("Divide by zero detected");
				$$=0;
				}
			} 
		| val1 MOD val0 
			{
			if($3)
				{
				$$=($1%$3);
				}
				else
				{
				yyerror("Divide by zero detected");
				$$=0;
				}
			} 
		| val0 ;

val2:           val2 ADD val1 
			{
			$$=($1+$3);
			} 
		| val2 SUB val1 
			{
			$$=($1-$3);
			} 
		| val1 ;

val3:           val3 LSH val2 
			{
			$$=($1<<$3);
			} 
		| val3 RSH val2 
			{
			$$=($1>>$3);
			} 
		| val2 ;

val4:		val4 LT val3
			{
			$$=($1<$3);
			}
		| val4 GT val3
			{
			$$=($1>$3);
			}
		| val4 LET val3
			{
			$$=($1<=$3);
			}
		| val4 GET val3
			{
			$$=($1>=$3);
			}
		| val3 ;


val5:		val5 EQV val4
			{
			$$=($1==$3);
			}
		| val5 NEQ val4
			{
			$$=($1!=$3);
			}
		| val4 ;


val6:		val6 BITAND val5
			{
			$$=($1&$3);
			} 
		| val5 ;


val7:		val7 XOR val6
			{
			$$=($1^$3);
			} 
		| val6 ;


val8:		val8 BITOR val7
			{
			$$=($1|$3);
			} 
		| val7 ;

val9:		val9 AAND val8
			{
			$$=($1&&$3);
			} 
		| val8 ;

val10:		val10 OOR val9
			{
			$$=($1||$3);
			} 
		| val9 ;

expression:     val10 
		{
		expval = $1;
		} ;


assign2conc:	idl range EQ asexpression2
                {
		int ldelta, rdelta, t, i;
		struct bitvec *b;
		b=(struct bitvec *)$4;

		vecheck(lhs);
		for(i=lorange;i<=hirange;i++)addsym(lhs,i);

		DEBUG("Assign: (%s[%d:%d]) = {bitvec}\n",
			lhs,lorange,hirange);

		ldelta=hirange-lorange+1; rdelta=b->length;

		if(ldelta!=rdelta)
			{
			mesg("** Bit count error in assign statement,\n\t'%s' line %d: %s (%d) and a bitvec of width (%d).\n",yyfilename, yylineno, lhs, ldelta, rdelta);
			}
			else
			{
			if(rangedir==-1)
				for(t=0;t<ldelta;t++)
					{
					addwire(addsym(lhs,lorange+t),b->bits[t]);
					}
				else /* reverse bit order of above */
				for(t=0;t<ldelta;t++)
					{
					addwire(addsym(lhs,lorange+t),b->bits[rdelta-t-1]);
					}
			}
		free(b);
		};


concexp:	LB bitlist RB 
		{
		$$=$2;
		}
		| LB expression concexp RB 
		{
		int cnt=0;
		int dupcnt,oldwidth,width,i,j;
		struct bitvec *b, *dup;
		struct symslot *t;


		dupcnt=$2;
		if(dupcnt<=0)
                        {
                        mesg("** Zero or negative duplication count found in '%s' on line: %d.\n",yyfilename, yylineno);
                        dupcnt=1;  /* to allow further compiling--will throw error at end */
                        }  

		dup=(struct bitvec *)$3;
		oldwidth=dup->length;
		width=oldwidth*dupcnt;

		b=(struct bitvec *)calloc(1,sizeof(struct bitvec)+
				(width*sizeof(struct symslot *)));
		b->length=width;

		for(i=0;i<dupcnt;i++)
		for(j=0;j<oldwidth;j++)
			{
			b->bits[cnt++]=dup->bits[j];
			}

		for(i=0;i<width;i++)
			{
			t=b->bits[i];
			DEBUG("Dup %d: %s\n",i,t->name);
			}
		DEBUG("----------\n");


		free(dup);
		$$=b;
		}
		;

bits2:		asexpression2 ;

bitlist:	bits2 
		| bits2 COMMA bitlist
		{
		struct bitvec *lft, *rgh, *b;
		struct symslot *t;
		int i,width,cnt=0;

		lft=(struct bitvec *)$1;
		rgh=(struct bitvec *)$3;
		
		width=(lft->length+rgh->length);

		b=(struct bitvec *)calloc(1,sizeof(struct bitvec)+
			(width*sizeof(struct symslot *)));
		b->length=width;

		for(i=0;i<lft->length;i++)
			{
			b->bits[cnt++]=lft->bits[i];
			}		

		for(i=0;i<rgh->length;i++)
			{
			b->bits[cnt++]=rgh->bits[i];
			}		

                for(i=0;i<width;i++)
                        {
                        t=b->bits[i];
                        DEBUG("Comma %d: %s\n",i,t->name);
                        }       
                DEBUG("----------\n");

		free(lft); free(rgh);
		$$=b;
		}
		;

scalarorvec:	IDENTIFIER
		{
		struct bitvec *b;
		struct symslot *t;
		int i, len;

		/* add sym[0] if not in, out, or wire */
		if(!(t=symfind((char *)($1))))
			{ 
			t=addsym2((char *)($1),0);

	                b=(struct bitvec *)calloc(1,sizeof(struct bitvec));
        	        b->length=1;
               	 	b->bits[0]=t;

	                DEBUG("Bit: %s\n----------\n",t->name);

                	$$=b;
			}
			else
			{
			len=t->hirange-t->lorange+1;
	                b=(struct bitvec *)calloc(1,sizeof(struct bitvec)+
				(sizeof(struct symslot *)*len));

			b->length=len;
			if(t->dir==1)
				for(i=0;i<len;i++)
				{	/* take bitvec from hi to lo order! */
				b->bits[len-i-1]=addsym2((char *)($1),t->lorange+i);
				}
				else
				for(i=0;i<len;i++)
				{	/* take bitvec from lo to hi order! */
				b->bits[i]=addsym2((char *)($1),t->lorange+i);
				}
			$$=b;
			}
		}

scalar:         IDENTIFIER LBRACK expression RBRACK {
			vecheck((char *)($1));
                        $$=addsym2((char *)($1),$3); } ;

bits:		scalar
                {
                struct bitvec *b;
                struct symslot *t;

                t=(struct symslot *)$1;

                b=(struct bitvec *)calloc(1,sizeof(struct bitvec));
                b->length=1;
                b->bits[0]=t;

                DEBUG("Bit: %s\n----------\n",t->name);

                $$=b;
                }
		| scalarorvec
		| IDENTIFIER LBRACK expression COLON expression RBRACK
		{
		int i, width;
		struct bitvec *b;
		struct symslot *t;
		int cnt=0;
		int lo, hi, dir;

		vecheck((char *)($1));

		lo=$3;
		hi=$5;

		if(lo<hi) 
			{ 
			width=hi-lo+1;
			dir=1;
 			} 
			else 
			{ 
			width=lo-hi+1;
			dir=-1;
			}

		DEBUG("lo:%d hi:%d width:%d dir:%d\n",lo,hi,width,dir); 

		b=(struct bitvec *)calloc(1,sizeof(struct bitvec)+
				(width*sizeof(struct symslot *)));
		b->length=width;

		if(dir==-1)
			{
			for(i=lo;i>=hi;i--)
				{
				b->bits[cnt++]=addsym((char *)($1),i);
				}
			}
			else
			{
			for(i=lo;i<=hi;i++)
				{
				b->bits[cnt++]=addsym((char *)($1),i);
				}
			}		

		for(i=0;i<width;i++)
			{
			t=b->bits[i];
			DEBUG("Bit %d: %s\n",i,t->name);
			}
		DEBUG("----------\n");

		$$=b;
		}
		| concexp | bitbase ;

bitbase:	opt_number BBASE 
		{
		int width, vlen, i;
		char *pnt, *pnt2, *bin, *bin2;
		struct symslot *t;
		struct bitvec *b;

		width=$1;
		pnt=(char *)$2;

                if(param_width == -1)
                {  /* BBASE does not result from parameter substitution */
                   width = $1;
                   if (width == -1)
                      /* no width specified. Verilog Manual says to
                         take machine-dependent width. We take 32. */
                      width = 32;
                }
                else
                { /* BBASE comes from parameter substitution */
                   width = param_width;
                   param_width = -1;
                }
		if(!width)
			{
			mesg("** Zero length binary string found in '%s' on line: %d\n"
                             "\tValue of string is '%s'.\n",yyfilename, yylineno,pnt);

			width=1;  /* to allow further compiling--will throw error at end */
			}

		vlen=strlen(pnt);	/* length of user supplied 'b vector */
		bin=(char *)calloc(1,width+1); /* alloc size we need + null terminator */

		for(i=0;i<width;i++)	/* empty "0" string */
			{
			bin[i]='0';
			}
		bin2=bin+width-1;	/* point to end of dest string */
		pnt2=pnt+vlen-1;	/* point to end of the source string */		

		for(;;)
			{
			*(bin2--)=*(pnt2--); /* copy in reverse order.. */
			if((bin2<bin)||(pnt2<pnt)) break;  /* ran out of src/dest buff */
			}

		DEBUG("BBASE String is: *%s*\n",bin);

		b=(struct bitvec *)calloc(1,sizeof(struct bitvec)+(width*sizeof(struct symslot *)));
		b->length=width;

		saveports();
		for(i=0;i<width;i++)
			{
			isgate=1; gptr=NULL;

			switch(bin[i])
				{
				case '0': gtype=CONST0; break;
				case '1': gtype=CONST1; break;
				case '?':
				case 'X':
				case 'x': gtype=CONSTX; break;
				case 'Z':
				case 'z': gtype=CONSTZ; break;
				}

			createtemp();
			t=astemp;			
			b->bits[i]=t;
			addgate();			

			isgate=0; gptr=NULL;
			}
		retrports();

		free(bin);
		$$=b;
		}
		| opt_number OBASE
		{
		int width, reallen, vlen, i;
		char *realpnt, *pnt, *pnt2, *bin, *bin2, *temp;
		struct symslot *t;
		struct bitvec *b;

		width=$1;
		realpnt=(char *)$2;

                if(param_width == -1)
                {  /* OBASE does not result from parameter substitution */
                   width = $1;
                   if (width == -1)
                      /* no width specified. Verilog Manual says to
                         take machine-dependent width. We take 32. */
                      width = 32;
                }
                else
                { /* OBASE comes from parameter substitution */
                   width = param_width;
                   param_width = -1;
                }
		if(!width)
			{
			mesg("** Zero length octal string found in '%s' on line: %d\n"
				"\tValue of string is '%s'.\n",yyfilename, yylineno,realpnt);
			width=1;  /* to allow further compiling--will throw error at end */
			}

		reallen=strlen(realpnt);
		vlen=reallen*3;	/* length of user supplied 'o vector conv to binary*/
		temp=pnt=(char *)calloc(1,vlen+1); /* convert buffer.. */

		for(i=0;i<reallen;i++)
			{
			switch(realpnt[i])
				{
				case '0': *(temp++)='0'; *(temp++)='0'; *(temp++)='0'; break;
				case '1': *(temp++)='0'; *(temp++)='0'; *(temp++)='1'; break;
				case '2': *(temp++)='0'; *(temp++)='1'; *(temp++)='0'; break;
				case '3': *(temp++)='0'; *(temp++)='1'; *(temp++)='1'; break;
				case '4': *(temp++)='1'; *(temp++)='0'; *(temp++)='0'; break;
				case '5': *(temp++)='1'; *(temp++)='0'; *(temp++)='1'; break;
				case '6': *(temp++)='1'; *(temp++)='1'; *(temp++)='0'; break;
				case '7': *(temp++)='1'; *(temp++)='1'; *(temp++)='1'; break;
				case 'X':
				case 'x': *(temp++)='x'; *(temp++)='x'; *(temp++)='x'; break;
				case 'Z':
				case 'z': *(temp++)='z'; *(temp++)='z'; *(temp++)='z'; break;
				}
			}

		bin=(char *)calloc(1,width+1); /* alloc size we need + null terminator */

		for(i=0;i<width;i++)	/* empty "0" string */
			{
			bin[i]='0';
			}
		bin2=bin+width-1;	/* point to end of dest string */
		pnt2=pnt+vlen-1;	/* point to end of the source string */		

		for(;;)
			{
			*(bin2--)=*(pnt2--); /* copy in reverse order.. */
			if((bin2<bin)||(pnt2<pnt)) break;  /* ran out of src/dest buff */
			}

		DEBUG("OBASE String is: *%s*\n",bin);

		b=(struct bitvec *)calloc(1,sizeof(struct bitvec)+(width*sizeof(struct symslot *)));
		b->length=width;

		saveports();
		for(i=0;i<width;i++)
			{
			isgate=1; gptr=NULL;

			switch(bin[i])
				{
				case '0': gtype=CONST0; break;
				case '1': gtype=CONST1; break;
				case 'X':
				case 'x': gtype=CONSTX; break;
				case 'Z':
				case 'z': gtype=CONSTZ; break;
				}

			createtemp();
			t=astemp;			
			b->bits[i]=t;
			addgate();			

			isgate=0; gptr=NULL;
			}
		retrports();

		free(pnt);
		free(bin);
		$$=b;
		}
		| opt_number HBASE
		{
		int width, reallen, vlen, i;
		char *realpnt, *pnt, *pnt2, *bin, *bin2, *temp;
		struct symslot *t;
		struct bitvec *b;

		width=$1;
		realpnt=(char *)$2;

                if(param_width == -1)
                {  /* HBASE does not result from parameter substitution */
                   width = $1;
                   if (width == -1)
                      /* no width specified. Verilog Manual says to
                         take machine-dependent width. We take 32. */
                      width = 32;
                }
                else
                { /* HBASE comes from parameter substitution */
                   width = param_width;
                   param_width = -1;
                }
		if(!width)
			{
			mesg("** Zero length hexadecimal string found in '%s' on line: %d\n"
                             "\tValue of string is '%s'.\n",yyfilename, yylineno,realpnt);
			width=1;  /* to allow further compiling--will throw error at end */
			}

		reallen=strlen(realpnt);
		vlen=reallen*4;	/* length of user supplied 'h vector conv to binary*/
		temp=pnt=(char *)calloc(1,vlen+1); /* convert buffer.. */

		for(i=0;i<reallen;i++)
			{
			switch(realpnt[i])
				{
				case '0': *(temp++)='0'; *(temp++)='0'; *(temp++)='0'; *(temp++)='0'; break;
				case '1': *(temp++)='0'; *(temp++)='0'; *(temp++)='0'; *(temp++)='1'; break;
				case '2': *(temp++)='0'; *(temp++)='0'; *(temp++)='1'; *(temp++)='0'; break;
				case '3': *(temp++)='0'; *(temp++)='0'; *(temp++)='1'; *(temp++)='1'; break;
				case '4': *(temp++)='0'; *(temp++)='1'; *(temp++)='0'; *(temp++)='0'; break;
				case '5': *(temp++)='0'; *(temp++)='1'; *(temp++)='0'; *(temp++)='1'; break;
				case '6': *(temp++)='0'; *(temp++)='1'; *(temp++)='1'; *(temp++)='0'; break;
				case '7': *(temp++)='0'; *(temp++)='1'; *(temp++)='1'; *(temp++)='1'; break;
				case '8': *(temp++)='1'; *(temp++)='0'; *(temp++)='0'; *(temp++)='0'; break;
				case '9': *(temp++)='1'; *(temp++)='0'; *(temp++)='0'; *(temp++)='1'; break;
				case 'A':
				case 'a': *(temp++)='1'; *(temp++)='0'; *(temp++)='1'; *(temp++)='0'; break;
				case 'B':
				case 'b': *(temp++)='1'; *(temp++)='0'; *(temp++)='1'; *(temp++)='1'; break;
				case 'C':
				case 'c': *(temp++)='1'; *(temp++)='1'; *(temp++)='0'; *(temp++)='0'; break;
				case 'D':
				case 'd': *(temp++)='1'; *(temp++)='1'; *(temp++)='0'; *(temp++)='1'; break;
				case 'E':
				case 'e': *(temp++)='1'; *(temp++)='1'; *(temp++)='1'; *(temp++)='0'; break;
				case 'F':
				case 'f': *(temp++)='1'; *(temp++)='1'; *(temp++)='1'; *(temp++)='1'; break;
				case 'X':
				case 'x': *(temp++)='x'; *(temp++)='x'; *(temp++)='x'; *(temp++)='x'; break;
				case 'Z':
				case 'z': *(temp++)='z'; *(temp++)='z'; *(temp++)='z'; *(temp++)='z'; break;
				}
			}

		bin=(char *)calloc(1,width+1); /* alloc size we need + null terminator */

		for(i=0;i<width;i++)	/* empty "0" string */
			{
			bin[i]='0';
			}
		bin2=bin+width-1;	/* point to end of dest string */
		pnt2=pnt+vlen-1;	/* point to end of the source string */		

		for(;;)
			{
			*(bin2--)=*(pnt2--); /* copy in reverse order.. */
			if((bin2<bin)||(pnt2<pnt)) break;  /* ran out of src/dest buff */
			}

		DEBUG("HBASE String is: *%s*\n",bin);

		b=(struct bitvec *)calloc(1,sizeof(struct bitvec)+(width*sizeof(struct symslot *)));
		b->length=width;

		saveports();
		for(i=0;i<width;i++)
			{
			isgate=1; gptr=NULL;

			switch(bin[i])
				{
				case '0': gtype=CONST0; break;
				case '1': gtype=CONST1; break;
				case 'X':
				case 'x': gtype=CONSTX; break;
				case 'Z':
				case 'z': gtype=CONSTZ; break;
				}

			createtemp();
			t=astemp;			
			b->bits[i]=t;
			addgate();			

			isgate=0; gptr=NULL;
			}
		retrports();

		free(pnt);
		free(bin);
		$$=b;
		}
		| opt_number DBASE
		{
		int width, vlen, i, decvalue;
		char *pnt, *pnt2, *bin, *bin2, *temp;
		struct symslot *t;
		struct bitvec *b;

		width=$1;
		pnt=(char *)$2;

                if(param_width == -1)
                {  /* DBASE does not result from parameter substitution */
                   width = $1;
                   if (width == -1)
                      /* no width specified. Verilog Manual says to
                         take machine-dependent width. We take 32. */
                      width = 32;
                }
                else
                { /* DBASE comes from parameter substitution */
                   width = param_width;
                   param_width = -1;
                }
		if(!width)
			{
			mesg("** Zero length decimal string found in '%s' on line: %d\n"
                             "\tValue of string is '%s'.\n",yyfilename, yylineno,pnt);
			width=1;  /* to allow further compiling--will throw error at end */
			}

		vlen=32;	/* make all decimal numbers 32 bits.. */
		temp=(char *)calloc(1,vlen+1); 
		decvalue=atoi(pnt);
		for(i=0;i<32;i++)	/* construct right justified decimal number.. */
			{
			temp[i]='0'+((decvalue>>(31-i))&1);
			}
		pnt=temp;	/* blast on top of old dec num with new binary one.. */

		bin=(char *)calloc(1,width+1); /* alloc size we need + null terminator */

		for(i=0;i<width;i++)	/* empty "0" string */
			{
			bin[i]='0';
			}
		bin2=bin+width-1;	/* point to end of dest string */
		pnt2=pnt+vlen-1;	/* point to end of the source string */		

		for(;;)
			{
			*(bin2--)=*(pnt2--); /* copy in reverse order.. */
			if((bin2<bin)||(pnt2<pnt)) break;  /* ran out of src/dest buff */
			}

		DEBUG("DBASE String is: *%s*\n",bin);

		b=(struct bitvec *)calloc(1,sizeof(struct bitvec)+(width*sizeof(struct symslot *)));
		b->length=width;

		saveports();
		for(i=0;i<width;i++)
			{
			isgate=1; gptr=NULL;

			switch(bin[i])
				{
				case '0': gtype=CONST0; break;
				case '1': gtype=CONST1; break;
				case 'X':
				case 'x': gtype=CONSTX; break;
				case 'Z':
				case 'z': gtype=CONSTZ; break;
				}

			createtemp();
			t=astemp;			
			b->bits[i]=t;
			addgate();			

			isgate=0; gptr=NULL;
			}
		retrports();

		free(bin);
		$$=b;
		}
		;

opt_number:     /* empty */ { $$ = -1; }
                | NUMBER { $$ = $1; };

always:         ALWAYS opt_event { pushline(); } statement
                {
                   struct symlist *evt_list = (struct symlist *)$2;
                   struct stmt *stmt = (struct stmt *)$4;
		   int evt_lineno = popline();

                   elaborate_always(stmt, evt_list, evt_lineno);
                };

opt_event:      /* empty */               { $$ = 0; }
                | EVENT event             { $$ = $2; }
                | EVENT LP event_list RP  { $$ = $3; };

event_list:     event { $$ = $1; }
                | event OR event_list
                {
                   struct symlist *e1 = (struct symlist *)$1;
                   struct symlist *e2 = (struct symlist *)$3;
                   if (e1 == 0)
                      $$ = $3;
                   else if ((e2 != 0)
                            && ((e1->flags != 0) ^ (e2->flags != 0)))
                   {
                      mesg("** Cannot have a clock and normal signals in the same event list in '%s' on line %d\n",
                           yyfilename, yylineno);
                      $$ = 0;
                   }
                   else
                   {
                      e1->next = e2;
                      $$ = e1;
                   }
                };

event:          IDENTIFIER
                {
                   struct symlist *sym = calloc(1, sizeof(struct symlist));
                   char *s = (char*)$1;
                   sym->symbol = symfind(s);
                   $$ = sym;
                   if (sym->symbol == 0)
                   {
                      char *bf = ver_alloca(10+strlen(s));
                      sprintf(bf,"%s/0",s);
                      sym->symbol = symfind(bf);
                      if (sym->symbol == 0)
                      {
                         warn("-- WARNING: Unknown identifier %s in event list at '%s' line: %d\n",
                              (char *)$1, yyfilename, yylineno);
                         $$ = 0;
                      }
                      else
                      {
                         iswirestmt = 1;
                         sym->symbol = addsym2(s,-1);
                         iswirestmt = 0;
                      }
                   }
                };
                | pos_neg_edge clock
                {
                   struct symlist *clk = (struct symlist *)$2;
                   if (clk)
                   {
                      if ($1==1)
                         /* clock or asynchronous set/reset (negedge) */
                         clk->flags = SYML_NEGEDGE;
                      else
                         /* clock or asynchronous set/reset (posedge) */
                         clk->flags = SYML_POSEDGE;
                   }
                   $$ = clk;
                }

clock:          IDENTIFIER
                {
                   int idlen;
		   char *tbuff;
                   struct symlist *sym = calloc(1, sizeof(struct symlist));

		   idlen=strlen((char *)$1);
		   tbuff=(char *)ver_alloca(idlen+3);
		   strcpy(tbuff,(char *)$1);
		   strcpy(tbuff+idlen,"/0");	/* construct bitref part of signame */

                   sym->symbol = symfind(tbuff);
                   if (sym->symbol == 0)
                   {
                      warn("-- WARNING: Unknown identifier %s in event list at '%s' line: %d\n",
                           (char *)$1, yyfilename, yylineno);
                      $$ = 0;
                   }
                   else
                      $$ = sym;
                }
		|
                IDENTIFIER LBRACK expression RBRACK
                {
                   int idlen;
		   char *tbuff;
                   struct symlist *sym = calloc(1, sizeof(struct symlist));

		   idlen=strlen((char *)$1);
		   tbuff=(char *)ver_alloca(idlen+20);	/* more than enough */
		   strcpy(tbuff,(char *)$1);
		   sprintf(tbuff+idlen,"/%d",$3);

                   sym->symbol = symfind(tbuff);
                   if (sym->symbol == 0)
                   {
                      warn("-- WARNING: Unknown identifier %s in event list at '%s' line: %d\n",
                           (char *)$1, yyfilename, yylineno);
                      $$ = 0;
                   }
                   else
                      $$ = sym;
                }


pos_neg_edge:   POSEDGE   { $$ = 0; }
                | NEGEDGE { $$ = 1; };

statement:      assign_stmt  { $$ = $1;}
                | group_stmt { $$ = $1;}
                | if_stmt    { $$ = $1;}
                | case_stmt  { $$ = $1;}
                | null_stmt  { $$ = $1;};

lvalue_list2:   lvalue_list
                { $$ = $1;}
                | lvalue_list COMMA lvalue_list2
                {
                   struct reg_lval *lv1 = (struct reg_lval *)$1, *lv;
                   for (lv = lv1; lv->next != 0; lv = lv->next)
                      ;
                   lv->next = (struct reg_lval *)$3;
                   $$ = lv1;
                }
                ;

lvalue_list:    reg_lvalue opt_range 
                { $$ = new_reg_lval((struct reg *)$1, (struct bit_range *)$2);}
                | LB lvalue_list2 RB
                { $$ = $2;}
                ;

assign_stmt:    lvalue_list assign_sym { pushline(); }
                opt_delay asexpression2 SEMI
                {
                   struct bitvec *rval = (struct bitvec *)$5;
                   struct reg_lval *lval_list = (struct reg_lval *)$1;
                   struct reg_lval *lhs;
                   int trouble = FALSE;
		   int lineval = popline();
                   stmt_flags non_blocking = (stmt_flags)$<i>2;
                   int lwidth = 0;
                   struct stmt *s = new_stmt(Stmt_Assign, lineval);
                   s->flags |= non_blocking;
                   $$ = s;

                   for (lhs = lval_list; lhs != 0; lhs = lhs->next)
                   {
                      struct reg *lval = lhs->reg;
                      struct bit_range *lrange = lhs->bit_range;
                      int lleft = lrange ? lrange->left : -1;
                      int lright = lrange ? lrange->right : -1;
                      int llo = lval->port->lorange;
                      int lhi = lval->port->hirange;
                      if (lrange != 0) /* a sub-range was specified */
                         if ((lleft < llo) || (lleft > lhi)
                             || (lright < llo) || (lright > lhi))
                         {
                            mesg("** Lvalue selector out of range in line: %d\n",
                                 lineval);
                            free(s);
                            $$ = new_stmt(Stmt_Null, lineval);
                            trouble = TRUE;
                            break;
                         }
                         else
                         {  /* set lo-hi to the specified range: */
                            llo = (lleft < lright) ? lleft  : lright;
                            lhi = (lleft < lright) ? lright : lleft;
                         }
                      lwidth += (1 + ABS(lhi - llo));
                   }
                   if (!trouble && (lwidth != rval->length))
                   {
                      if (lval_list->next == 0)
                         warn("-- WARNING: value assigned to %s has too %s bits in line: %d\n",
                              lval_list->reg->port->name,
                              (lwidth < rval->length) ? "many" : "few", lineval);
                      else
                         warn("-- WARNING: right-hand side of assignment has too %s bits in line %d\n",
                              (lwidth < rval->length) ? "many" : "few", lineval);
                      resize_bitvec(&rval, lwidth); /* fix the situation */
                   }

		   for (lhs = lval_list; (lhs != 0) && !trouble; lhs = lhs->next)
		   {
                      struct reg *lval = lhs->reg;
                      struct bit_range *lrange = lhs->bit_range;
                      int lleft = lrange ? lrange->left : -1;
                      int lright = lrange ? lrange->right : -1;
                      int llo = lval->port->lorange;
                      int lhi = lval->port->hirange;
                      struct bitvec *this_rval = 0;

                      if (lrange != 0) /* a sub-range was specified */
                      {  /* set lo-hi to the specified range: */
                         llo = (lleft < lright) ? lleft  : lright;
                         lhi = (lleft < lright) ? lright : lleft;
                         free(lrange); /* dont need this anymore */
                      }
                      split_bitvec(&rval, &this_rval, 1 + lhi - llo);
                      stmt_assign_expr(s, llo, lhi, lval, this_rval);
                      free(lval);
                   }
                };

reg_lvalue:     IDENTIFIER
                {  /* verify that this has been declared REG */
                   struct reg *r = currentmod->reg;
                   char *id = (char*)$1;
                   int id_len = strlen(id);
                   for (; r != 0; r = r->next)
                   {
                      char *p = strrchr(r->port->name,'/');
                      int n = strlen(r->port->name);
                      if (p != 0)
                         n = p - r->port->name;
                      if ((n == id_len)
                          && strncmp(r->port->name, id, n) == 0)
                         break;
                   }
                   if (r == 0)
                   {
                      mesg("** Signal %s must be of type REG in '%s' on line: %d\n",
                           id, yyfilename, yylineno);
                      $$ = 0;
                   }
                   else
                   {  /* create a copy of reg and return it */
                      struct reg *newr = calloc(1, sizeof(struct reg));
                      struct symslot *sym = symfind(id);
                      newr->port = sym;
                      $$ = newr;
                   }
                };

opt_range:      /* empty */ { $$ = 0; }
                | LBRACK expression RBRACK
                {
                   struct bit_range *r = calloc(1, sizeof(struct bit_range));
                   r->left = r->right = $2;
                   $$ = r;
                }
                | LBRACK expression COLON expression RBRACK
                {
                   struct bit_range *r = calloc(1, sizeof(struct bit_range));
                   r->left = $2;
                   r->right = $4;
                   $$ = r;
                };

assign_sym:     EQ    { $<i>$ = (int)Null_Flag; }
                | LET { $<i>$ = (int)Is_Non_Blocking; };

opt_delay:      /* empty */
                | POUND NUMBER;

group_stmt:     BEGIN_TOKEN { pushline(); } opt_stmt_list END
                {
                   struct stmt *group = new_stmt(Stmt_Group, popline());
                   $$ = stmt_group_append(group, (struct stmt *)$3);
                };

opt_stmt_list:  /* empty */ { $$ = 0; }
                | statement opt_stmt_list
                  {
                     struct stmt *s1 = (struct stmt *)$1;
                     struct stmt *s2 = (struct stmt *)$2;
                     s1->next = s2;
                     if (s2 != 0) s2->prev = s1;
                     $$ = s1;
                  };

if_stmt:        IF LP condition RP { pushline(); } statement opt_else
                {
                   struct stmt *branch = new_stmt(Stmt_Branch, popline());
                   stmt_branch_true(branch, (struct bitvec *)$3,
                                    (struct stmt *)$6);
                   $$ = stmt_branch_false(branch, (struct stmt *)$7);
                };

condition:      asexpression2 { $$ = $1; };

opt_else:       /* empty */      { $$ = 0; }
                | ELSE statement { $$ = $2; };

case_stmt:      case_xz { pushline(); } LP switch_expr RP
                {pragma_expected(TRUE);} opt_pragma1 {pragma_expected(FALSE);}
                choice_list ENDCASE
                {
                   int lineno = popline();
                   struct stmt *this = new_stmt(Stmt_Case, lineno);
                   struct bitvec *switch_expr = (struct bitvec *)$4;
                   struct choice *choice_list = (struct choice *)$9;
                   if (choice_list == 0)
                   {
                      free(this); /* empty case statement */
                      $$ = new_stmt(Stmt_Null, lineno);
                   }
                   else
                   {
                      stmt_flags with_dontcare = (stmt_flags)$1;
                      stmt_flags parallel_case = (stmt_flags)$7;
                      stmt_case_setflag(this, with_dontcare);
                      stmt_case_setflag(this, parallel_case);
                      stmt_case_init(this, switch_expr, choice_list);
                      $$ = this;
                   }
                };

case_xz:        CASE    { $$ = (int)Null_Flag; }
                | CASEX { $$ = (int)Case_With_Dontcare; }
                | CASEZ { $$ = (int)Case_With_Dontcare; };

switch_expr:    asexpression2;

opt_pragma1:    /* empty */ { $$ = 0; }
                | PRAGMA
                {
                   char *prag = (char *)$1;
                   if (strcasecmp(prag,"parallel_case") == 0)
                      $$ = (int)Case_Parallel;
                   else
                   {
                      warn("-- WARNING: pragma %s not recognized in '%s' on line: %d\n",
                           prag, yyfilename, yylineno);
                      $$ = (int)Null_Flag;
                   }

		   free(prag);
                };

choice_list:    /* empty */ { $$ = 0; }
                | choice choice_list
                {
                   struct choice *left = (struct choice *)$1;
                   struct choice *right = (struct choice *)$2;
                   left->next = right;
                   $$ = left;
                };

choice:         selector { pushline(); } statement
                {
                   struct choice_selector *sel = (struct choice_selector *)$1;
                   struct stmt *st = (struct stmt *)$3;
                   $$ = new_choice(sel, st, popline());
                };

selector:       value_list COLON    { $$ = $1; };
                | DEFAULT opt_colon { $$ = 0; };

value_list:     asexpression2
                {
                   struct bitvec *const_val = (struct bitvec *)$1;
                   $$ = new_choice_selector(const_val);
                }
                | asexpression2 COMMA value_list
                {
                   struct bitvec *const_val = (struct bitvec *)$1;
                   struct choice_selector *s = new_choice_selector(const_val);
                   s->next = (struct choice_selector*)$3;
                   $$ = s;
                };

opt_colon:      /* empty */ | COLON;

null_stmt:      SEMI {$$ = new_stmt(Stmt_Null, yylineno);};

%%
