/* Copyright (C) 1996 Himanshu M. Thaker

This file is part of vpp.

Vpp is distributed in the hope that it will be useful,
but without any warranty.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.

Everyone is granted permission to copy, modify and redistribute
vpp, but only under the conditions described in the
document "vpp copying permission notice".   An exact copy
of the document is supposed to have been given to you along with
vpp so that you can know how you may redistribute it all.
It should be in a file named COPYING.  Among other things, the
copyright notice and this notice must be preserved on all copies.  */

/*
 * $Header: /home/cvspsrv/cvsroot/eda/vbpp/vpp.c,v 1.2 2001/01/25 07:38:03 jching Exp $
 * Program :
 * Author : Himanshu M. Thaker
 * Date : Apr. 16, 1995
 * Description :
 * $Log: vpp.c,v $
 * Revision 1.2  2001/01/25 07:38:03  jching
 * Return non-zero when parsing error detected.
 *
 * Revision 1.1.1.1  2000/02/21 04:05:37  ahvezda
 * Import of vbpp
 *
 * Revision 1.3  1995/04/21  18:11:58  hemi
 * *** empty log message ***
 *
 * Revision 1.3  1995/04/21  18:11:58  hemi
 * *** empty log message ***
 *
 * Revision 1.2  1995/04/18  00:23:43  hemi
 * *** empty log message ***
 *
 * Revision 1.1  1995/04/16  19:38:15  hemi
 * Initial revision
 */


/*
 * include files...
 */
#include "vpp_common.h"
#include "vpp_yacc_stuff.h"
#include <stdio.h>
#include <string.h>

static char *stdin_name="<stdin>";
struct preproc_buff_t *preproc_buff_head=NULL, *preproc_buff_current=NULL;
extern int vppcommentdepth, vppcommentline;

/*
 * printf to memory..
 */
int mprintf(const char *fmt, ... )
{
int rc;
va_list args;
struct preproc_buff_t *bt = (struct preproc_buff_t *)calloc(1, sizeof(struct preproc_buff_t));
char buff[65537];

va_start(args, fmt);
rc=vsprintf(buff, fmt, args);
bt->len = strlen(buff);
bt->str = malloc(bt->len+1);
strcpy(bt->str, buff);

if(!preproc_buff_current)
	{
	preproc_buff_head = preproc_buff_current = bt;
	}
	else
	{
	preproc_buff_current->next = bt;
	preproc_buff_current = bt;	
	}

return(rc);
}

/*
 * get maxsiz chars from memory and copy into buf
 * (return actual number of characters copied)
 */
int mgetstr(char *buf, int maxsiz)
{
struct preproc_buff_t *bt = preproc_buff_head;
int cpylen;

if(!bt) return(0);

cpylen = bt->len-bt->pos;
if(cpylen <= maxsiz)
	{
	memcpy(buf, bt->str + bt->pos, cpylen);
	free(bt->str);

	if(bt != preproc_buff_current)
		{
		preproc_buff_head = bt->next;
		}
		else
		{
		preproc_buff_head = preproc_buff_current = NULL;
		}
	free(bt);
	}
	else
	{
	memcpy(buf, bt->str + bt->pos, maxsiz);
	bt->pos += (cpylen=maxsiz);
	}

return(cpylen);
}

/*
 * sample print to stdout of memory
 */
void mstdout(void)
{
char buff[16];
int rc;

while((rc=mgetstr(buff, 15)))
	{
	buff[rc] = 0;
	printf("%s", buff);
	}
}


int main(int argc, char **argv)
{
  int i;
  int use_stdin = FALSE;
  struct vpp_filename *vpp_filename_head=NULL, *vpp_filename_current=NULL;

  if(argc<2)
	{
	use_stdin = TRUE;
	}
	else
	{
	for(i=1;i<argc;i++)
		{
		if(!strcmp(argv[i], "-h"))
			{
			fprintf(stderr, "Usage:\n------\n%s [options] [files]\n\n"
					"-\t\taccept input from stdin\n"
					"-nc\t\tdo not generate numbered comments\n"
					"-exp\t\texpand `define macros\n"
					"-license\tprint vpp usage license to stdout\n\n",
					argv[0]);
			exit(0);
			}
		else
		if(!strcmp(argv[i], "-nc"))
			{
			dont_add_comments = TRUE;
			}	
		else
		if(!strcmp(argv[i], "-exp"))
			{
			expand_defines = TRUE;
			}	
		else
		if(!strcmp(argv[i], "-"))
			{
			use_stdin = TRUE;
			}
		else
		if(!strcmp(argv[i], "-license"))
			{
			print_license();
			exit(0);
			}
		else
			{
			if(!vpp_filename_current)
				{
				vpp_filename_head = vpp_filename_current = calloc(1, sizeof(struct vpp_filename));
				vpp_filename_current->name = argv[i];
				}
				else
				{
				struct vpp_filename *t = calloc(1, sizeof(struct vpp_filename));
				t->name = argv[i];
				vpp_filename_current->next = t;
				vpp_filename_current = t;
				}
			}
		}
	}

if ((use_stdin) && (vpp_filename_head))
		{
		fprintf(stderr, "'%s': using stdin and a file list is not allowed, exiting.\n", argv[0]);
		exit(255);
		}


	fprintf(stderr, "VPP: Verilog Preprocessor 1.0.4-6 Copyright (C) 1996 Himanshu M. Thaker\n\n");

if(use_stdin)
	{
        nl_count = 1;
        current_file = malloc(strlen(stdin_name)+1);	/* otherwise vbpp crashes! */
	strcpy(current_file, stdin_name);
        fprintf(stderr, "Preprocessing '%s'\n",stdin_name);
        do_comment_count(1,nl_count);
	wwin = stdin;
	wwparse();
	mstdout();
	if(vppcommentdepth)
		{
		printf("\n*/  // <-- UNTERMINATED COMMENT FROM LINE %d FORCED CLOSE!!\n\n", vppcommentline);
		vppcommentdepth;
		fprintf(stderr,"ERROR : Unterminated Comment in file %s\n", stdin_name);
		++wwerror_count;
		}
    	}
    	else
    	{
	vpp_filename_current = vpp_filename_head;
	while(vpp_filename_current) 
		{
    		nl_count = 1;
    		fprintf(stderr, "Preprocessing '%s'\n", vpp_filename_current->name);
    		if ((wwin = fopen(vpp_filename_current->name, "r")) == NULL) 
			{
      			fprintf(stderr,"ERROR : could not open file %s\n", vpp_filename_current->name);
    			}
    			else 
			{
      			current_file = malloc(strlen(vpp_filename_current->name)+1);
      			strcpy(current_file, vpp_filename_current->name);
      			do_comment_count(1,nl_count);
      			wwparse();
      			fclose(wwin);
			mstdout();
			if(vppcommentdepth)
				{
				printf("\n*/  // <-- UNTERMINATED COMMENT FROM LINE %d FORCED CLOSE!!\n\n", vppcommentline);
				vppcommentdepth=0;
				fprintf(stderr,"ERROR : Unterminated Comment in file %s\n", current_file);
				++wwerror_count;
				}
      			free(current_file);
    			}
		vpp_filename_current=vpp_filename_current->next;
  		}
   	}

exit(wwerror_count);
}

