
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        sequ.c
**
**   Functions:   enqueue_seq (static), node_effects (static)
**                condition, propogate_effects, sig_assign_1,
**                sig_assign_2, var_assign_1, var_assign_2
**
**   Description:
**
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**		  Anthony J. Bybell(AJB) - Penn State University
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial code writing and debug.
**   08/04/88    ARM      Code cleanup; added comments.
**   08/30/97    AJB	  Removed compiler warnings.
**
**
*/

#include "net.h"

static nlptr    prop_list;
	
void node_effects (nptr); /* proto def gets rid of compiler warnings */


/*
**   Function:     enqueue_seq (static)
**
**   File:         sequ.c
**
**   Called by:    sig_assign_1 (sequ.c), sig_assign_2 (sequ.c)
**
**   Description:
**
**   Arguments:    n -
**                 newvalue -
**                 delta -
**                 delay_type -
**
**   Return Value: None.
*/
 /*
  * assign node its new value and add it to the list of nodes whose values
  * are to be propagated
  */
static
enqueue_seq (n, newvalue, delta, delay_type)
    register nptr   n;
    int             newvalue;
    long            delta;
    int             delay_type;
{
    char            nv;
    long            etime;

    if (newvalue == 0)
	nv = '0';
    else
	nv = (char) newvalue;

    if (delta)
    {				/* delay so let event queue take over	 */
	new_out (n, newvalue, delta, delay_type);
    }
    else
    {
	/* need to delete old events as needed 	 */
/***
***  THIS IS WRONG!!!   this call has the wrong number of arguments
*/
	etime = cur_delta + delta;

	delete_events (n, etime, delay_type);

	/* change inited flag if node was inited 	 */
	if (n->nflags & INITED)
	    n->nflags &= ~INITED;

	/* give node its new value and output trace if necessary	 */
	if ((n->nflags & TRACED) || (debug & DEBUG_SYSTEM))
	    lprintf (outfile, "[event #%ld] node %s: %c -> %c @ %2.1fns\n",
		     nevent, n->nname, n->nvalue, nv, 0.1 * etime);

	/*  only update the time if we are really changing value  */
	if (n->nvalue != nv)
	  {
	    n->ctime = etime;
	    n->delta_stamp = delta_timestamp;

	    /*  clear the rising and falling bits  */ 
	    n->nflags &= ~(FLAG_GATE_RISING | FLAG_GATE_FALLING);

	    if (n->nvalue == '1') {
	      /*  we are transitioning from 1 to 0  */
	      n->nflags |= FLAG_GATE_FALLING;

	    } else if (n->nvalue == '0') {
	      /*  we are transitioning from 0 to 1  */
	      n->nflags |= FLAG_GATE_RISING;

	    }
	  }

	/*
	 *    HERE IS WHERE WE ASSIGN THE NEW VALUE TO THE NODE
	 */
	n->nvalue = nv;

	/* add to list of nodes which changed value during process	 */
	node_effects (n);

	/* fix */
	/* see if we want to halt if this node changes value	 */
	if (n->nflags & STOPONCHANGE)
	    ;
    }
}


/*
**   Function:     node_effects (static)
**
**   File:         sequ.c
**
**   Called by:
**
**   Description:
**
**   Arguments:
**
**   Return Value: None.
*/
void
node_effects (n)
    nptr            n;
{
    nlptr           nl;
    int             inlist = 0;

    if (prop_list)
    {
	for (nl = prop_list; nl; nl = nl->next)
	    if (nl->this = n)
	    {
		inlist = 1;
		break;
	    }
	if (!inlist)
	{
	    nl->next = (nlptr) malloc (sizeof (struct nlist));
	    nl = nl->next;
	    nl->this = n;
	    nl->next = (nlptr) 0;
	}
    }
    else
    {
	prop_list = (nlptr) malloc (sizeof (struct nlist));
	prop_list->this = n;
	prop_list->next = (nlptr) 0;
    }
}


/*
**   Function:     condition
**
**   File:         sequ.c
**
**   Called by:    user written 'process' blocks
**                 compiled by the user
**
**   Description:
**
**   Arguments:
**
**   Return Value: None.
*/
int
condition (s)
    char           *s;
{
    nptr            n = find (s);

    if (n->nvalue == '1' )
	return 1;
    else
	return 0;	/* 0, u, x, z */
}


/*
**   Function:     propagate_effects
**
**   File:         sequ.c
**
**   Called by:    user written 'process' blocks
**                 compiled by the user
**
**   Description:
**
**   Arguments:
**
**   Return Value: None.
*/
propagate_effects ()
{
    nlptr           nl;
    oblptr          bl;

    for (nl = prop_list; nl; nl = nl->next)
    {
	/*
	 * for each object controlled by event node, evaluate object and post
	 * new events
	 */
	for (bl = nl->this->in; bl != NULL; bl = bl->next)
	    eval_obj (bl->this);
    }
}


/*
**   Function:     sig_assign_1
**
**   File:         sequ.c
**
**   Called by:    user written 'process' blocks
**                 compiled by the user
**
**   Description:  assign the signal s2 to signal s1;
**                 enqueue event for the value of
**                 s2 to be assigned to s1
**
**   Arguments:    s1 - the signame being assigned to
**                 s2 - the signal being assigned from
**                 delay_type - type of delay
**                 delay - delta time from now for
**                         assignment to occur
**
**   Return Value: None.
*/
sig_assign_1 (s1, s2, delay_type, delay)
    char           *s1;
    char           *s2;
    int             delay_type;
    long            delay;
{
    nptr            n1 = find (s1);
    nptr            n2 = find (s2);

    inited_node (n2);
    enqueue_seq (n1, n2->nvalue, delay, delay_type);
}


/*
**   Function:     sig_assign_2
**
**   File:         sequ.c
**
**   Called by:    user written 'process' blocks
**                 compiled by the user
**
**   Description:  perform operation: signal = variable
**                 enqueue an event to assign the value
**                 of variable 'v' to signal 's'
**
**   Arguments:    s - the signal being assigned to
**                 v - the variable being assigned from
**                 delay_type - type of delay
**                 delay - delta time from now for
**                         assignment to occur
**
**   Return Value: None.
*/
sig_assign_2 (s, v, delay_type, delay)
    char           *s;
    char            v;
    int             delay_type;
    long            delay;
{
    nptr            n = find (s);

    enqueue_seq (n, v, delay, delay_type);
}


/*
**   Function:     var_assign_1
**
**   File:         sequ.c
**
**   Called by:    user written 'process' blocks
**                 compiled by the user
**
**   Description:  routine to assign the value of
**                 variable v1 to be that of variable v2
**
**   Arguments:    v1 - variable to be assigned to
**                 v2 - variable to be assigned from
**
**   Return Value: None.
*/
var_assign_1 (v1, v2)
    int            *v1;
    int             v2;
{
    *v1 = v2;
}


/*
**   Function:     var_assign_2
**
**   File:         sequ.c
**
**   Called by:    user written 'process' blocks
**                 compiled by the user
**
**   Description:  assign the value of signal s
**                 to variable v
**
**   Arguments:    v - the variable assigned to
**                 s - the signal the value is
**                     assigned from
**
**   Return Value: None.
*/
var_assign_2 (v, s)
    int            *v;
    char           *s;
{
    nptr            n = find (s);

    inited_node (n);
    *v = n->nvalue;
}
